;Lawson Labs, Inc.
;15 Phoenixville Pike                             DRIVR250  Library  Rev. 5
;Malvern, PA  19355                               For microcode version 6
;406 257-5355  or  800 725-8800                   September 12, 1996
;
;
;!* marks the start of a block of code
;I* marks initialization routines        M* marks misc routines
;S* marks scanning routines              C* marks self-calibration routines
;L* marks load .cal file routines        D* marks display routines
;T* marks timer interrupt routines
;*****************************************************************************
;
; LPT port interface definitions:
;
;       host writes right-justified data to output pins
;       host reads left-justified data from input pins
;       the most significant data bit must be inverted after read
;
;       BIT #                7  6  5  4  3  2  1  0           ADDRESS
;
;       OUTPORT: DB25 PIN #  9  8  7  6  5  4  3  2     port 3BC 378 278
;                           A1 A0 RD WR D3 D2 D1 D0
;                           __
;       INPORT:  DB25 PIN # 11 10 12 13                 port 3BD 379 279
;                           D7 D6 D5 D4
;
;       A0 and A1 select the parallel register for both read and write
;       RD and WR are active low enables
;
;       A0 & A1 select one of four 4-bit registers
;
;       Only the OUTPORT address need be referenced for writing.
;       Data nibble to write should be at D3, D2, D1 & D0
;
;       To write, combine register selection and data nibbles and
;       place at OUTPORT with RD and WR strobes inactive (high).
;       Then, drop and raise write strobe with data and register
;       selection bits still in place.
;
;       Register 0 is used as a status register. The other three 
;       registers are used for passing commands and/or data.
;       The acknowledge and newdata bits are used for handshaking.
;
;          D7     D6     D5     D4     D3        D2        D1        D0
;          A1     A0     RD     WR          
; Status    0      0                  error   stop scan  NewData   HostAck
; Reg1      0      1                  data      data      data      data 
; Reg2      1      0                  data      data      data      data 
; Reg3      1      1                  data      data      data      data 
;
;
;      To read, first select register (with RD strobe low) at OUTPORT.
;      Then, input data from selected register at INPORT.  
;      Return RD strobe high when done.
;
;       Status definition is slightly different for read operations
;
;          A1     A0                   D7        D6        D5        D4
;           0      0    (Status)    not error   BIT13    NewData   M250Ack 
;
;
; ERRORCODE% definitions    0  =  NO ERROR
;               10h =  M250 raised error bit while host was reading
;               11h =  timeout while host was attempting to read
;               12h =  no response to reset, reset failed
;               13h =  timeout while host was attempting to write
;               14h =  M250 raised error bit while host was writing
;               15h =  invalid command word 
;               16h =  dac data is out of range (self-calibration)
;               17h =  checksum wrong
;               18h =  time stamp scan checksum wrong (internal)
;               19h =  abort time stamp scan (ESC key) (internal)
;               1Ah =  average loop shift counter = 0
;               1Bh =  time stamp scanning buffer full
;               1Ch =  time stamp scanning buffer empty
;               1Dh =  digitial input pending flag
;               1Eh =  M250 error raised in time stamp scanning. Data lost
;*****************************************************************************
.MODEL  medium
.DATA
;!***********************   START OF DATA SEGMENT   **************************
protect_array   DB 256 DUP(?)   ;PROTECT QUICKBASIC IDE ENVIRONMENT 
OUTPORT         DW ?            ;port address for data to the M250
INPORT          DW ?            ;port address for data from the M250
                                 
ERROR_OFS       EQU [BP + 06H]  ;offset to ERRORCODE%
ERROR_SEG       EQU [BP + 08h]  ;segment of ERRORCODE%    
L_RESPONSE_OFS  EQU [BP + 0AH]  ;offset to RESPONSE& (or DATA1&)
L_RESPONSE_SEG  EQU [BP + 0CH]  ;segment of RESPONSE& (or DATA1&)
COMMAND_OFS     EQU [BP + 0EH]  ;offset to COMMAND%
COMMAND_SEG     EQU [BP + 10H]  ;segment of COMMAND%

HIGH12          DW ?            ;HIGH12 word of conversion
LOW14           DW ?            ;LOW14 word of conversion
RESPONSE        DW ?            ;response from M250
CHECKSUM        DW 0            ;current running checksum
RWDELAY         DW 0            ;delay for read/write (long cables)
STATUS          DB 30h          ;write to outport to select status reg.
NEWDATA         DB 0            ;toggle bit for new data from M250 (20h or 0)
M250ACK         DB 10h          ;toggle bit for M250 ready to read command
OUTB1           DB 0            ;first nibble of command
OUTB2           DB 0            ;second nibble of command
OUTB3           DB 0            ;third nibble of command
ERRORCODE       DB 0            ;error code byte
HOLD            DB 1            ;track/hold status  set to hold
AVERAGE1        DB 0            ;current low nibble of remote average word
AVERAGE2        DB 0            ;current high nibble of remote average word
CHANNEL         DB 0            ;current channel selected
EXTCODE         DB 0            ;current external control code selected

OFFSET_HI       DW 0            ;high part of offset factor word
OFFSET_LO       DW 0            ;low part of offset factor word
OFFSET_FLAG     DW 0            ;flag for offset channel 
FULLSCALE_HI    DW 0            ;high part of full scale count
FULLSCALE_LO    DW 0            ;low part of full scale count
POWER2          DW 0            ;power of 2 for calibrations

TESTDATA        DW 0            ;current dac value for self-calibration
DACDATA         DW 0            ;temp dac value
HOLDSTEP        DW -10          ;holdstep value
MONODELAY       DW 0            ;self-cal delay in usec (minimum of 17)
AVRGSHIFT       DB 6            ;shift value for AVRG (inner) loop counter
AVRG1SHIFT      DB 6            ;shift value for AVRG1 (outer) loop counter
AVRG            DB 64           ;inner loop counter for self-cal
AVRG1           DB 64           ;outer loop counter for self-cal
TOTREMLO        DW ?            ;low part of remainder
TOTREMHI        DW ?            ;high part of remainder
TEMPYHI         DW ?            ;low part of y remainder
TEMPYLO         DW ?            ;high part of y remainder
TEMPXHI         DW ?            ;low part of x remainder
TEMPXLO         DW ?            ;high part of x remainder
TEMPY           DW ?            ;temporary storage of average y remainder
SINGLECAL       DB ?            ;flag for 1 calibration point
Y_HI_BOUND      DW 1800h        ;upper bounds for y remainders
Y_LO_BOUND      DW 0C00h        ;lower bounds for y remainders
X_HI_BOUND      DW 2000h        ;upper bounds for x remainders
X_LO_BOUND      DW 1400h        ;lower bounds for x remainders
COR_16          DW 10h          ;mask for filling block copy array
MASK_16         DW 10h          ;mask for exception code
STEP_512        DW 200h         ;fill_table step size
TEMPXY_ARRAY    DW 20h DUP (0)  ;block copy array for selfcal

CHANNEL0        DW 0            ;channel configuration block for scanning
CHANNEL1        DW 0            ;store channel control codes here
CHANNEL2        DW 0
CHANNEL3        DW 0
CHANNEL4        DW 0
CHANNEL5        DW 0
SCANSET         DW 0            ;scan mode command 
                      ;(# of channels to scan for MULTI-CHANNEL FAST SCAN)
                      ;(MULTI-CHANNEL FAST SCAN IS NOT CURRENTLY IMPLEMENTED) 
SCANHI          DW 1            ;inter-scan hi delay
SCANDELAYS      DW 100h         ;inter-scan delay and scan rate 
INTERCHAN       DB 1            ;inter-channel delay  (multiples of 17 Usec)
NUM_DATAPOINTS  DB 1            ;# of points for normal/time stamp scanning
TOT_SCANS       DW 1            ;total scans for normal scans
                                ;buffer size for time-stamped scan
NUMSCANS        DW ?            ;# of scans performed
NSCANLO         DW 0
NSCANHI         DW 0

START_SEG       DW ?            ;starting segment for scan storage
FAST_SCAN_FLAG  DB 0            ;1 = fast scan   0 = normal/time stamp scan
STOP_FLAG       DB 0            ;flag for end of scan
TRIGGERTYPE     DB 0            ;type of trigger 0 = none  
                                ;                1 = positive edge
                                ;                2 = negative edge
OFS_HI          DW 0            ;high part of scan offset factor word
OFS_LO          DW 0            ;low part of scan offset factor word
ZCOR_HI         DW 0            ;high part of scan zero correction
ZCOR_LO         DW 0            ;low part of scan zero correction

CURRENT_DI      DW 0            ;current DI index for time stamp scanning
CURRENT_ES      DW ?            ;current ES segment for time stamp scanning
STAMP_ARRAY     DW 200h DUP (0) ;store one time stamp scan
TICK_COUNTER    DW 0            ;counter for when to service old INT 8
DO_OLDINT8      DW 400H         ;compare for servicing INT 8
HICOUNT         DB 00H          ;high byte for 8254 timer
LOCOUNT         DB 40H          ;low byte for 8254 timer
HEAD_POINTER    DW 0            ;offset into timer array for writing
TAIL_POINTER    DW 0            ;offset into timer array for reading
WRAP_POINTER    DW 0            ;counter of number of wraps
START_ISR       DB 0            ;flag for gathering data in ISR
                                ;0 = gather data
                                ;1 = stop gathering data
                                ;2 = abort gathering data
ISR_ERROR       DB 0            ;error flag for ISR
STAMP_CHKSUM    DB 0            ;checksum for one time stamped scan
STORE_CHKSUM    DB 0            ;place to store the time stamp checksum
END_OFFSET      DW ?            ;offset to the end of the buffer
END_SEGMENT     DW ?            ;segment for the end of the buffer
DO_DIG_CMDS     DB 0            ;Flag for digin and digout commands during 
                                ;time-stamped interrupt scanning
DIGOUTWORD      DW 0            ;digital output word to send
DIGIN_WORD      DW 0            ;digital input word to send back to host
FIRST_WORD      DW 0            ;temp storage of a possible digital input word
DIGIN_STATUS    DB 0            ;0= no digin command 1 = digin command pending
                                ;2= got digin, not sent to host
STORE_1ST_WORD  DW 0            ;first half of last time stamp number
STORE_2ND_WORD  DW 0            ;first half of last time stamp number
S_1ST_WORD      DW 0            ;temp storage of first half of time stamp #
S_2ND_WORD      DW 0            ;temp storage of second half of time stamp #
INTERRUPT_FLAG  DB 0            ;flag for current status of INT 8 vector
NUM_OF_SCANS    DW 0            ;number of scans to get at a time
LESS_FLAG       DB 0

SRC_HANDLE      DW 0                    ;source file handle
COR_FILE        DB "250AXXXX.CAL",0     ;correction array file name
COR_ARRAY       DW 1000h DUP (0)        ;correction array
CORERR          DB 0                    ;correction array file read error
ABS_FILE        DB "FACTXXXX.CAL",0     ;absolute accuracy file name
ABS_ARRAY       DW 2000h DUP (0)        ;absolute accuracy array
ABSERR          DB 0                    ;absolute acc. array file read error

.FARDATA?                       ;segment to store timer interrupt buffer
TIMER_BUFFER    DB 0FFFFh DUP(0)
.CODE
PUBLIC M250
;!********  standard BASIC interface, save registers and set stack **********

M250    PROC
        
        PUSH    BP                      ;entry sequence - save old BP
        MOV     BP, SP                  ;set stack framepointer
        PUSH    BX                      ;store registers on stack
        PUSH    CX
        PUSH    SS
        PUSH    DS
        PUSH    ES
        PUSH    SI
        PUSH    DI
        PUSHF                           ;store flags on stack
        MOV     ERRORCODE, 0            ;clear ERRORCODE
        MOV     RESPONSE, 0             ;clear RESPONSE
        MOV     ES, COMMAND_SEG         ;segment and offset to command
        MOV     BX, COMMAND_OFS
        MOV     AX, ES:[BX]             ;get COMMAND word

; ***************  parse COMMAND word  ******************

SETPORTCMD:                             ;set port command ?
        CMP     AX, 0F000h              
        JB      SENDSETUPCMD
        CALL    INIT                    ;initialize model 250
        JMP     SEND_RESPONSE           ;send back response and error to host

SENDSETUPCMD:                           ;send scan configuration block ?
        CMP     AX, 0E000h              
        JB      SETSCANCHANNELSCMD
        CALL    SENDMODE
        JMP     SEND_RESPONSE           ;send back response and error to host

SETSCANCHANNELSCMD:                     ;setup scan configuration block ?
        CMP     AX, 0D000h
        JB      SETRATECMD
        AND     AL, 0FH                 ;get index
        MOV     OUTB1, AL               ;store index location
        CALL    SETSCANCHANNELS
        JMP     SEND_RESPONSE           ;send back response and error to host

SETRATECMD:                             ;setup scan rate ?
        CMP     AX, 0C000h
        JB      LASTBYTE
        CALL    SETSCANRATE
        JMP     SEND_RESPONSE           ;send back response and error to host

LASTBYTE:                               ;get last byte ?
        CMP     AX, 0B000h
        JB      GETCALFILESCMD
        CALL    CHK1                    ;read the last byte
        JMP     SEND_RESPONSE           ;send back response and error to host

GETCALFILESCMD:                         ;get cal files ?
        CMP     AX, 0A000h
        JB      GETSTAMPSCANCMD
        MOV     SI, OFFSET COR_FILE     ;get offset to cal file name
        ADD     SI, 3                   ;index to letter
        CMP     AL, 0                   ;is there a character ?
        JE      NOT_LETTER              ;no, skip storing letter
        MOV     [SI], AL                ;yes, store the letter
NOT_LETTER:
        INC     SI                      ;increment index to start of serial #
        CALL    GET_SERIAL              ;get the serial number and put it in
                                        ;the cal file name
        CALL    GET_COR_FILE            ;load correction array
        
        MOV     SI, OFFSET ABS_FILE     ;get offset to fact cal file name
        ADD     SI, 4                   ;index to start of serial #
        CALL    GET_SERIAL              ;get the serial number and put it in
                                        ;the factory file name
        CALL    GET_ABS_FILE            ;load array
        
        MOV     AL, CORERR              ;send back file read errors, if any
        MOV     AH, ABSERR
        MOV     ES, ERROR_SEG           ;segment and offset to ERRORCODE%
        MOV     BX, ERROR_OFS
        MOV     ES:[BX], AX             ;load file errors, if any
        JMP     EXIT_M250               

GETSTAMPSCANCMD:                        ;get one time stamp scan ?
        CMP     AX, 09000h
        JB      ENDSTAMPSCANCMD
        CMP     AL, 1
        JE      GET_MANY_SCANS
        CALL    GET_TIMER_SCAN
        JMP     SEND_ERROR              ;send an error, if any

GET_MANY_SCANS:                         ;get RESPONSE scans
        MOV     ES, L_RESPONSE_SEG
        MOV     BX, L_RESPONSE_OFS
        MOV     CX, ES:[BX]             ;number of scans to get in RESPONSE
        CMP     CX, 0
        JNE     NOT_ZERO_SCANS
        MOV     CX, 1                   ;if CX = 0 then CX = 1
NOT_ZERO_SCANS:
        MOV     NUM_OF_SCANS, CX
GET_TIMER_AGAIN:
        PUSH    CX
        CALL    GET_TIMER_SCAN
        POP     CX
        CMP     ERRORCODE, 0
        JNE     MANY_SCANS_ERROR
        LOOP    GET_TIMER_AGAIN
        JMP     EXIT_MANY_SCANS
MANY_SCANS_ERROR:        
        MOV     AX, NUM_OF_SCANS
        SUB     AX, CX
        MOV     ES, L_RESPONSE_SEG
        MOV     BX, L_RESPONSE_OFS
        MOV     ES:[BX], AX
EXIT_MANY_SCANS:        
        JMP     SEND_ERROR

ENDSTAMPSCANCMD:                        ;end time stamp scan ?        
        CMP     AX, 08000h
        JB      CALLOFFSETCMD
        CALL    END_TIMER_SCAN
        JMP     SEND_ERROR              ;send an error, if any

CALLOFFSETCMD:                          ;offset calibration ?
        CMP     AX, 07000h
        JB      CALLFULLSCALECMD
        CALL    OFFSET_CAL              ;get offset calibration
        JMP     SEND_ERROR              ;send an error, if any

CALLFULLSCALECMD:                       ;fullscale calibration ?
        CMP     AX, 06000h
        JB      SEND_NUM_BYTES
        CALL    FULLSCALE_CAL           ;get fullscale calibration
        JMP     SEND_ERROR              ;send an error, if any

SEND_NUM_BYTES:
        CMP     AX, 05000h
        JB      STOP_INTERRUPT_SCAN
        CALL    GET_NUM_WORDS           ;returns # of bytes in AX
        MOV     RESPONSE, AX
        CMP     BX, 0
        JE      NOT_FULL
        MOV     ERRORCODE, 1BH
NOT_FULL:
        MOV     CX, BX
        MOV     ES, L_RESPONSE_SEG      ;segment and offset to response
        MOV     BX, L_RESPONSE_OFS
        MOV     ES:[BX], AX             ;store number of bytes in buffer
        MOV     ES:[BX + 2], CX
        JMP     SEND_ERROR

STOP_INTERRUPT_SCAN:
        CMP     AX, 04000h
        JAE     DO_DIGIN_STUFF
        JMP     SET_TIMER_FREQ
DO_DIGIN_STUFF:
        CMP     AL, 3                   
        JAE     DIGIN_CMDS              ;if AL >= 3 then digital commands
        MOV     START_ISR, AL           ;stop isr 1 = stop gathering scans
                                        ;         2 = abort current scan
        JMP     EXIT_M250

DIGIN_CMDS:                             ;CHECK DIGITAL COMMANDS
        CMP     AL, 3                   ;if AL = 3 then send digital ouput
        JNE     IS_DIGIN_CMD
        MOV     ES, L_RESPONSE_SEG      ;AL = 3  send digital output
        MOV     BX, L_RESPONSE_OFS
        MOV     AX, ES:[BX]             ;get digital output word 
        MOV     DIGOUTWORD, AX          ;digitial output word to send
        MOV     DO_DIG_CMDS, 3          ;signal send digital output
NOT_SENT_YET:        
        CMP     DO_DIG_CMDS, 0         ;wait for send of digital output ack
        JNE     NOT_SENT_YET
        JMP     SEND_RESPONSE           ;check to see if buffer is full
IS_DIGIN_CMD:                           ;send digital input request, wait      
        CMP     DIGIN_STATUS, 1
        JE      GET_FIRST               ;command has been sent, attempt to
                                        ;get digital input word
        CMP     DIGIN_STATUS, 2         ;got the word, send it back
        JE      SEND_BACK_DIGIN
        MOV     DO_DIG_CMDS, 4          ;sends digital input command in ISR
NOT_SENT_YET1:        
        CMP     DO_DIG_CMDS, 0         ;wait for send of digital input request
        JNE     NOT_SENT_YET1
GET_FIRST:        
        CALL    CHK_FOR_DIGIN           ;check for digital input word
SEND_BACK_DIGIN:        
        CMP     ERRORCODE, 0            ;if errorcode = 1Dh then digin pending
        JE      GOT_DIGIN
        JMP     SEND_RESPONSE
GOT_DIGIN:
        MOV     AX, DIGIN_WORD          ;else if errorcode = 0 then send back
        MOV     RESPONSE, AX            ;digital input word
        MOV     STAMP_CHKSUM, 0
        MOV     DIGIN_STATUS, 0
        JMP     SEND_RESPONSE


SET_TIMER_FREQ:                         ;SETUP INTERRUPT/ RESTORE INTERRUPT
        CMP     AX, 03000h
        JB      CMDS_2000h
        CMP     AL, 1                   ;IF AL = 1 THEN SETUP INTERRUPT
        JE      SET_INTERRUPT
        CMP     AL, 2                   ;IF AL = 2 THEN RESTORE INTERRUPT
        JE      RESTORE_INTERRUPT
        MOV     ES, L_RESPONSE_SEG      ;AL = 0 THEN SETUP 
        MOV     BX, L_RESPONSE_OFS
        MOV     AX, ES:[BX]
        MOV     LOCOUNT, AL             ;setup timer counter word
        MOV     HICOUNT, AH
        MOV     AX, ES:[BX + 2]
        MOV     DO_OLDINT8, AX          ;setup compare for jump to old INT 8
        JMP     EXIT_M250
SET_INTERRUPT:
        MOV     RESPONSE, 2             ;flag already set
        MOV     START_ISR, 1            ;flag don't start gathering scans yet
        CMP     INTERRUPT_FLAG, 1
        JE      ALREADY_SET
        CALL    SET_INT8
        MOV     RESPONSE, 1             ;flag interrupt set
ALREADY_SET:
        JMP     SEND_RESPONSE

RESTORE_INTERRUPT:
        MOV     RESPONSE, 2             ;flag interrupt was not set
        CMP     INTERRUPT_FLAG, 1
        JNE     NOT_SET
        CALL    RESTORE_INT8
        MOV     RESPONSE, 1             ;flag interrupt restored
NOT_SET:        
        JMP     SEND_RESPONSE

CMDS_2000H:                             ;COMMAND 2000h INVALID
        CMP     AX, 02000h
        JB      TESTCONVCMD
        JMP     INVALID_COMMAND         ;flag invalid command

TESTCONVCMD:                            ;test conversion command ?
        CMP     AX, 01000h              ;for troubleshooting only 
        JB      PARSE_CMD_WORD
        MOV     CX, LOW14               ;put LOW14 and HIGH12 in response
        MOV     AX, HIGH12
        MOV     ES, L_RESPONSE_SEG      ;segment and offset to response
        MOV     BX, L_RESPONSE_OFS
        MOV     ES:[BX], CX             ;LOW14
        MOV     ES:[BX + 2], AX         ;HIGH12
        JMP     EXIT_M250               

PARSE_CMD_WORD:                         ;partition command word
        MOV     OUTB1, AL
        AND     OUTB1, 0Fh              ;first nibble of command
        SHR     AL,1                    ;DIVIDE BY 16
        SHR     AL,1
        SHR     AL,1
        SHR     AL,1
        AND     AL, 0Fh
        MOV     OUTB2, AL               ;second nibble of command
        AND     AH, 0Fh
        MOV     OUTB3, AH               ;third nibble of command
        
        CMP     OUTB3, 0                ;if OUTB3 <> 0 then data commands
        JE      NO_DATA_COMMANDS
        JMP     DATA_COMMANDS

;!*************************   COMMANDS WITHOUT DATA   ***********************

NO_DATA_COMMANDS:                       ;COMMANDS THAT DO NOT SEND DATA
        CMP     OUTB1, 1                ;COMMAND INVALID
        JNE     DOSTAMPCALSCANCMD
        JMP     INVALID_COMMAND         ;flag invalid command

DOSTAMPCALSCANCMD:                      ;time stamp scan with calibration ?
        CMP     OUTB1, 2
        JNE     VERSIONCMD 
        CALL    SET_TRIGGER             ;send trigger type to M250
        CMP     ERRORCODE, 0            ;check for error
        JNE     SET_CALSTAMP_TRIGGER
        CALL    STAMPSCAN               ;start time stamp scan with calibration
SET_CALSTAMP_TRIGGER:        
        JMP     SEND_ERROR              ;send an error, if any

VERSIONCMD:                             ;version command ?
        CMP     OUTB1, 3
        JNE     CHECKSUMCMD
        JMP     SEND_COMMAND            ;send command word to M250

CHECKSUMCMD:                            ;checksum command ?
        CMP     OUTB1, 4                
        JNE     DIGITALINCMD
        CALL    GETCHKSUM
        JMP     SEND_ERROR              ;send an error, if any

DIGITALINCMD:                           ;digitial input command ?
        CMP     OUTB1, 5
        JNE     DONORMALSCANCMD
        JMP     SEND_COMMAND            ;send command word to M250

DONORMALSCANCMD:                        ;normal scan command ?
        CMP     OUTB1, 6                
        JNE     DOSTAMPSCANCMD
        CALL    SET_TRIGGER             ;send trigger type to M250
        CMP     ERRORCODE, 0            ;check for error
        JNE     SET_NORMAL_TRIGGER
        CALL    NORMALSCAN              ;do normal scan
SET_NORMAL_TRIGGER:        
        JMP     SEND_ERROR              ;send an error, if any

DOSTAMPSCANCMD:                         ;time stamp scan command ?
        CMP     OUTB1, 7                
        JNE     DOFASTSCANCMD
        CALL    SET_TRIGGER             ;send trigger type to M250
        CMP     ERRORCODE, 0            ;check for error
        JNE     SET_STAMP_TRIGGER
        CALL    STAMPSCAN               ;start time stamp scan
SET_STAMP_TRIGGER:        
        JMP     SEND_ERROR              ;send an error, if any

DOFASTSCANCMD:                          ;fast scan command ?
        CMP     OUTB1, 8                
        JNE     NO_DATA9
        CALL    SET_TRIGGER             ;send trigger type to M250
        CMP     ERRORCODE, 0            ;check for error
        JNE     SET_FAST_TRIGGER
        CALL    FASTSCAN                ;do fast single-channel scan
SET_FAST_TRIGGER:        
        JMP     SEND_ERROR              ;send an error, if any

NO_DATA9:                               ;COMMAND INVALID
        CMP     OUTB1, 9
        JNE     PORTTESTCMD 
        JMP     INVALID_COMMAND         ;flag invalid command

PORTTESTCMD:                            ;check port communications command ?
        CMP     OUTB1, 10
        JNE     SCANCALCMD
        CALL    CHECKPORT
        JMP     SEND_RESPONSE           ;send back response and error to host

SCANCALCMD:                             ;pre-scan calibration ?
        CMP     OUTB1, 11
        JNE     CMDS_12_TO_15
        CALL    OFFSET_CORRECTION       ;get offset correction
        JMP     SEND_ERROR              ;send an error, if any

CMDS_12_TO_15:                          ;COMMANDS FROM 12 TO 15 INVALID
        JMP     INVALID_COMMAND         ;flag invalid command

;!***************************   DATA COMMANDS   *****************************

DATA_COMMANDS:        

CHANNELCODECMD:                         ;select input channel (control code) ?
        CMP     OUTB3, 1                
        JNE     DIGITALOUTCMD
        MOV     AL, OUTB2
        MOV     CHANNEL, AL             ;store current channel
        MOV     AL, OUTB1
        MOV     EXTCODE, AL             ;store current external code
        JMP     SEND_COMMAND            ;send command word to M250

DIGITALOUTCMD:                          ;digitial out ? 
        CMP     OUTB3, 2
        JNE     TRACKHOLDCMD
        JMP     SEND_COMMAND            ;send command word to M250

TRACKHOLDCMD:                           ;track/hold ?
        CMP     OUTB3, 3
        JNE     DATA_CMD4
        MOV     AL, OUTB1
        MOV     HOLD, AL                ;store current track/hold status
        JMP     SEND_COMMAND            ;send command word to M250

DATA_CMD4:                              ;COMMAND INVALID
        CMP     OUTB3, 4
        JNE     REMOTEAVGCMD
        JMP     INVALID_COMMAND         ;flag invalid command

REMOTEAVGCMD:                           ;set remote average ?
        CMP     OUTB3, 5
        JNE     CONVERTCMD
        MOV     AL, OUTB1               ;store current average word
        MOV     AVERAGE1, AL
        MOV     AL, OUTB2
        MOV     AVERAGE2, AL
        JMP     SEND_COMMAND            ;send command word to M250

CONVERTCMD:                             ;get conversion ?
        CMP     OUTB3, 6
        JNE     SETDACSCMD
        MOV     AL, OUTB2
        MOV     CHANNEL, AL             ;store current channel
        MOV     AL, OUTB1
        MOV     EXTCODE, AL             ;store current external code
        JMP     SEND_COMMAND            ;send command word to M250

SETDACSCMD:                             ;set dacs ?
        CMP     OUTB3, 7                
        JNE     DATA_CMDS_8_TO_11
        CALL    SETDACS
        JMP     SEND_RESPONSE           ;send back response and error to host

DATA_CMDS_8_TO_11:                      ;COMMANDS FROM  8 TO 11 INVALID
        CMP     OUTB3, 11
        JA      SETCALCMD
        JMP     INVALID_COMMAND         ;flag invalid command

SETCALCMD:                              ;setup long term self-calibration ?
        CMP     OUTB3, 12                
        JNE     STARTCALCMD
        CALL    SETSELFCAL
        JMP     SEND_RESPONSE           ;send back response and error to host

STARTCALCMD:                            ;start long term selt-calibration ?
        CMP     OUTB3, 13                
        JNE     SETSTAMPRATECMD
        CALL    SELFCAL
        JMP     SEND_ERROR              ;send an error, if any

SETSTAMPRATECMD:                        ;set time stamp scan rate ?
        CMP     OUTB3, 14
        JNE     DATA_CMD15
        CALL    SETSTAMPRATE
        JMP     SEND_ERROR              ;send an error, if any

DATA_CMD15:                             ;COMMAND INVALID
        JMP     INVALID_COMMAND         ;flag invalid command

;!************************   SEND COMMAND TO M250   *************************

SEND_COMMAND:
        CALL    SENDIT                  ;send command word to M250
        CMP     ERRORCODE, 0            ;check for error
        JNE     SEND_RESPONSE           ;send back response and error to host
        CMP     OUTB3, 0                
        JNE     READ_CONVERSION         ;if not 0, get conversion command

READ_VERSION:
        CMP     OUTB1, 3                ;get version ?
        JE      ONE_BYTE_RESPONSE

READ_DIGITAL_INPUT:        
        CMP     OUTB1, 5                ;get digital input ?
        JE      ONE_BYTE_RESPONSE
        JMP     SEND_RESPONSE           

READ_CONVERSION:                              
        CMP     OUTB3, 6                ;get conversion command ?
        JNE     SEND_RESPONSE
        CALL    GET_CONV
        CMP     ERRORCODE, 0            ;check for error
        JNE     SEND_RESPONSE           ;if error, send response and error
        JMP     SEND_ERROR              ;send an error, if any

INVALID_COMMAND:                        ;INVALID COMMANDS
        MOV     ERRORCODE, 15h          ;flag an invalid command
        JMP     SEND_ERROR              ;send an error, if any

ONE_BYTE_RESPONSE:
        CALL    CHK1                    ;get response

SEND_RESPONSE:                          ;load response for basic
        MOV     ES, L_RESPONSE_SEG      ;segment and offset to response
        MOV     BX, L_RESPONSE_OFS
        MOV     AX, RESPONSE
        MOV     ES:[BX], AX             ;store response
        XOR     AX, AX
        MOV     ES:[BX + 2], AX
SEND_ERROR:                             ;load ERRORCODE for basic
        XOR     AH, AH                  ;clear AH
        MOV     AL, ERRORCODE
        MOV     ES, ERROR_SEG           ;segment and offset to error
        MOV     BX, ERROR_OFS
        MOV     ES:[BX], AX
EXIT_M250:
        POPF                            ;restore flags
        POP     DI                      ;restore registers
        POP     SI
        POP     ES
        POP     DS
        POP     SS
        POP     CX
        POP     BX
        MOV     SP, BP
        POP     BP                      ;restore old framepointer
        RET     12
M250    ENDP

;!**********************   INITIALIZATION ROUTINES   ************************
;I******   This routine will initalize the port and reset the M250.   *******

INIT    PROC    NEAR

        AND     AX, 0FFFh               ;put port address in AX
        MOV     OUTPORT, AX             ;address for output data
        INC     AX
        MOV     INPORT, AX              ;address for input data
        MOV     ES, L_RESPONSE_SEG      ;segment and offset to response
        MOV     BX, L_RESPONSE_OFS
        MOV     AX, ES:[BX]
        INC     AX
        MOV     RWDELAY, AX             ;read/write delay
        MOV     ERRORCODE, 0            ;clear ERRORCODE
DO_INIT:
        MOV     CX, 64                  ;setup resetloop counter
RESETLOOP:
        PUSH    CX
        CALL    RESET_250               ;reset M250
        MOV     CX, 256                 ;setup readloop counter
READLOOP:        
        CALL    READ_REGS               ;read registers, sum returned in BL
        JZ      REGS_ZERO               ;registers zero?
        LOOP    READLOOP
        
        JMP     BAD_REGS                ;registers not zero, try again
REGS_ZERO:        
        MOV     STATUS, 30h             ;initialize STATUS
        MOV     NEWDATA, 0              ;initialize NEWDATA
        MOV     M250ACK, 10h            ;initialize M250ACK
        MOV     CHECKSUM, 0             ;initialize CHECKSUM
        CALL    GETCHKSUM               ;get checksum
        CMP     ERRORCODE, 0            ;check for error
        JE      GOOD_INIT               ;if no error, then reset successful
BAD_REGS:        
        MOV     STATUS, 30h             ;write STATUS with error bit low
        MOV     AL, STATUS
        CALL    WRITENIBBLE
        POP     CX                      ;restore resetloop counter
        LOOP    RESETLOOP               ;try to reset again
        
        MOV     ERRORCODE, 12h          ;UNABLE TO RESET M250
        JMP     DONE_INIT               ;skip POP CX
GOOD_INIT:
        POP     CX                      ;pop CX off stack
DONE_INIT:
        RET
INIT    ENDP

;I***************************   RESET THE M250   ****************************

RESET_250 PROC NEAR
        
        MOV     AL, 070H                ;clear reg1
        CALL    WRITENIBBLE
        MOV     AL, 0B0h                ;clear reg2
        CALL    WRITENIBBLE
        MOV     AL, 0F0h                ;clear reg3
        CALL    WRITENIBBLE
        MOV     STATUS, 38h             ;write STATUS with error bit high
        MOV     AL, STATUS
        CALL    WRITENIBBLE
        RET
RESET_250 ENDP

;I************************   READ M250 REGISTERS   **************************

READ_REGS  PROC NEAR
        
        XOR     BX, BX
        MOV     AL, 50h                 ;read reg1
        CALL    READNIBBLE
        ADD     BX, AX                  ;add up regs
        MOV     AL, 90h                 ;read reg2
        CALL    READNIBBLE
        ADD     BX, AX                  ;add up regs
        MOV     AL, 0D0h                ;read reg3
        CALL    readnibble
        ADD     BX, AX                  ;add up regs
        MOV     AL, 10h                 ;read reg0
        CALL    READNIBBLE
        ADD     BX, AX                  ;add up regs
        RET
READ_REGS ENDP

;!***************************   MISC ROUTINES   *****************************
;M***  This routine writes 3 nibbles to the M250 and updates the status.  ***
;the three nibbles written are OUTB1, OUTB2, and OUTB3

SENDIT  PROC NEAR
        
        MOV     DX, OUTPORT             ;set port to OUTPORT
        MOV     AL, STATUS
        AND     AL, 0DFH                ;enable reading status reg
        OUT     DX, AL
        CALL    DELAYRW                 ;delay for write strobe
        INC     DX                      ;INPORT
        MOV     ERRORCODE, 0
        MOV     CX, 0FFFFh              ;lots of tries
CHECKACK:
        IN      AL, DX                  ;get M250 status in high nibble
        XCHG    AL, AH
        IN      AL, DX                  ;get M250 status in high nibble again
        AND     AX, 0F0F0H              ;mask off low bits
        CMP     AL, AH                  ;check to see if equal
        JNE     CHECKACK
        AND     AH, 10H                 ;mask ack bit
        CMP     AH, M250ACK             ;has last command been read?
        JNE     PORTREADY
        LOOP    CHECKACK                ;port not ready, try again
        MOV     ERRORCODE, 13h          ;acknowledgement timeout
        TEST    AL, 80H                 ;check for error bit
        JE      SEND_ERROR_HI
        JMP     ERRACK                  ;timeout, don't write
PORTREADY:
        TEST    AL, 80H                 ;check for error bit
        JNE     NOERROR
SEND_ERROR_HI:        
        MOV     ERRORCODE, 14h          ;flag M250 raised its error bit
        JMP     NOREADERROR
NOERROR:                        ;now, host knows M250 is ready to accept data 
                                        ;write 3 nibbles and STATUS
        MOV     AL, OUTB1
        OR      AL, 070H                ;select reg1
        CALL    WRITENIBBLE
        MOV     AL, OUTB2
        OR      AL, 0B0h                ;select reg2
        CALL    WRITENIBBLE
        MOV     AL, OUTB3
        OR      AL, 0F0h                ;select reg3
        CALL    WRITENIBBLE
        
        XOR     M250ACK, 10h            ;toggle ack
        XOR     STATUS, 2               ;toggle NewData
        MOV     AL, STATUS              ;reg0 select lines = 0
        CALL    WRITENIBBLE
ERRACK:
        RET
SENDIT  ENDP

;M** This routine gets a word from the M250 and puts it in CX and RESPONSE **

CHK1    PROC NEAR
        
        MOV     DX, OUTPORT             ;set port to OUTPORT
        MOV     AL, STATUS
        AND     AL, 0DFH                ;enable reading status reg
        OUT     DX, AL
        CALL    DELAYRW                 ;delay for write strobe
        INC     DX                      ;INPORT
        MOV     CX, 0FFFFh              ;lots of tries
CHKSTAT:
        IN      AL, DX                  ;get M250 status in high nibble
        XCHG    AL, AH
        IN      AL, DX                  ;get M250 status in high nibble again
        AND     AX, 0F0F0H              ;mask off low bits
        CMP     AL, AH                  ;check to see if equal
        JNE     CHKSTAT
        AND     AH, 20H                 ;mask for NEWDATA bit
        CMP     AH, NEWDATA             ;data ready ?
        JNE     OUTCHECK                ;read new data
CHECKREAD:
        LOOP    CHKSTAT                 ;no new data so try again
        MOV     ERRORCODE, 11h          ;no new data timeout
        TEST    AL, 80H                 ;check for error bit
        JE      OUTCHK1                 ;get error code
        JMP     NOTPICERR
OUTCHECK:
        ; DX is INPORT at this point   status reg read is selected
        MOV     CX, 1000H
        TEST    AL, 80H                 ;check for error bit
        JNE     NOREADERROR
OUTCHK1:                            
        MOV     ERRORCODE, 10h          ;flag M250 raised its error bit
NOREADERROR:                            ;read 3 nibbles and update STATUS
        AND     AL, 40H
        JNZ     NOT13
        XOR     CX, CX
NOT13:  MOV     AL, 0D0h                ;select reg3
        CALL    READNIBBLE
        SHR     AL, 1                   ;divide by 16
        SHR     AL, 1
        SHR     AL, 1
        SHR     AL, 1
        ADD     CX, AX                  ;save reg3

        MOV     AL, 90h                 ;select reg2
        CALL    READNIBBLE
        ADD     CX, AX                  ;save reg2

        MOV     AL, 50h                 ;select reg1
        CALL    READNIBBLE

        SHL     AX, 1                   ;mult by 16
        SHL     AX, 1
        SHL     AX, 1
        SHL     AX, 1
        ADD     CX, AX                  ;save reg1
        CMP     ERRORCODE, 0            ;check for error
        JNE     NOTPICERR
        XOR     NEWDATA, 20H            ;toggle NEWDATA
        XOR     STATUS, 1               ;toggle HostAck

NOTPICERR:
        MOV     AL, STATUS              ;reg0, select lines both = 0
        CALL    WRITENIBBLE
        MOV     RESPONSE, CX
        RET
CHK1    ENDP

;M** This routine gets the conversion result and sends it back to the host **
GET_CONV PROC NEAR
        MOV     LOW14, 0                ;clear LOW14
        CALL    CHK1                    ;read word
        CMP     ERRORCODE, 0            ;check for error
        JNE     CONV_ERROR              ;send back response and error to host

        CLC
        RCR     CX, 1                   ;carry contains msb of LOW14
        MOV     HIGH12, CX              ;save HIGH12
        JNC     DC1
        MOV     LOW14, 2000h            ;if carry set, set bit in LOW14
DC1:
        CALL    CHK1                    ;read second word
        CMP     ERRORCODE, 0            ;check for error 
        JNE     CONV_ERROR              ;send back response and error to host
        ADD     CX, LOW14               ;LOW14 in CX, add msb
        MOV     LOW14, CX               ;save LOW14
        
        CALL    GETCOUNTS               ;returns low word of count in CX,
                                        ;high word in AX
        MOV     ES, L_RESPONSE_SEG      ;segment and offset to response
        MOV     BX, L_RESPONSE_OFS
        
        SUB     CX, OFFSET_LO           ;subtract offset correction
        SBB     AX, OFFSET_HI
        MOV     ES:[BX], CX             ;store counts
        MOV     ES:[BX + 2], AX
CONV_ERROR:        
        RET
GET_CONV ENDP

;M*******************   NULL LOOP FOR HARDWARE SETTLING  ********************
;TIMING IS COMPUTER EXECUTION SPEED-DEPENDENT
;LOOP EXECUTES RWDELAY + 1 TIMES

DELAYRW PROC    NEAR
        TEST    RWDELAY, 0FFFEh
        JNZ     DELAY1
        RET
DELAY1:
        PUSH    CX
        MOV     CX, WORD PTR RWDELAY
WAIT_HERE:

        LOOP    WAIT_HERE
        POP     CX
        RET
DELAYRW ENDP

;M***   convert HIGH12 and LOW14 to one long integer count for voltage   ****
;expects LOW14 in CX

GETCOUNTS  PROC NEAR

        MOV     BX, HIGH12              ;get index for HIGH12 word
        SHL     BX, 1                   ;make word index
        MOV     SI, OFFSET COR_ARRAY    ;get offset to correction array
        MOV     AX, [SI + BX]           ;index to HIGH12 word
        ADD     CX, AX                  ;add correction to LOW14
        CMP     HOLD, 1                 ;track/hold ?
        JNE     NOT_HOLD
        ADD     BX, 8192                ;index to second dimension
NOT_HOLD:
        MOV     SI, OFFSET ABS_ARRAY    ;get offset to absolute accuracy array
        MOV     AX, [SI + BX]           ;index to HIGH12 word
        ADD     CX, AX                  ;add absolute accuracy to LOW14
        MOV     AX, CX                  ;and put it in AX
        CWD                             ;convert word to dword in AX:DX
        XOR     BX, BX                  ;clear high part of HIGH12
        MOV     CX, HIGH12
        ROL     CX, 1                   ;multiply by 8
        ROL     CX, 1
        ROL     CX, 1
        MOV     BL, CH                  ;multiply by 256 (2048 = 8 * 256)
        MOV     CH, CL
        XOR     CL, CL
        CLC
        ADD     CX, AX                  ;(HIGH12 * 2048) + LOW14
        ADC     BX, DX                  ;low part of count in CX
        MOV     AX, BX                  ;high part of count in AX
        RET
GETCOUNTS ENDP

;M**********************   offset calibration   *************************

OFFSET_CAL PROC NEAR
        
        MOV     OUTB1, AL
        AND     OUTB1, 0Fh              ;first nibble of command
        SHR     AL,1                    ;divide by 16
        SHR     AL,1
        SHR     AL,1
        SHR     AL,1
        AND     AL, 0Fh
        MOV     OUTB2, AL               ;second nibble of command
        
        MOV     ES, L_RESPONSE_SEG      ;segment and offset to response
        MOV     BX, L_RESPONSE_OFS
        MOV     CX, ES:[BX]                ;number of conversions
        MOV     AX, ES:[BX + 2]
        MOV     POWER2, AX              ;number of shifts
        
        MOV     OFFSET_HI, 0            ;clear offset dword
        MOV     OFFSET_LO, 0
OFFSET_LOOP:        
        PUSH    CX                      ;store loop counter
        MOV     OUTB3, 6                ;get conversion command
        CALL    SENDIT                  ;get a conversion on current channel
        CMP     ERRORCODE, 0            ;check for error
        JNE     EXIT_CAL_CX
        
        MOV     LOW14, 0                ;clear LOW14
        CALL    CHK1                    ;read word
        CMP     ERRORCODE, 0            ;check for error
        JNE     EXIT_CAL_CX
        CLC
        RCR     CX, 1                   ;carry contains msb of LOW14
        MOV     HIGH12, CX              ;save HIGH12
        JNC     DC7
        MOV     LOW14, 2000h            ;if carry set, set bit in LOW14
DC7:
        CALL    CHK1                    ;read second word
        CMP     ERRORCODE, 0            ;check for error
        JNE     EXIT_CAL_CX
        ADD     CX, LOW14               ;LOW14 in CX, add msb
        
        CALL    GETCOUNTS               ;returns low word of count in CX,
                                        ;high word in AX
        ADD     OFFSET_LO, CX           ;accumulate conversions
        ADC     OFFSET_HI, AX
        POP     CX                      ;pop loop counter
        LOOP    OFFSET_LOOP
        
        MOV     AX, OFFSET_LO           ;setup to divide offset
        MOV     BX, OFFSET_HI
        
        MOV     CX, POWER2              ;power of 2
        JCXZ    NO_OFFSET_POWER
OFFSETDIVIDE:                           ;shift 1 to 9 powers of 2
        CLC
        SHR     BX, 1
        RCR     AX, 1
        LOOP    OFFSETDIVIDE
NO_OFFSET_POWER:
        
        MOV     OFFSET_LO, AX           ;store new offset count
        MOV     OFFSET_HI, BX
        JMP     EXIT_CAL
EXIT_CAL_CX:        
        POP     CX                      ;if error while in loop, pop CX
EXIT_CAL:
        RET
OFFSET_CAL ENDP

;M*****************  calculate full scale calibration   ********************

FULLSCALE_CAL PROC NEAR
        
        MOV     OUTB1, AL
        AND     OUTB1, 0Fh              ;first nibble of command
        SHR     AL,1                    ;divide by 16
        SHR     AL,1
        SHR     AL,1
        SHR     AL,1
        AND     AL, 0Fh
        MOV     OUTB2, AL               ;second nibble of command
        
        MOV     ES, L_RESPONSE_SEG      ;segment and offset to response
        MOV     BX, L_RESPONSE_OFS
        MOV     CX, ES:[BX]                ;number of conversions
        MOV     AX, ES:[BX + 2]
        MOV     POWER2, AX              ;number of shifts
        
        MOV     FULLSCALE_HI, 0         ;clear full scale dword
        MOV     FULLSCALE_LO, 0
FULLSCALE_LOOP:        
        PUSH    CX                      ;push loop counter
        MOV     OUTB3, 6                ;get a conversion on current channel
        CALL    SENDIT
        CMP     ERRORCODE, 0            ;check for error
        JE      NO_FULL_EXIT
        JMP     EXIT_FULL_CX
NO_FULL_EXIT:        
        MOV     LOW14, 0
        CALL    CHK1                    ;read word
        CMP     ERRORCODE, 0            ;check for error
        JE      NO_FULL_EXIT1
        JMP     EXIT_FULL_CX
NO_FULL_EXIT1:        
        CLC
        RCR     CX, 1                   ;carry contains msb of LOW14
        MOV     HIGH12, CX              ;save HIGH12
        JNC     DC10
        MOV     LOW14, 2000h            ;if carry set, set bit in LOW14
DC10:
        CALL    CHK1                    ;read second word
        CMP     ERRORCODE, 0            ;check for error 
        JNE     EXIT_FULL_CX
        ADD     CX, LOW14               ;LOW14 in CX, add msb
        
        CALL    GETCOUNTS               ;returns low word of count in CX,
                                        ;high word in AX
        ADD     FULLSCALE_LO, CX        ;accumulate conversions
        ADC     FULLSCALE_HI, AX
        POP     CX                      ;pop loop counter
        LOOP    FULLSCALE_LOOP

        MOV     AX, FULLSCALE_LO        ;setup to divide fullscale
        MOV     DX, FULLSCALE_HI
        
        MOV     CX, POWER2              ;power of 2
        JCXZ    NO_FULL_POWER
FULLDIVIDE:                             ;shift 1 to 9 powers of 2
        CLC
        SHR     DX, 1
        RCR     AX, 1
        LOOP    FULLDIVIDE
NO_FULL_POWER:
        
        SUB     AX, OFFSET_LO           ;subtract offset correction
        SBB     DX, OFFSET_HI
        
        MOV     ES, L_RESPONSE_SEG      ;segment and offset to response
        MOV     BX, L_RESPONSE_OFS
        
        MOV     ES:[BX], AX             ;store dword full scale factor
        MOV     ES:[BX + 2], DX
        JMP     EXIT_FULL
EXIT_FULL_CX:        
        POP     CX                      ;if error while in loop, pop CX
EXIT_FULL:
        RET
FULLSCALE_CAL ENDP

;M*************************************************************************** 
; This routine tests the communications of the port for errors.  It will
; repeat the following loop 8192 times (set in CX).
;
; Request a conversion on channel 6.
; Request a conversion on channel 7.
; Request a conversion on channel 6.
; Request a digital input.
; Get a checksum.
;
; If a failure occurs, the read/write delay is incremented, the M250 reset,
; and the procedure will start again.  Procedure fails if the read/write delay
; exceeds 25.
;
CHECKPORT  PROC NEAR
        
        MOV     RWDELAY, 0              ;clear RWDELAY
        MOV     CX, 8192                ;start loop counter
CHECKING_PORT:
        PUSH    CX                      ;push loop counter
        
        CALL    AVG1                    ;.v5
        CMP     ERRORCODE, 0            ;.v5 check for error
        JNE     CHECK_FAILED            ;.v5
        
        MOV     OUTB2, 7                ;get a conversion on channel 7
        CALL    DUMMYCONV               ;simulate get conversion
        CMP     ERRORCODE, 0            ;check for error
        JNE     CHECK_FAILED

        MOV     OUTB2, 6                ;get a conversion on channel 6
        CALL    DUMMYCONV               ;simulate get conversion
        CMP     ERRORCODE, 0            ;check for error
        JNE     CHECK_FAILED

        MOV     OUTB2, 7                ;get a conversion on channel 7
        CALL    DUMMYCONV               ;simulate get conversion
        CMP     ERRORCODE, 0            ;check for error
        JNE     CHECK_FAILED

        CALL    DIG1
        CMP     ERRORCODE, 0            ;check for error
        JNE     CHECK_FAILED
        
        CALL    CHK1                    ;read digitial input word
        CMP     ERRORCODE, 0            ;check for error
        JNE     CHECK_FAILED

        CALL    GETCHKSUM               ;get checksum
        CMP     ERRORCODE, 0            ;check for error
        JNE     CHECK_FAILED
        
        POP     CX
        MOV     ERRORCODE, 0            ;clear ERRORCODE
        
        MOV     AX, 100h
;;if there was a keystroke, exit scan
        INT     16H
        JZ      CHECK_OUTKEY           ;key was not pressed
        JMP     OUT_OF_CHECK            ;key pressed, exit loop
CHECK_OUTKEY:
        LOOP    CHECKING_PORT
        
        JMP     OUT_OF_CHECK
CHECK_FAILED:
        POP     CX                      ;if failure while in loop, pop CX
        CALL    DO_INIT                 ;if check fails, re-initialize
        CMP     ERRORCODE, 0            ;.v5
        JNE     OUT_OF_CHECK            ;.v5
        INC     RWDELAY                 ;increment read/write delay
        MOV     CX, 8193                ;start loop counter
        CMP     RWDELAY, 25             ;read/write delay > 25 ?
        JL      CHECK_OUTKEY
        MOV     RWDELAY, 100            ;portcheck error
OUT_OF_CHECK:
        CALL    DO_INITIALIZATION
        MOV     AX, RWDELAY             ;put read/write delay in RESPONSE
        MOV     RESPONSE, AX
        RET
CHECKPORT  ENDP
        
;.v5 set averaging to 1
AVG1 PROC NEAR        
        MOV     OUTB3, 5
        MOV     OUTB1, 0               ;set averaging to 1
        MOV     OUTB2, 0
        CALL    SENDIT        
        RET
AVG1 ENDP

DIG1 PROC NEAR        
        MOV     OUTB3, 0                ;request a digital input
        MOV     OUTB2, 0
        MOV     OUTB1, 5
        CALL    SENDIT
        RET
DIG1 ENDP

;M*****   This routine simulates getting a conversion from the M250.   ******

DUMMYCONV  PROC NEAR              ; the answer is discarded
        
        MOV     OUTB3, 6                ;do conversion
        MOV     OUTB1, 0
        
        CALL    SENDIT
        CMP     ERRORCODE, 0
        JNE     EXIT_DUMMY
        
        CALL    CHK1
        CMP     ERRORCODE, 0
        JNE     EXIT_DUMMY
        
        CALL    CHK1
EXIT_DUMMY:        
        RET
DUMMYCONV  ENDP

;M*************   This routine writes one nibble to M250   **************
; AL = data for OUTPORT    (A1 A0 1 1 x x x x)      x x x x = data nibble

WRITENIBBLE  PROC NEAR
        
        MOV     DX, OUTPORT             ;select port
        OUT     DX, AL                  ;data nibble and select reg
        AND     AL, 0EFh                ;prepare write strobe active
        CALL    DELAYRW                 ;delay for write strobe
        OUT     DX, AL                  ;strobe active
        CALL    DELAYRW                 ;delay for write strobe
        OR      AL, 10h                 ;strobe inactive
        OUT     DX, AL                  ;re-select reg
        RET
WRITENIBBLE  ENDP

;M*******   This routine reads one nibble from the a M250 register.   *********
; on entry    AL = data for OUTPORT    (A1 A0 0 1 x x x x)
; on exit     high nibble of AL is data

READNIBBLE  PROC NEAR
        
        MOV     DX, OUTPORT             ;select port
        OUT     DX, AL                  ;select reg for reading
        INC     DX                      ;INPORT
        CALL    DELAYRW                 ;delay for addr
        XOR     AH, AH                  ;clear AH
        IN      AL, DX                  ;get reg
        XOR     AL, 80H                 ;correct data
        AND     AL, 0F0H                ;mask data
        ADD     CHECKSUM, AX            ;compute checksum
        RET
READNIBBLE  ENDP

;M**************************    SET DAC FOR SELF_CAL   ********************

SETDACS     PROC NEAR
        
        MOV     ES, L_RESPONSE_SEG      ;segment and offset to response
        MOV     BX, L_RESPONSE_OFS
        MOV     AX, ES:[BX]             ;setdacs data
        MOV     OUTB3, 7                ;set low dac
        MOV     OUTB2, AL
        AND     OUTB2, 0Fh              ;second nibble of command
        SHR     AL,1                    ;divide by 16
        SHR     AL,1
        SHR     AL,1
        SHR     AL,1
        AND     AL, 0Fh
        MOV     OUTB1, AH               ;first nibble of command
        CALL    SENDIT
        CMP     ERRORCODE, 0
        JNE     SETDAC_ERROR
        MOV     OUTB3, 8                ;set high dac
        MOV     OUTB2, AH
        AND     OUTB2, 0Fh              ;second nibble of command
        SHR     AH,1                    ;divide by 16
        SHR     AH,1
        SHR     AH,1
        SHR     AH,1
        AND     AH, 03h
        MOV     OUTB1, AH               ;first nibble of command
        CALL    SENDIT
SETDAC_ERROR:        
        RET
SETDACS ENDP

;M**********************   GET AND COMPARE CHECKSUM   ***********************

GETCHKSUM PROC NEAR
        
        MOV     OUTB1, 4                ;get checksum command
        MOV     OUTB2, 0
        MOV     OUTB3, 0                ;get checksum
        MOV     AX, CHECKSUM            ;format current checksum
        SHR     AX, 1                   ;divide by 16
        SHR     AX, 1
        SHR     AX, 1
        SHR     AX, 1
        AND     AX, 0FFh
        PUSH    AX                      ;store current checksum

        CALL    SENDIT
        CMP     ERRORCODE, 0            ;check for error
        JNE     CSUM_SEND_ERROR
        CALL    DELAYRW                 ;delay between writing command and
                                        ;reading word
        MOV     CHECKSUM, 0
        CALL    CHK1                    ;read M250 checksum, passed back in CX
                                        
        POP     AX                      ;pop current checksum 
        AND     CX, 0FFh                ;format M250 checksum
        CMP     AX, CX                  ;compare to see if equal
        JE      CSUM_ERROR              ;not a checksum error
        CMP     ERRORCODE, 0            ;check for error
        JNE     CSUM_ERROR
        MOV     ERRORCODE, 17h          ;set checksums did not match error
        JMP     CSUM_ERROR
CSUM_SEND_ERROR:
        POP     AX                      ;if error while sending, pop AX
CSUM_ERROR:
        RET
GETCHKSUM ENDP

;!*************************   SCANNING ROUTINES   **************************
;S***********   WAIT FOR M250 TO SEND WORD,  HOST KEYPRESS CANCELS  ********
;      this routine parallels CHK1 except there is no timeout

CHK2    PROC NEAR
        
        MOV     DX, OUTPORT             ;set port to OUTPORT
        MOV     AL, STATUS
        AND     AL, 0DFH                ;enable reading status reg
        OUT     DX, AL
        CALL    DELAYRW                 ;delay for write strobe
        INC     DX                      ;INPORT
CHKSTAT2:
        IN      AL, DX                  ;get M250 status in high nibble
        XCHG    AL, AH
        IN      AL, DX                  ;get M250 status in high nibble again
        AND     AX, 0F0F0H              ;mask off low bits
        CMP     AL, AH                  ;check to see if equal
        JNE     CHKSTAT2
        AND     AH, 20H                 ;mask for NEWDATA bit
        CMP     AH, NEWDATA             ;data ready ?
        JNE     DONECHECK               ;read new data
        
        TEST    AL, 80H                 ;check for error bit
        JE      DONECHECK               ;get error code
        
        MOV     AX, 100h                ;check for keystroke
        INT     16H
        JZ      CHKSTAT2                ;key was not pressed
        CMP     AL, 27                  ;esc key ?
        JE      ABORT_IT
        XOR     AX, AX                  ;get key
        INT     16H
        JMP     CHKSTAT2                ;not esc key
ABORT_IT:        
        MOV     ERRORCODE, 19h          ;set for ESC key to abort time stamp
                                        ;scans
DONECHECK:
        RET
CHK2    ENDP

;S********************   SET TRIGGER TYPE FOR A SCAN   **********************

SET_TRIGGER PROC NEAR
        
        XOR     AX, AX
        MOV     AL, OUTB1
        PUSH    AX                      ;store scan type
        CMP     TRIGGERTYPE, 0          ;default no trigger
        JE      NO_TRIGGER
        MOV     OUTB1, 9                ;set for positive trigger
        CMP     TRIGGERTYPE, 1          ;positive trigger ?
        JE      SEND_TRIGGERTYPE
        MOV     OUTB1, 10               ;set for negative trigger
        CMP     TRIGGERTYPE, 2          ;negitive trigger ?
        JE      SEND_TRIGGERTYPE
        MOV     TRIGGERTYPE, 0          ;invalid type, no trigger selected
        JMP     NO_TRIGGER              ;exit
SEND_TRIGGERTYPE:
        MOV     OUTB2, 0
        MOV     OUTB3, 0
        CALL    SENDIT                  ;send trigger type
NO_TRIGGER:        
        POP     AX
        MOV     OUTB1, AL               ;restore scan type
        RET
SET_TRIGGER ENDP

;S************************   SETUP SCAN CHANNELS   **************************
;OUTB1 = position index   0   scan type             
;                         1-6 scan channel  
;                         7   number of datapoints

SETSCANCHANNELS     PROC NEAR               
        
        MOV     ES, L_RESPONSE_SEG      ;segment and offset to response
        MOV     BX, L_RESPONSE_OFS      
        CMP     OUTB1, 0                ;scan type ?
        JNE     CHAN0
        MOV     FAST_SCAN_FLAG, 1       ;set fast scan flag
        MOV     AX, ES:[BX]
        CMP     AX, 0                   ;AX <> 0 then = fast scan channels
        JNZ     FAST_SCAN
        MOV     FAST_SCAN_FLAG, 0       ;flag normal / time stamp scan
FAST_SCAN:
        MOV     AH, 4                   ;set scan mode command 
        MOV     SCANSET, AX             ;set scan mode command data
        JMP     EXIT_SET
CHAN0:                                  ;channel 0 ?
        CMP     OUTB1, 1                
        JNE     CHAN1

        CALL    SETCHANS                ;setup channel 0 control codes,
                                        ;passed back in AX
        MOV     CHANNEL0, AX            ;store channel 0 data
        JMP     EXIT_SET
CHAN1:                                  ;channel 1 ?
        CMP     OUTB1, 2
        JNE     CHAN2

        CALL    SETCHANS                ;setup channel 1 control codes,
                                        ;passed back in AX
        MOV     CHANNEL1, AX            ;store channel 1 data
        JMP     EXIT_SET
CHAN2:                                  ;channel 2 ?
        CMP     OUTB1, 3
        JNE     CHAN3

        CALL    SETCHANS                ;setup channel 2 control codes,
                                        ;passed back in AX
        MOV     CHANNEL2, AX            ;store channel 2 data
        JMP     EXIT_SET
CHAN3:                                  ;channel 3 ?
        CMP     OUTB1, 4
        JNE     CHAN4

        CALL    SETCHANS                ;setup channel 3 control codes,
                                        ;passed back in AX
        MOV     CHANNEL3, AX            ;store channel 3 data
        JMP     EXIT_SET
CHAN4:                                  ;channel 4 ?
        CMP     OUTB1, 5
        JNE     CHAN5

        CALL    SETCHANS                ;setup channel 4 control codes,
                                        ;passed back in AX
        MOV     CHANNEL4, AX            ;store channel 4 data
        JMP     EXIT_SET
CHAN5:                                  ;channel 5 ?
        CMP     OUTB1, 6
        JNE     POSITION7

        CALL    SETCHANS                ;setup channel 5 control codes,
                                        ;passed back in AX
        MOV     CHANNEL5, AX            ;store channel 5 data
        JMP     EXIT_SET
POSITION7:
        CMP     OUTB1, 7
        JNE     EXIT_SET
        MOV     AX, ES:[BX]
        MOV     NUM_DATAPOINTS, AL      ;number of datapoints
EXIT_SET:
        RET
SETSCANCHANNELS ENDP

;S*****************   sets up the channel control codes   *******************

SETCHANS PROC NEAR
        
        MOV     AX, ES:[BX]             ;get control codes
        CMP     FAST_SCAN_FLAG, 1       ;if = 1 then fast scan
        JE      FAST_CHANS
        MOV     AH, AL                  ;parse out control codes
        SHR     AL,1                    ;divide by 16
        SHR     AL,1
        SHR     AL,1
        SHR     AL,1
        AND     AL, 0Fh                 ;AL = STARTCC
        AND     AH, 0Fh                 ;AH = ENDCC
        JMP     NOTCHAN6_LO

FAST_CHANS:                             ;the code below is for fast multi-
                                        ;channel scans, currently not in use
        MOV     CX, AX
        AND     AH, 7                   ;clear bit 3
        TEST    AH, 1
        JZ      NO2BIT_HI
        OR      AH, 8                   ;set bit 3
NO2BIT_HI:
        AND     CH, 7
        CMP     CH, 6
        JNE     NOTCHAN6_HI
        OR      AH, 8                   ;set bit 3
NOTCHAN6_HI:
        AND     AL, 7                   ;clear bit 3
        TEST    AL, 1
        JZ      NO2BIT_LO
        OR      AL, 8                   ;set bit 3
NO2BIT_LO:
        AND     CL, 7
        CMP     CL, 6
        JNE     NOTCHAN6_LO
        OR      AL, 8                   ;set bit 3
NOTCHAN6_LO:
        RET
SETCHANS ENDP

;S**************   SETUP SCAN RATE AND INTER-CHANNEL DELAY   ****************

SETSCANRATE   PROC NEAR
        
        MOV     INTERCHAN, AL           ;inter channel delay
        MOV     ES, L_RESPONSE_SEG      ;segment and offset to response
        MOV     BX, L_RESPONSE_OFS
        MOV     AX, ES:[BX]
        MOV     SCANDELAYS, AX          ;scan delay data
        MOV     AX, ES:[BX + 2]         ;high word for normal scanning 
        MOV     SCANHI, AX              ;inter-scan delay
        RET
SETSCANRATE ENDP

;S*********************   SETUP TIME STAMP SCAN RATE   **********************

SETSTAMPRATE PROC NEAR
        
        MOV     AL, OUTB2               ;get the first byte
        SHL     AL, 1
        SHL     AL, 1
        SHL     AL, 1
        SHL     AL, 1                   ;multiply by 16
        ADD     AL, OUTB1
        PUSH    AX                      ;store first byte on stack
        MOV     ES, L_RESPONSE_SEG      ;segment and offset to response
        MOV     BX, L_RESPONSE_OFS
        MOV     OUTB1, 0
        MOV     OUTB2, 0
        MOV     OUTB3, 10               ;send set time stamp scan rate command
        CALL    SENDIT
        POP     AX                      ;restore first byte from stack
        INC     AX
        CALL    SEND_AL                 ;send first byte
        MOV     AX, ES:[BX]             
        INC     AX
        CALL    SEND_AL                 ;send second byte
        MOV     AX, ES:[BX]             
        MOV     AL, AH
        INC     AX
        CALL    SEND_AL                 ;send third byte
        RET
SETSTAMPRATE ENDP

;S*******************   SEND SCAN CONFIGUTATION BLOCK   *********************

SENDMODE        PROC NEAR
        
        MOV     TRIGGERTYPE, AL         ;AL = TRIGGERTYPE
        MOV     ES, L_RESPONSE_SEG      ;segment and offset to response
        MOV     BX, L_RESPONSE_OFS
        MOV     AX, ES:[BX]
        MOV     TOT_SCANS, AX           ;total number of scans
        MOV     AX, SCANSET
        MOV     OUTB2, AL               ;number of channels for multi-channel
                                        ;fast scanning (currently not in use)
        MOV     OUTB3, 4
        CALL    SENDIT                  ;send scan mode command
        MOV     AX, CHANNEL5               
        MOV     OUTB2, AL
        MOV     OUTB3, AH
        CALL    SENDIT                  ;send CHANNEL5 control codes
        MOV     AX, CHANNEL4               
        MOV     OUTB2, AL
        MOV     OUTB3, AH
        CALL    SENDIT                  ;send CHANNEL4 control codes
        MOV     AX, CHANNEL3               
        MOV     OUTB2, AL
        MOV     OUTB3, AH
        CALL    SENDIT                  ;send CHANNEL3 control codes
        MOV     AX, CHANNEL2               
        MOV     OUTB2, AL
        MOV     OUTB3, AH
        CALL    SENDIT                  ;send CHANNEL2 control codes
        MOV     AX, CHANNEL1               
        MOV     OUTB2, AL
        MOV     OUTB3, AH
        CALL    SENDIT                  ;send CHANNEL1 control codes
        MOV     AX, CHANNEL0               
        MOV     OUTB2, AL
        MOV     OUTB3, AH
        CALL    SENDIT                  ;send CHANNEL0 control codes
        MOV     AL, INTERCHAN           
        INC     AX
        CALL    SEND_AL                 ;interchannel delay
        MOV     AX, SCANDELAYS          
        MOV     AL, AH
        INC     AX
        CALL    SEND_AL                 ;high scan delay
        MOV     AX, SCANDELAYS          
        INC     AX
        CALL    SEND_AL                 ;low scan delay
        MOV     AX, SCANHI
        INC     AX
        CALL    SEND_AL                 ;hi inter-scan delay
        RET
SENDMODE ENDP

;S***************************  SEND AL TO M250   ****************************

SEND_AL PROC NEAR
        
        MOV     OUTB2, AL
        AND     OUTB2, 0Fh              ;first nibble of AL
        SHR     AL,1                    ;divide by 16
        SHR     AL,1
        SHR     AL,1
        SHR     AL,1
        AND     AL, 0Fh
        MOV     OUTB3, AL               ;second nibble of AL
        CALL    SENDIT                  ;send AL to M250
        RET
SEND_AL ENDP

;S************   This routine performs buffered fast scanning   *************

FASTSCAN  PROC  NEAR

        MOV     ES, L_RESPONSE_SEG      ;segment and offset to response
        MOV     BX, L_RESPONSE_OFS
        MOV     AX, ES:[BX]
        MOV     START_SEG, AX           ;get start segment byte 
        MOV     ES, AX                  ;set seg to save data
        MOV     DI, 0                   ;start at 0 offset

        MOV     NUMSCANS, 0             ;no scans done
        MOV     STOP_FLAG, 0            ;clear STOP_FLAG

        OR      STATUS, 4               ;"run" bit in status nibble
        CALL    SENDIT                  ;send fastscan command
        CMP     ERRORCODE, 0            ;check for error
        JE      CHECK_FTRIGGER
        JMP     FASTSCANERR             ;pic didn't read

CHECK_FTRIGGER:
        CMP     TRIGGERTYPE, 0          ;check for trigger scan start
        JE      START_FAST
        CALL    CHK2                    ;wait for trigger or ESC to cancel
        CMP     ERRORCODE, 0            ;check for error
        JE      START_FAST
        JMP     FASTSCANERR             ;exit scan with error

START_FAST:        
        MOV     AX, TOT_SCANS           ;check to see if last scan
        CMP     AX, NUMSCANS 
        JNE     NEXTFASTSCAN            ;not last scan, don't set STOP_FLAG
F_ENDSCAN:        
        AND     STATUS, 3Bh             ;clear bit #2, "stop scan" bit
        MOV     STOP_FLAG, 1            ;flag send end if scan
NEXTFASTSCAN:
        CALL    CHK1                    ;get one 12 bit word from the M250
        CMP     ERRORCODE, 0            ;check for error
        JNE     FASTEXIT                ;exit scan
        CMP     STOP_FLAG, 1            ;if end of scan, don't add to buffer
        JE      DC2
        
        MOV     LOW14, 0                ;clear LOW14
        CLC
        RCR     CX, 1                   ;carry contains msb of LOW14
        MOV     HIGH12, CX              ;save HIGH12
        JNC     DC2
        MOV     LOW14, 2000h            ;if carry, set msb of LOW14
DC2:
        CALL    CHK1                    ;get one 12 bit word from the M250
        CMP     ERRORCODE, 0            ;check for error
        JNE     FASTSCANERR             ;data was not ready
        CMP     STOP_FLAG, 1            ;if end of scan, don't add to buffer
        JE      NOT_STOP_TIME
        
        ADD     CX, LOW14               ;LOW14 in CX, add msb
        
        CALL    GETCOUNTS               ;returns low word of count in CX,
                                        ;high word in AX
        SUB     CX, OFFSET_LO           ;subtract offset correction
        SBB     AX, OFFSET_HI
        
        MOV     ES:[DI], CX             ;put dword in scan buffer array
        ADD     DI, 2                   
        MOV     ES:[DI], AX             
        ADD     DI, 2                   ;increment pointer to next location

        JNZ     SAME_SEG                ;data is in the same segment
        MOV     AX, ES                  ;set ES to next segment
        ADD     AX, 1000H
        MOV     ES, AX
SAME_SEG:
        INC     NUMSCANS                ;did a scan
NOT_STOP_TIME:        
        MOV     AX, 100H                ;read keyboard buffer
;if there was a keystroke, exit scan
        INT     16H
        JZ      F_OUTKEY                ;no key, get next scan
        JMP     F_ENDSCAN
F_OUTKEY:
        JMP     START_FAST              ;get next scan

FASTEXIT:                     
        CMP     STOP_FLAG, 1            ;check for end of scan
        JNE     FASTSCANERR
        CMP     ERRORCODE, 11h          ;check for timeout error
        JNE     FASTSCANERR             ;if timeout error, normal exit
        MOV     ERRORCODE, 0            ;clear ERRORCODE

FASTSCANERR:        
        JMP     EXIT_SCANS              ;send information back to host
FASTSCAN  ENDP

;S***************   This routine performes normal scanning   ****************

NORMALSCAN PROC    NEAR

        MOV     ES, L_RESPONSE_SEG      ;segment and offset to response
        MOV     BX, L_RESPONSE_OFS
        MOV     AX, ES:[BX]
        MOV     START_SEG, AX           ;get start segment byte
        MOV     ES, AX                  ;set seg to save data
        MOV     DI, 0                   ;start at 0 offset

        MOV     NUMSCANS, 0             ;no scans done
        MOV     STOP_FLAG, 0            ;clear STOP_FLAG
         
        OR      STATUS, 4               ;"run" bit in staus nibble
        CALL    SENDIT                  ;send normal scan command
        CMP     ERRORCODE, 0            ;check for error
        JE      CHECK_NTRIGGER
        JMP     EXIT_SCANS
CHECK_NTRIGGER:
        CMP     TRIGGERTYPE, 0          ;check for trigger scan start
        JE      STARTSCAN
        CALL    CHK2                    ;wait for trigger or ESC to cancel
        CMP     ERRORCODE, 0            ;check for error
        JE      STARTSCAN
        JMP     EXIT_SCANS

STARTSCAN:
        MOV     OFFSET_FLAG, 0
        MOV     ZCOR_LO, 0              ;clear zero correction dword
        MOV     ZCOR_HI, 0
        XOR     CX, CX                  ;clear CX
        MOV     CL, NUM_DATAPOINTS      ;setup loop counter for scan
        MOV     AX, TOT_SCANS           ;check to see if last scan
        CMP     AX, NUMSCANS
        JNE     DATAPOINTS              ;if on last scan, then one more scan
N_ENDSCAN:        
        AND     STATUS, 3Bh             ;clear bit #2, "stop scan" bit
        MOV     STOP_FLAG, 1            ;flag send end if scan
DATAPOINTS:
        PUSH    CX                      ;store loop counter on stack
        
        JMP     GET_OFFSET              ;get normal scan offset correction
RET_GET_OFFSET:
        
        CALL    CHK1                    ;get one 12 bit word from the M250
        CMP     ERRORCODE, 0            ;check for error
        JNE     NORMALEXIT              ;exit scan
        
        CMP     STOP_FLAG, 1            ;if end of scan, don't add tu buffer
        JE      DC3
        
        MOV     LOW14, 0                ;clear LOW14
        CLC
        RCR     CX, 1                   ;carry contains msb of LOW14
        
        MOV     HIGH12, CX              ;save HIGH12
        
        JNC     DC3
        MOV     LOW14, 2000h            ;if carry, set msb of LOW14
DC3:
        CALL    LOOPCHK1                ;call CHK1 until data ready or timeout
        CMP     ERRORCODE, 0            ;check for error
        JNE     N_SCANERR               ;exit scan with error
        
        CMP     STOP_FLAG, 1            ;if end of scan, don't add tu buffer
        JE      SAME_SCAN_SEG
        
        ADD     CX, LOW14               ;LOW14 in CX, add msb
        
        CALL    GETCOUNTS               ;returns low word of count in CX,
                                        ;high word in AX
        SUB     CX, ZCOR_LO             ;subtract offset + normal scan
        SBB     AX, ZCOR_HI             ;interaction correction
        
        MOV     ES:[DI], CX             ;put dword in scan buffer array
        ADD     DI, 2                   
        MOV     ES:[DI], AX             
        ADD     DI, 2                   ;increment pointer to next location
        JNZ     SAME_SCAN_SEG           ;data is in the same segment
        MOV     BX, ES                  ;set ES to next segment
        ADD     BX, 1000H
        MOV     ES, BX
SAME_SCAN_SEG:
        POP     CX                      ;pop loop counter off of stack
        LOOP    DATAPOINTS
        CMP     STOP_FLAG, 1            ;end of scan?
        JE      N_SKIP_NUMSCANS         ;skip inc numscans
        INC     NUMSCANS                ;did one scan
N_SKIP_NUMSCANS:        
        MOV     AX, 100h
;;if there was a keystroke, exit scan
        INT     16H
        JZ      N_OUTKEY                ;no key, get next scan
        JMP     N_ENDSCAN
N_OUTKEY:
        JMP     STARTSCAN               ;start next scan
NORMALEXIT:
        CMP     STOP_FLAG, 1            ;end of scan ?
        JNE     N_SCANERR
        MOV     ERRORCODE, 0            ;clear ERRORCODE
N_SCANERR:      
        POP     CX                      ;pop stored loop counter if error

EXIT_SCANS:                             ;send information back to host        
        MOV     AX, NUMSCANS
        MOV     ES, L_RESPONSE_SEG      ;segment and offset to response
        MOV     BX, L_RESPONSE_OFS
        MOV     ES:[BX], AX             ;number of scans completed
        XOR     AX, AX
        MOV     ES:[BX+2], AX
T_EXIT_SCANS:
        CMP     ERRORCODE, 0            ;check for error
        JE      SCAN_EXIT               ;no error, return
        MOV     AX, RESPONSE            ;there was an error, pass back 
        MOV     ES:[BX], AX             ;M250 error code
SCAN_EXIT:
        RET
NORMALSCAN ENDP

;get offset channel in normal scanning
;this is a jump procedure rather than a called procedure so that when an error
;occurs, it can jump directly to the correct error label

GET_OFFSET:
        CALL    CHK1                    ;get one 12 bit word from the m250
        CMP     ERRORCODE, 0            ;check for error
        JNE     NORMALEXIT              ;exit scan
        
        CMP     STOP_FLAG, 1            ;if end of scan, don't add to buffer
        JE      DC8
        
        MOV     LOW14, 0                ;clear LOW14
        CLC
        RCR     CX, 1                   ;carry contains msb of LOW14
        
        MOV     HIGH12, CX              ;save HIGH12
        
        JNC     DC8
        MOV     LOW14, 2000h            ;set msb of LOW14
DC8:
        CALL    LOOPCHK1                ;call CHK1 until data ready or timeout
        CMP     ERRORCODE, 0            ;check for error
        JNE     N_SCANERR               ;exit scan with error
        
        CMP     STOP_FLAG, 1            ;if end of scan, don't add to buffer
        JE      N_SKIP_LO
        
        ADD     CX, LOW14               ;LOW14 in CX, add msb
        
        CALL    GETCOUNTS               ;returns low word of count in CX,
                                        ;high word in AX
        SUB     CX, OFS_LO              ;load offset correction dword
        SBB     AX, OFS_HI                      
        CLC                             ;divide difference by 4
        SAR     AX, 1
        RCR     CX, 1
        CLC
        SAR     AX, 1   
        RCR     CX, 1                   
        MOV     ZCOR_LO, CX             ;store zero correction
        MOV     ZCOR_HI, AX
        MOV     AX, OFS_LO
        MOV     BX, OFS_HI
        ADD     ZCOR_LO, AX             ;save offset + zero correction
        ADC     ZCOR_HI, BX
N_SKIP_LO:        
        JMP     RET_GET_OFFSET

;S*************  This routine calls CHK1 up to 100 times    ***************

LOOPCHK1  PROC NEAR
        MOV     CX, 100                 ;# of chk1's to call before timeout
TIMEOUT_CHK:
        PUSH    CX                      ;store loop counter on stack
        MOV     ERRORCODE, 0            ;clear ERRORCODE
        CALL    CHK1                    ;get one 12 bit word from the M250
        CMP     ERRORCODE, 11h          ;timeout error ?
        JNE     NOT_TIMEOUT             ;not timeout, exit loop
        POP     CX                      ;pop loop counter off stack
        LOOP    TIMEOUT_CHK
        JMP     SKIP_TIME_CX            ;skip pop
NOT_TIMEOUT:                            ;out of loop        
        POP     BX                      ;pop temp loop off stack into BX
SKIP_TIME_CX:        
        RET
LOOPCHK1  ENDP

;S*************   This routine starts time-stamped scanning    **************

STAMPSCAN PROC    NEAR
        
        MOV     ES, L_RESPONSE_SEG      ;segment and offset to response
        MOV     BX, L_RESPONSE_OFS
        MOV     AX, ES:[BX]
        MOV     START_SEG, AX           ;get start segment byte
        
        MOV     NUMSCANS, 0             ;no scans done
        MOV     STOP_FLAG, 0            ;clear STOP_FLAG
        MOV     NSCANLO, 0
        MOV     NSCANHI, 0

        MOV     STORE_1ST_WORD, 0       ;initialize storage vars
        MOV     STORE_2ND_WORD, 0
        MOV     S_1ST_WORD, 0
        MOV     S_2ND_WORD, 0
        MOV     STAMP_CHKSUM, 0        
        
        XOR     CX, CX
        MOV     CL, NUM_DATAPOINTS
        INC     CX
        SHL     CX, 1 
        SHL     CX, 1                   ;NUMBER OF WORDS PER SCAN IN CX
        XOR     DX, DX
        MOV     AX, TOT_SCANS           ;NUMBER OF SCANS IN AX:DX
        MUL     CX                      ;MULT BY NUMBER OF WORDS IN CX
        MOV     END_OFFSET, AX          ;end offset value for buffer
        MOV     AX, DX                  ;DX = # OF SEGMENTS
        XOR     DX, DX
        MOV     CX, 1000h               ;GET # OF SEGMENTS * 1000h
        MUL     CX
        MOV     CX, START_SEG           ;ADD TO STARTING SEGMENT
        ADD     CX, AX
        MOV     END_SEGMENT, CX         ;end segment value for buffer
        
        OR      STATUS, 4               ;"run" bit in staus nibble
        CALL    SENDIT                  ;send start stamp scan command
        CMP     ERRORCODE, 0            ;check for error
        JE      CHECK_STRIGGER          ;command sent ok
        MOV     AX, NSCANLO
        MOV     ES, L_RESPONSE_SEG      ;segment and offset to response
        MOV     BX, L_RESPONSE_OFS
        MOV     ES:[BX], AX             ;number of scans completed
        MOV     AX, NSCANHI
        MOV     ES:[BX+2], AX
        JMP     T_EXIT_SCANS            ;else exit scan with error

CHECK_STRIGGER:
        CMP     TRIGGERTYPE, 0          ;check for trigger scan start
        JE      EXITSTARTSTAMP          ;no, exit library
        CALL    CHK2                    ;wait for trigger or ESC to cancel
EXITSTARTSTAMP:
        MOV     ISR_ERROR, 0            ;clear isr error
        MOV     START_ISR, 0            ;flag to start gathering data
        MOV     AX, NSCANLO
        MOV     ES, L_RESPONSE_SEG      ;segment and offset to response
        MOV     BX, L_RESPONSE_OFS
        MOV     ES:[BX], AX             ;number of scans completed
        MOV     AX, NSCANHI
        MOV     ES:[BX+2], AX
        JMP     T_EXIT_SCANS            ;else exit scan with error
STAMPSCAN ENDP

;S*******   This routine performes special offset correction scan.   ********

OFFSET_CORRECTION PROC    NEAR
        
        MOV     ES, L_RESPONSE_SEG      ;segment and offset to response
        MOV     BX, L_RESPONSE_OFS
        MOV     AX, ES:[BX]
        MOV     TOT_SCANS, AX           ;total number of scans
        MOV     AX, ES:[BX + 2]
        MOV     POWER2, AX              ;number of shifts
        
        MOV     NUMSCANS, 0             ;no scans done
        MOV     STOP_FLAG, 0            ;clear STOP_FLAG
        MOV     OFS_HI, 0               ;initialize scan offset dword
        MOV     OFS_LO, 0            
        MOV     ZCOR_HI, 0              ;initialize zero correction dword
        MOV     ZCOR_LO, 0
        
        OR      STATUS, 4               ;"run" bit in staus nibble
        MOV     OUTB1, 6                ;signal normal scan
        CALL    SENDIT                  ;send normal scan command
        CMP     ERRORCODE, 0            ;check for error
        JE      CHECK_OTRIGGER
        JMP     EXIT_SCANS
CHECK_OTRIGGER:
        CMP     TRIGGERTYPE, 0          ;check for trigger scan start
        JE      START_OFFSET_SCAN
        CALL    CHK2                    ;wait for trigger or ESC to cancel
        CMP     ERRORCODE, 0            ;check for error
        JE      START_OFFSET_SCAN
        JMP     EXIT_SCANS

START_OFFSET_SCAN:
        MOV     OFFSET_FLAG, 0          ;clear OFFSET_FLAG 
        XOR     CX, CX
        MOV     CL, NUM_DATAPOINTS      ;setup loop counter for scan
        MOV     AX, TOT_SCANS           ;check to see if last scan
        CMP     AX, NUMSCANS
        JNE     OFFSET_POINTS           ;if on last scan, then one more scan
OENDSCAN:        
        AND     STATUS, 3Bh             ;clear bit #2, "stop scan" bit
        MOV     STOP_FLAG, 1            ;flag send end if scan
OFFSET_POINTS:
        PUSH    CX                      ;store loop counter on stack
        CALL    CHK1                    ;get one 12 bit word from the M250
        CMP     ERRORCODE, 0            ;check for error
        JNE     OFFSET_EXIT
        
        CMP     STOP_FLAG, 1            ;if end of scan, don't add to
        JE      DC6                     ;accumulators
        
        MOV     LOW14, 0                ;clear LOW14
        CLC
        RCR     CX, 1                   ;carry contains msb of LOW14
        
        MOV     HIGH12, CX              ;save HIGH12
        
        JNC     DC6
        MOV     LOW14, 2000h            ;set msb of LOW14
DC6:
        CALL    LOOPCHK1                ;call CHK1 until data ready or timeout
        CMP     ERRORCODE, 0            ;check for error
        JNE     O_SCANERR               ;exit scan with error
        
        CMP     STOP_FLAG, 1            ;if end of scan, don't add to
        JE      SKIP_OFFSET_CALC        ;accumulators
        
        ADD     CX, LOW14               ;LOW14 in CX, add msb
        
        CALL    GETCOUNTS               ;returns low word of count in CX,
                                        ;high word in AX
        CALL    FIND_OFFSETS            ;get zero correction channel and
                                        ;offset channel
SKIP_OFFSET_CALC:        
        POP     CX                      ;pop loop counter off of stack
        LOOP    OFFSET_POINTS
        CMP     STOP_FLAG, 1            ;end of scan?
        JE      O_SKIP_NUMSCANS         ;skip inc numscans
        INC     NUMSCANS                ;did a scan inc numscans
O_SKIP_NUMSCANS:        
        MOV     AX, 100H
;;if there was a keystroke, exit scan
        INT     16H
        JZ      O_OUTKEY                ;no key, get next scan
        JMP     OENDSCAN
O_OUTKEY:
        JMP     START_OFFSET_SCAN
OFFSET_EXIT:
        CMP     STOP_FLAG, 1            ;end of scan ?
        JNE     O_SCANERR
        MOV     ERRORCODE, 0            ;clear ERRORCODE
O_SCANERR:
        POP     CX                      ;pop stored loop counter if error
        
        MOV     AX, ZCOR_LO             ;setup to divide zero correction
        MOV     BX, ZCOR_HI
        
        MOV     CX, POWER2              ;power of 2
        JCXZ    NO_ZCOR_POWER
ZCORDIVIDE:                             ;shift 1 to 9 powers of 2
        CLC
        SHR     BX, 1
        RCR     AX, 1
        LOOP    ZCORDIVIDE
NO_ZCOR_POWER:
        MOV     ZCOR_HI, BX             ;store average zero correction
        MOV     ZCOR_LO, AX
        
        MOV     AX, OFS_LO              ;setup to divide scan offset
        MOV     BX, OFS_HI
        
        MOV     CX, POWER2              ;power of 2
        JCXZ    NO_OFS_POWER
OFSDIVIDE:                              ;shift 1 to 9 powers of 2
        CLC
        SHR     BX, 1
        RCR     AX, 1
        LOOP    OFSDIVIDE
NO_OFS_POWER:
        MOV     OFS_HI, BX              ;store average offset 
        MOV     OFS_LO, AX
        
        MOV     CX, ZCOR_LO             ;load average zero correction
        MOV     DX, ZCOR_HI
        SUB     AX, CX                  ;offset - zcor
        SBB     BX, DX
        
        CLC                             ;divide difference by 4
        SAR     BX, 1
        RCR     AX, 1
        CLC                             
        SAR     BX, 1
        RCR     AX, 1

        ADD     OFS_LO, AX              ;store offset + (offset - zcor)/4
        ADC     OFS_HI, BX              ;in OFS_LO and OFS_HI
        
        MOV     ES, L_RESPONSE_SEG      ;segment and offset to response
        MOV     BX, L_RESPONSE_OFS
        CMP     ERRORCODE, 0            ;check for error
        JE      OFFSET_COR_EXIT         ;no error, return
        MOV     AX, RESPONSE            ;put M250 error code in response
        MOV     ES:[BX], AX
OFFSET_COR_EXIT:
        RET
OFFSET_CORRECTION ENDP

;S************   Get the zero correction and offset channels   **************

FIND_OFFSETS PROC NEAR
        
        CMP     OFFSET_FLAG, 0          ;zero channel offset ?
        JNE     OFFSET_CHANNEL          ;no, get zero channel
        CLC
        ADD     ZCOR_LO, CX             ;add zero correction value
        ADC     ZCOR_HI, AX
        JMP     EXIT_FIND_OFFSETS       ;exit find offsets
OFFSET_CHANNEL:                         
        CLC
        ADD     OFS_LO, CX              ;add offset value
        ADC     OFS_HI, AX
EXIT_FIND_OFFSETS:
        INC     OFFSET_FLAG             ;inc OFFSET_FLAG for zero channel
        RET
FIND_OFFSETS ENDP

;!*******************   M250 SELF CALIBRATION ROUTINES   ********************
;C****************   setup M250 long-term self-calibration   ****************

SETSELFCAL     PROC NEAR
       
        MOV     ES, L_RESPONSE_SEG      ;segment and offset to response
        MOV     BX, L_RESPONSE_OFS      ;self calibration data
        CMP     OUTB1, 0                ;index 0 ?
        JNE     NEXT_DATA1
        MOV     AX, ES:[BX]             ;get single calibration flag
        MOV     SINGLECAL, AL           
        MOV     AX, ES:[BX + 2]         ;get TESTDATA start
        MOV     TESTDATA, AX            
        JMP     EXIT_SELF
NEXT_DATA1:
        CMP     OUTB1, 1                ;index 1 ?
        JNE     NEXT_DATA2
        MOV     AX, ES:[BX]
        MOV     AVRG, AL                ;AVRG (inner loop counter)
        MOV     AVRGSHIFT, AH           ;number of shifts for avrg
        CMP     AH, 0                   ;is AVRGSHIFT 0 ?
        JNE     AVRG_NOT_ZERO           ;no, AVRGSHIFT ok
        MOV     ERRORCODE, 1Ah          ;set ERRORCODE to
        JMP     EXIT_SELF               ;avg shift counter zero error
AVRG_NOT_ZERO:
        MOV     AX, ES:[BX + 2]
        MOV     AVRG1, AL               ;AVGR1 (outer loop counter)
        MOV     AVRG1SHIFT, AH          ;number of shifts for avrg1
        CMP     AH, 0                   ;is AVRGSHIFT1 0 ?
        JNE     EXIT_SELF               ;no, AVRGSHIFT1 ok
        MOV     ERRORCODE, 1Ah          ;set ERRORCODE to
        JMP     EXIT_SELF               ;avg1 shift counter zero error
NEXT_DATA2:
        CMP     OUTB1, 2                ;index 2 ?
        JNE     EXIT_SELF
        MOV     AX, ES:[BX]
        MOV     MONODELAY, AX           ;delay for mono in usec
        MOV     AX, ES:[BX + 2]
        MOV     START_SEG, AX           ;get start segment byte 
EXIT_SELF:
        RET
SETSELFCAL ENDP

;C**************   preform M250 long-term self-calibration   ****************

SELFCAL PROC    NEAR
        
        MOV     ES, L_RESPONSE_SEG      ;segment and offset to response
        MOV     BX, L_RESPONSE_OFS
        MOV     AX, ES:[BX]
        MOV     DI, AX                  ;get offset for cor array
        MOV     AX, START_SEG
        MOV     ES, AX                  ;set seg to save data
        
        XOR     AX, AX
        MOV     ES:[DI], AX             ;zero first element
        
        MOV     ERRORCODE, 0            ;clear ERRORCODE
        MOV     STEP_512, 200h          ;fill table step size
        MOV     MASK_16, 10h            ;mask for exception code
        MOV     COR_16, 10h             ;flag for filling block copy array
        
        MOV     SI, 2                   ;start of block copy array
        MOV     BX, OFFSET TEMPXY_ARRAY ;offset to block copy array
        
        MOV     AX, TESTDATA
        SHL     AX, 1                   ;mult by 2
        ADD     DI, AX                  ;start at TESTDATA offset
TG1:    
        CALL    DOMONODELAY
        CALL    DOMONODELAY
        CALL    DOMONODELAY
        MOV     AX, 100h
;;if there was a keystroke, exit scan
        INT     16H
        JZ      CALOUTKEY               ;key was not pressed
        JMP     EXITCAL
CALOUTKEY:
        MOV     AX, TESTDATA            
        SUB     AX, HOLDSTEP            ;get step size
        MOV     DACDATA, AX
        CMP     DACDATA, 0
        JB      STEPERROR               ;is DACDATA in bounds ?
        CMP     DACDATA, 4095
        JLE     DACDATAOK
STEPERROR:
        MOV     ERRORCODE, 16H          ;set dac data out of bounds error
        MOV     HOLDSTEP, 0             ;re-initialize HOLDSTEP
        JMP     EXITCAL
DACDATAOK:
        CALL    SETBOTHDACS             ;set dacs
        CMP     ERRORCODE, 0            ;check for error
        JE      SETOK
        JMP     EXITCAL
SETOK:
        XOR     CX, CX
        MOV     CL, AVRG1               ;outer loop counter
        MOV     TEMPYHI, 0
        MOV     TEMPYLO, 0
        MOV     TEMPXHI, 0
        MOV     TEMPXLO, 0
OUTLOOP:
        PUSH    CX                      ;store outer loop counter
        JMP     DO_REMAINDER            ;GET REMAINDERS
REMAINDER_DONE:
        POP     CX                      ;pop loop counter off stack
        LOOP    OUTLOOP

        CALL    GETCHKSUM
        CMP     ERRORCODE, 17h          ;check for checksum error
        JNE     CHKSUMOK
        CALL    DO_INIT                 ;re-initialize if wrong checksum 
        JMP     DACDATAOK               ;checksum error, try again

CHKSUMOK:
        CMP     ERRORCODE, 0            ;check for error
        JE      CONTLOOP                ;other error, not chksum so exit
        JMP     EXITCAL
CONTLOOP:
        MOV     DX, TEMPYHI             ;setup to divide y remainders
        MOV     AX, TEMPYLO

        XOR     CX, CX
        MOV     CL, AVRG1SHIFT          ;divide by power of 2 of AVRG1
TEMPYLOOP:
        CLC
        SHR     DX, 1                   ;rotate DX through carry
        RCR     AX, 1                   ;into AX
        LOOP    TEMPYLOOP
        MOV     TEMPY, AX

        MOV     DX, TEMPXHI             ;setup to divide x remainders
        MOV     AX, TEMPXLO

        XOR     CX, CX
        MOV     CL, AVRG1SHIFT          ;divide by power of 2 of AVRG1
TEMPXLOOP:
        CLC
        SHR     DX, 1                   ;rotate DX through carry
        RCR     AX, 1                   ;into AX
        LOOP    TEMPXLOOP
        CMP     SINGLECAL, 0
        JZ      SC2
        JMP     EXITSINGLECAL           ;if single, exit
SC2:
        MOV     DX, TEMPY
        SUB     AX, DX
        SUB     AX, 2044                ;TEMPX-TEMPY-2044
        MOV     CX, COR_16
        CMP     TESTDATA, CX            ;block copy array value ?
        JGE     NOT_ARRAY               ;exception case, put in cal array
        MOV     DS:[BX + SI], AX        ;store result
        INC     SI
        INC     SI                      ;point to next value
NOT_ARRAY:
        ADD     AX, ES:[DI - 2]         ;AX = TEMPX-TEMPY-2044+LASTCOR
        
        MOV     ES:[DI], AX
        ADD     DI, 2
        MOV     DX, 1030H
        CALL    DISP_AX                 ;display current cor count
        MOV     AX, TESTDATA
        MOV     DX, 1040H
        CALL    DISP_AX                 ;display current TESTDATA
        INC     TESTDATA                ;inc to next TESTDATA
        CMP     TESTDATA, 4096          ;if TESTDATA = 4096, done
        JE      EXITCAL

        MOV     CX, MASK_16             ;test for exception
        CMP     TESTDATA, CX            ;if not exception, get next 
        JNE     JMP_TG1                 ;value from block copy array
        MOV     SI, 2                   ;reset array pointer
        ADD     MASK_16, 10h            ;set exception mask to next value
        MOV     CX, STEP_512
        CMP     TESTDATA, CX            ;next step size ?
        JNE     NOT_NEW_ARRAY
        MOV     CX, STEP_512
        MOV     COR_16, CX              ;update flag for filling
        ADD     COR_16, 10h             ;block copy array
        ADD     STEP_512, 200h          ;next step size value
        MOV     SI, 0                   ;reset array pointer
NOT_NEW_ARRAY:                          ;do next value
        JMP     TG1
JMP_TG1:
        MOV     CX, COR_16
        CMP     TESTDATA, CX            ;still flling block copy array ?
        JG      REPEAT_512              ;no, put array value in ax
        JMP     TG1                     ;yes, get next value

REPEAT_512:
        MOV     AX, [BX + SI]           ;get TEMPX-TEMPY-2044 value
        ADD     SI, 2                   ;point to next array value
        JMP     NOT_ARRAY               ;put it in cal array

gototg1:
        MOV     DX, 500                 ;wait 500 usec
        XOR     CX, CX
        MOV     AX, 8600h
        INT     15h
        POP     CX
        JMP     TG1
POPCXEXIT:
        POP     CX                      ;pop CX off stack if error
EXITCAL:
        MOV     ES, L_RESPONSE_SEG      ;segment and offset to response
        MOV     BX, L_RESPONSE_OFS
        MOV     AX, RESPONSE
        MOV     ES:[BX], AX
        XOR     AX, AX
        MOV     ES:[BX + 2], AX
        JMP     EXITRET
EXITSINGLECAL:      
        PUSH    AX
        MOV     ES, L_RESPONSE_SEG      ;segment and offset to response
        MOV     BX, L_RESPONSE_OFS
        MOV     ES:[BX], AX             ;TEMPX in AX
        MOV     AX, TEMPY
        MOV     ES:[BX + 2], AX
        MOV     DI, OFFSET COR_ARRAY
        CLD                             ;select auto-increment
        MOV     CX, 8192
        XOR     AX, AX
        REP     STOSW
        POP     AX
        MOV     DX, TEMPY
        SUB     AX, DX
        SUB     AX, 2044                ;TEMPX-TEMPY-2044
        MOV     BX, TESTDATA            ;get index for TESTDATA word
        SHL     BX, 1                   ;make word index
        MOV     SI, OFFSET COR_ARRAY    ;get offset to correction array
        MOV     [SI + BX], AX           ;store result
EXITRET:      
        RET
SELFCAL ENDP


DO_REMAINDER:                           ;get 2 remainders and check bounds
        CALL    DOMONODELAY
        MOV     AX, TESTDATA
        DEC     AX
        CALL    GETREM                  ;get remainders on TESTDATA, passed
                                        ;back in AX
        CMP     ERRORCODE, 0            ;check for error
        JE      REMYOK
        JMP     POPCXEXIT
REMYOK:
        CMP     AX, X_HI_BOUND          ;x remainder in bounds ?
        JBE     CHECK_LO_X              ;yes, check low bound
        INC     HOLDSTEP                ;no, inc holdstep and try again
        jmp     gototg1
CHECK_LO_X:
        CMP     AX, X_LO_BOUND          ;x remainder in bounds ?
        JAE     REMAINDER_X_OK          ;yes, accumulate x remainder
        DEC     HOLDSTEP                ;no, dec holdstep and try again
        jmp     gototg1
REMAINDER_X_OK:
        CLC
        ADD     TEMPXLO, AX             ;accumulate x remainders
        JNC     NO_X_CARRY
        INC     TEMPXHI                 ;carry bit
NO_X_CARRY:
        CALL    DUMMYCONV               ;get conversion, throw away result
        CMP     ERRORCODE, 0            ;check for error
        JE      XDUMMY_OK
        JMP     POPCXEXIT               ;if error, exit
XDUMMY_OK:        
        CALL    DOMONODELAY             ;wait MONODELAY usec
        MOV     AX, TESTDATA
        CALL    GETREM                  ;get remainders on TESTDATA, passed
                                        ;back in AX
        CMP     ERRORCODE, 0            ;check for error
        JE      REMXOK
        JMP     POPCXEXIT
REMXOK:
        CMP     AX, Y_HI_BOUND          ;y remainder in bounds ?
        JBE     CHECK_LO_Y              ;yes, check low bound
        INC     HOLDSTEP                ;no, inc holdstep and try again
        jmp     gototg1
CHECK_LO_Y:
        CMP     AX, Y_LO_BOUND          ;y remainder in bounds ?
        JAE     REMAINDER_Y_OK          ;yes, accumulate y remainder
        DEC     HOLDSTEP                ;no, dec holdstep and try again
        jmp     gototg1
REMAINDER_Y_OK:
        CLC
        ADD     TEMPYLO, AX             ;accumulate y remainders
        JNC     NO_Y_CARRY
        INC     TEMPYHI                 ;carry bit
NO_Y_CARRY:
        CALL    DUMMYCONV               ;get conversion, throw away result
        CMP     ERRORCODE, 0            ;check for error
        JE      YDUMMY_OK
        JMP     POPCXEXIT               ;if error, exit
YDUMMY_OK:
        JMP     REMAINDER_DONE


;C****************   SET BOTH DACS IN SELF-CALIBRATION   ********************

SETBOTHDACS PROC NEAR

        SHL     AX, 1                   ;mult by 4
        SHL     AX, 1
        MOV     DACDATA, AX
        MOV     OUTB3, 7
        MOV     OUTB2, AH
        AND     OUTB2, 0Fh              ;second nibble of command
        SHR     AH,1                    ;divide by 16
        SHR     AH,1
        SHR     AH,1
        SHR     AH,1
        AND     AH, 0Fh
        MOV     OUTB1, AH               ;third nibble of command
        CALL    SENDIT
        CMP     ERRORCODE, 0            ;check for error
        JNE     SETDACERR

        MOV     AX, DACDATA
        MOV     OUTB3, 8
        MOV     OUTB2, AL
        AND     OUTB2, 0Fh              ;second nibble of command
        SHR     AL,1                    ;divide by 16
        SHR     AL,1
        SHR     AL,1
        SHR     AL,1
        AND     AL, 0Fh
        MOV     OUTB1, AL               ;third nibble of command
        CALL    SENDIT
        CMP     ERRORCODE, 0            ;check for error
        JNE     SETDACERR
      
        MOV     DX, 65                  ;wait 65 usec
        XOR     CX, CX
        MOV     AX, 8600h
        INT     15h

SETDACERR:
        RET
SETBOTHDACS ENDP

DOMONODELAY PROC NEAR
        MOV     DX, MONODELAY           ;wait monodelay usec 
        SHL     DX,1                    ;multiply by 16
        SHL     DX,1
        SHL     DX,1
        SHL     DX,1
        XOR     CX, CX                  ;no high byte
        MOV     AX, 8600h
        INT     15h
        RET
DOMONODELAY ENDP

;C****************   GET REMAINDERS FOR SELF-CALIBRATION   ******************

GETREM PROC NEAR

        SHL     AX, 1                   ;mult by 4
        SHL     AX, 1
        ADD     AX, 2                   ;add 2
        MOV     DACDATA, AX
        CALL    MONO                    ;remainder passed back in AX
        CMP     ERRORCODE, 0            ;check for error
        JE      MONO1OK
        JMP     REMERR
MONO1OK:
        MOV     TOTREMLO, 0             ;clear remainder dword
        MOV     TOTREMHI, 0

        XOR     CX, CX
        MOV     CL, AVRG
REMLOOP:
        PUSH    CX

        MOV     DX, MONODELAY           ;wait monodelay usec 
        XOR     CX, CX                  ;no high byte
        MOV     AX, 8600h
        INT     15h

        MOV     AX, DACDATA
        CALL    MONO                    ;remainder passed back in AX
        CMP     ERRORCODE, 0            ;check for error
        JE      MONO2OK
        POP     CX
        JMP     REMERR
MONO2OK:
        CLC
        ADD     TOTREMLO, AX            ;accumulate remainders
        JNC     NOCARRY
        INC     TOTREMHI                ;carry bit
NOCARRY:
        POP     CX
        LOOP    REMLOOP

        MOV     DX, TOTREMHI            ;setup to divide remainders
        MOV     AX, TOTREMLO

        xor     cx, cx
        MOV     CL, AVRGSHIFT           ;divide by power of 2 of AVRG
ROTATELOOP:
        CLC
        SHR     DX, 1                   ;rotate DX through carry
        RCR     AX, 1                   ;into AX
        LOOP    ROTATELOOP
REMERR:
        RET
GETREM  ENDP

;C*************  GET REMAINDER CONVERSION FOR SELF CALIBRATION   ************

MONO    PROC NEAR

        MOV     OUTB3, 9                ;table generation command
        MOV     OUTB1, AH
        AND     OUTB1, 0Fh              ;first nibble of command
        SHR     AH,1                    ;divide by 16
        SHR     AH,1
        SHR     AH,1
        SHR     AH,1
        AND     AH, 0Fh
        MOV     OUTB2, AH               ;second nibble of command
        CALL    SENDIT
        CMP     ERRORCODE, 0            ;check for error
        JNE     MONOERR

        MOV     AX, DACDATA
        MOV     OUTB3, AL
        AND     OUTB3, 0Fh              ;third nibble of command
        SHR     AL,1                    ;divide by 16
        SHR     AL,1
        SHR     AL,1
        SHR     AL,1
        AND     AL, 0Fh
        MOV     OUTB2, AL               ;second nibble of command
        CALL    SENDIT
        CMP     ERRORCODE, 0            ;check for error
        JNE     MONOERR

        CALL    CHK1
        CMP     ERRORCODE, 0            ;check for error
        JNE     MONOERR

        AND     CX, 1

        ROR     CX, 1                   ;mult by 8192
        ROR     CX, 1
        ROR     CX, 1
        PUSH    CX

        CALL    CHK1
        POP     AX                      ;put CX in AX
        ADD     AX, CX                  ;save remainder in AX
MONOERR:
        RET
MONO    ENDP

;!*************************   LOAD FILE ROUTINES   **************************
;L*************************   GET SERIAL NUMBER   ***************************

GET_SERIAL PROC NEAR        
        MOV     ES, L_RESPONSE_SEG      ;segment and offset to response
        MOV     BX, L_RESPONSE_OFS
        MOV     AX, ES:[BX]
        MOV     AL, AH                  ;get first 2 digits
        AAM                             ;split denary digits
        OR      AX,"00"                 ;convert to ASCII
        XCHG    AL,AH                   ;reverse digits
        MOV     [SI], AX                ;store them
        ADD     SI, 2                   ;index to next location
        
        MOV     AX, ES:[BX]             ;get second 2 digits
        AAM                             ;split denary digits
        OR      AX,"00"                 ;convert to ASCII
        XCHG    AL,AH                   ;reverse digits
        MOV     [SI], AX                ;store them
        RET        
GET_SERIAL ENDP        

;L***********************   LOAD CALIBRATION FILE   *************************

GET_COR_FILE PROC   NEAR        
        CLD                             ;string moves forward
        MOV     CORERR, 0               ;no file read error
        MOV     AX, 3D00H               ;open file for reading
        LEA     DX, COR_FILE            ;load DX with the name of the file
        INT     21H
        MOV     [SRC_HANDLE], AX        ;save source handle
        JNC     COR_FILE_EXISTS         ;if carry flag set, then file does
                                        ;not exist
        MOV     CORERR, 1               ;file read error
        JMP     EXIT_COR                ;exit reading the file with error set
COR_FILE_EXISTS:
        MOV     AX, 4202H               ;find file size
        MOV     BX, [SRC_HANDLE]        ;handle for source file
        XOR     CX, CX                  ;from zero
        XOR     DX, DX
        INT     21h                     ;file size returned in AX
        JNC     FIND_COR_SIZE           ;if carry flag set, then movement of
                                        ;file pointer failed
        MOV     CORERR, 3               ;file pointer movement failed
        JMP     CLOSE_COR_FILE          ;close file and exit
FIND_COR_SIZE:        
        CMP     AX, 8199                ;size of the file in bytes
        JE      GOOD_COR_FILE           ;if equal, correct file size
        MOV     CORERR, 2               ;incorrect file size error
        JMP     CLOSE_COR_FILE          ;close file and exit
GOOD_COR_FILE:                          ;load the file into COR_ARRAY
        MOV     AX, 4200H               ;set file pointer
        MOV     BX, [SRC_HANDLE]        ;handle for source file
        XOR     CX, CX                  ;no high pointer (used for files > 64k)
        MOV     DX, 7                   ;low pointer offset into file
        INT     21h
        JNC     MOVE_COR_START          ;if carry flag set, then movement of
                                        ;file pointer failed
        MOV     CORERR, 3               ;file pointer movement failed
        JMP     CLOSE_COR_FILE          ;close file and exit
MOVE_COR_START:        
        MOV     AX, 3F00H               ;read file
        MOV     BX, [SRC_HANDLE]        ;from this handle
        LEA     DX, COR_ARRAY           ;load dx with the pointer to COR_ARRAY
        MOV     CX, 8192                ;get 8192 bytes
        INT     21H                     
        JNC     CLOSE_COR_FILE          ;if carry flag set, then read file
                                        ;failed
        MOV     CORERR, 4               ;read file failed
CLOSE_COR_FILE:                         ;close the file        
        MOV     AH,3EH                  ;close handle fn
        MOV     BX,[SRC_HANDLE]         ;handle for source file
        INT     21H                     
        JNC     EXIT_COR                ;if carry flag set, then close file
                                        ;failed
        MOV     CORERR, 5               ;close file failed
EXIT_COR:        
        RET
GET_COR_FILE ENDP


;L********************   LOAD ABSOLUTE ACCURACY FILE   **********************

GET_ABS_FILE PROC   NEAR
        CLD                             ;string moves forward
                                        ; below is x40
        MOV     BX, OFFSET ABS_ARRAY    ;load bx with pointer to ABS_ARRAY
        XOR     DI, DI                  ;clear 16384 bytes
        XOR     AL, AL                  ;clear data
CLRIT:  MOV     [BX + DI], AL 
        INC     DI
        CMP     DI, 16384
        JNE     CLRIT
                                        ; above is x40
        MOV     ABSERR, 0               ;no file read error
        MOV     AX, 3D00H               ;open file for reading
        LEA     DX, ABS_FILE            ;load DX with the name of the file
        INT     21H
        MOV     [SRC_HANDLE], AX        ;save source handle
        JNC     ABS_FILE_EXISTS         ;if carry flag set, then file does
                                        ;not exist
        MOV     ABSERR, 1               ;file read error
        JMP     EXIT_ABS                ;exit reading the file with error set
ABS_FILE_EXISTS:
        MOV     AX, 4202H               ;find file size
        MOV     BX, [SRC_HANDLE]        ;handle for source file
        XOR     CX, CX                  ;from zero
        XOR     DX, DX
        INT     21h                     ;file size returned in AX
        JNC     FIND_ABS_SIZE           ;if carry flag set, then movement of
                                        ;file pointer failed
        MOV     ABSERR, 3               ;file pointer movement failed
        JMP     CLOSE_ABS_FILE          ;close file and exit
FIND_ABS_SIZE:        
        CMP     AX, 16391               ;size of the file in bytes
        JE      GOOD_ABS_FILE           ;if equal, correct file size
        MOV     ABSERR, 2               ;ncorrect file size error
        JMP     CLOSE_ABS_FILE          ;close file and exit
GOOD_ABS_FILE:                          ;load the file into ABS_ARRAY    
        MOV     AX, 4200H               ;set file pointer
        MOV     BX, [SRC_HANDLE]        ;handle for source file
        XOR     CX, CX                  ;no high pointer (used for files > 64k)
        MOV     DX, 7                   ;low pointer offset into file
        INT     21h
        JNC     MOVE_ABS_START          ;if carry flag set, then movement of
                                        ;file pointer failed
        MOV     ABSERR, 3               ;file pointer movement failed
        JMP     CLOSE_ABS_FILE          ;close file and exit
MOVE_ABS_START:        
        MOV     AX, 3F00H               ;read file
        MOV     BX, [SRC_HANDLE]        ;from this handle
        LEA     DX, ABS_ARRAY           ;load dx with the pointer to ABS_ARRAY
        MOV     CX, 16384               ;get 16384 bytes
        INT     21H                     
        JNC     CLOSE_ABS_FILE          ;if carry flag set, then read file
                                        ;failed
        MOV     ABSERR, 4               ;read file failed
CLOSE_ABS_FILE:                         ;close the file        
        MOV     AH,3EH                  ;close handle fn
        MOV     BX,[SRC_HANDLE]         ;handle for source file
        INT     21H                     
        JNC     EXIT_ABS                ;if carry flag set, then close file
                                        ;failed
        MOV     ABSERR, 5               ;close file failed
EXIT_ABS:
        RET
GET_ABS_FILE ENDP


;!**************************   DISPLAY ROUTINES   ***************************
;D********   This procedure will display the contents of AX in hex   ********
;display AX in hex at DH row, DL col

DISP_AX PROC    NEAR
        PUSH    BX
        PUSH    AX
        XOR     BX, BX
        MOV     AH, 2
        INT     10H
        POP     AX
        CALL    HEX4
        POP     BX
        RET
DISP_AX ENDP

;D*****************   write AX as 4 hex digits to std out   *****************

HEX4    PROC NEAR       
        PUSH    AX                      ;save register
        MOV     AL,AH                   ;show high digits first
        CALL    HEX2                    ;display AL
        POP     AX                      ;restore low digits in AL

;D*****************   write AL as 2 hex digits to std out   *****************

HEX2    PROC NEAR       
        PUSH    AX                      ;save register
        PUSH    CX                      ;save CX during shift
        MOV     CL,4
        SHR     AL,CL                   ;get high 4 bits
        POP     CX                      ;restore CX
        CALL    H2OUT                   ;display upper AL digit
        POP     AX                      ;restore lower
        AND     AL,0FH                  ;mask and display

;D*********   write lower 4 bits of AL as 1 hex digit to std out   **********

H2OUT   PROC NEAR
        PUSH    DX
        ADD     AL,90H                  ;convert AL to ASCII
        DAA
        ADC     AL,40H
        DAA
        MOV     DL,AL                   ;char in DL
        MOV     AH,2                    ;display
        INT     21H                     ;thru DOS
        POP     DX
        RET
H2OUT   ENDP
HEX2    ENDP
HEX4    ENDP

;!**********************   TIMER INTERRUPT ROUTINES   ***********************
;T*** setup 8254 mode 2 timer and hook the INT 8 timer service interrupt ****

SET_INT8   PROC NEAR
        
        MOV     SI, 0                   ;start at 0 offset
        MOV     HEAD_POINTER, SI        ;store starting offset for head and 
        MOV     TAIL_POINTER, SI        ;tail pointers        
        MOV     WRAP_POINTER, 0
        CALL    SETIMER                 ;setup timer
        PUSH    DS                      ;store data seg
        PUSH    ES                      ;store extra seg
        MOV     AL, 8                   ;interrupt number
        MOV     DI, OFFSET OLDINT8      ;store vector here
        MOV     DX, OFFSET INT_8        ;new interrupt procedure
        PUSH    CS
        POP     DS                      ;force data seg = code seg
        
;       AL = vector number.
;       ES:DI = destination for old address.
;       DS:DX = new interrupt address.
        
        PUSH    AX                      ;save vector # in AL
        MOV     AH,35h                  ;get Int vector
        INT     21h                     ;BX = offset, ES = segment to vector
        MOV     [DI+0],BX               ;save address in ES:DI
        MOV     [DI+2],ES
        POP     AX                      ;get AL back
        MOV     AH,25h                  ;set Int vector
        INT     21h
        
        POP     ES                      ;restore extra seg
        POP     DS                      ;restore data seg
        MOV     INTERRUPT_FLAG, 1       ;flag interrupt set
        RET
SET_INT8 ENDP

;T******* remove our hook to INT 8 and restore the 8254 mode 2 timer ********

RESTORE_INT8 PROC NEAR
        
        PUSH    DS                      ;store data seg
        PUSH    ES                      ;store extra seg
        PUSH    CS
        POP     DS                      ;force data seg = code seg
        MOV     AL,8                    ;interrupt number
        MOV     DI,OFFSET OLDINT8       ;put old INT 8 vector back
        MOV     DX,[DI+0]               ;save address in ES:DI
        MOV     DS,[DI+2]
        MOV     AH,25h                  ;set Int vector
        INT     21h
        POP     ES                      ;restore extra seg
        POP     DS                      ;restore data seg
        MOV     LOCOUNT, 0              
        MOV     HICOUNT, 0
        CALL    SETIMER                 ;reset 8254 mode 2 timer
        MOV     INTERRUPT_FLAG, 0
        RET
RESTORE_INT8 ENDP

;T*************   This routine gets control each timer tick.   **************

INT_8 PROC FAR
        
        STI                             ;enable interrupts
        PUSH    AX                      ;save AX
        PUSH    DS                      ;save data seg
        MOV     AX, @DATA
        MOV     DS, AX                  ;set data seg

INT8_ISR:                               ;perform our ISR routine
       
        CMP     START_ISR, 0            ;start gathering data?
        JE      ISR_READY               ;yes, perform ISR
        JMP     EXIT_DIG_CMDS   ;EXIT_ISR                ;no, exit ISR
ISR_READY:
        
        PUSH    BX                      ;store used registers
        PUSH    CX
        PUSH    DX
        PUSH    SI
        PUSH    ES
        MOV     AX, @FARDATA?
        MOV     ES, AX                  ;set seg to save data
        MOV     SI, HEAD_POINTER        ;get current head pointer into array
        
        MOV     DX, OUTPORT             ;set DX to OUTPORT
        
        CMP     ISR_ERROR, 0            ;check for errors
        JBE     CHK_DATAREADY           ;no error, normal status read
        
        MOV     AL, STATUS
        AND     AL, 0DFH                ;enable reading status reg
        OUT     DX, AL
        CALL    DELAYRW                 ;delay for write strobe
        INC     DX                      ;INPORT
CHKSTAT6:
        IN      AL, DX                  ;get M250 status in high nibble
        MOV     AH, AL
        IN      AL, DX                  ;get M250 status in high nibble again
        AND     AX, 0F0F0H              ;mask off low bits
        CMP     AL, AH                  ;check to see if equal
        JNE     CHKSTAT6
        TEST    AL, 80H                 ;check for error bit
        JNE     ERROR_RECOVER           ;if not equal, no error
        MOV     ISR_ERROR, 17h          ;set to go through error recover
        JMP     NO_DATAREADY            ;there was an error, exit ISR
ERROR_RECOVER:
        CMP     ISR_ERROR, 27h          ;wait for error bit to go high
        JNE     DO_RECOVER
        JMP     NO_DATAREADY
DO_RECOVER:
        
        AND     STATUS, 3Eh             ;initialize STATUS
        MOV     NEWDATA, 0              ;initialize NEWDATA
        MOV     ISR_ERROR, 0            ;reset ISR_ERROR
        MOV     DO_DIG_CMDS, 0
        MOV     DIGIN_STATUS, 0         ;clear flag for getting digital input
                                        ;word
        MOV     AL, STATUS              ;set new status
        CALL    WRITENIBBLE
        JMP     NO_DATAREADY            ;exit ISR
CHK_DATAREADY:        
        MOV     AL, STATUS
        AND     AL, 0DFH                ;enable reading status reg
        OUT     DX, AL
        CALL    DELAYRW                 ;delay for write strobe
        INC     DX                      ;INPORT
CHKSTAT5:
        IN      AL, DX                  ;get M250 status in high nibble
        MOV     AH, AL
        IN      AL, DX                  ;get M250 status in high nibble again
        AND     AX, 0F0F0H              ;mask off low bits
        CMP     AL, AH                  ;check to see if equal
        JNE     CHKSTAT5
        AND     AH, 20H                 ;mask for NEWDATA bit
        CMP     AH, NEWDATA             ;data ready ?
        JNE     OUTCHECK2               ;read new data
        TEST    AL, 80H                 ;check for error bit
        JNE     NDE1                    ;no error bit
NDE:                                    ;there was an error bit
        MOV     ES:[SI], 4000h          ;flag bad scan
        ADD     SI, 2
        JNZ     NOT_WRAP
        INC     WRAP_POINTER            ;inc WRAP_POINTER if pointer wrapped
                                        ;to zero
NOT_WRAP:        
        MOV     ISR_ERROR, 17h          ;flag error bit hi
NDE1:        
        JMP     NO_DATAREADY            ;there was an error, exit ISR
OUTCHECK2:        
        ; DX is INPORT at this point   status reg read is selected
        TEST    AL, 80H                 ;check for error bit
        JNE     NOREADERR1
        JMP     NDE                     ;there was an error
NOREADERR1:
        MOV     CX, 1000H
        AND     AL, 40H
        JNZ     NOT13I
        XOR     CX, CX
NOT13I: MOV     AL, 0D0h                ;select reg3
        CALL    READNIBBLE
        SHR     AL, 1                   ;divide by 16
        SHR     AL, 1
        SHR     AL, 1
        SHR     AL, 1
        ADD     CX, AX                  ;save reg3
        MOV     AL, 90h                 ;select reg2
        CALL    READNIBBLE
        ADD     CX, AX                  ;save reg2
        MOV     AL, 50h                 ;select reg1
        CALL    READNIBBLE
        SHL     AX, 1                   ;mult by 16
        SHL     AX, 1
        SHL     AX, 1
        SHL     AX, 1
        ADD     CX, AX                  ;save reg1
        
        XOR     NEWDATA, 20H            ;toggle NEWDATA
        XOR     STATUS, 1               ;toggle HostAck
        MOV     AL, STATUS              ;reg0, select lines both = 0
        CALL    WRITENIBBLE
        
        MOV     ES:[SI], CX             ;put word in timer interrupt array
        ADD     SI, 2                   ;increment to next location
        JNZ     NO_DATAREADY
        INC     WRAP_POINTER            ;inc WRAP_POINTER if pointer wrapped
                                        ;to zero
NO_DATAREADY:
        MOV     HEAD_POINTER, SI        ;store new head pointer
        POP     ES                      ;restore used registers
        POP     SI
        POP     DX
        POP     CX
        POP     BX
EXIT_ISR:
        CMP     DO_DIG_CMDS, 0
        JE      EXIT_DIG_CMDS           ;no commands to send
        CMP     DO_DIG_CMDS, 3          ;digital output command ?
        JNE     NO_DIGOUT
        CALL    SEND_DIGOUT             ;yes, send command
        JMP     EXIT_DIG_CMDS
NO_DIGOUT:
        CMP     DO_DIG_CMDS, 4          ;digitial input command ? 
        JNE     EXIT_DIG_CMDS           ;no, exit
        CALL    SEND_DIGIN              ;send digitial input command
EXIT_DIG_CMDS:
        MOV     AX, DO_OLDINT8          ;move compare 
        INC     TICK_COUNTER            ;increment counter
        CMP     TICK_COUNTER, AX        ;time to service interrupt?
        JB      NOT_TO_DOS              ;no, exit 
        
        MOV     TICK_COUNTER, 0         ;restart counter
        POP     DS                      ;restore data seg
        POP     AX                      ;restore AX
        DB      0EAH                    ;Opcode for jump far
OLDINT8 DD      -1                      ; located here

NOT_TO_DOS:
        MOV     AL, 20h                 ;tell host that we did interrupt
        OUT     20h, AL
        POP     DS                      ;restore data seg
        POP     AX                      ;restore AX
        IRET                            ;interrupt return
INT_8 ENDP


;T**   This procedure compares two DWORDS in a time-stamp number format   ***

CHK_STAMP_NUM  PROC NEAR
        CALL    GET_STAMP_NUM           ;stamp number is passed back in AX:DX
        PUSH    AX
        PUSH    DX
        MOV     BX, STORE_1ST_WORD
        MOV     DX, STORE_2ND_WORD
        CALL    GET_STAMP_NUM           ;stamp number is passed back in AX:DX
        POP     CX
        POP     BX
                
        SUB     BX, AX                  ;subtract the 2 time stamp numbers
        SBB     CX, DX
        CMP     CX, 0
        JNE     BAD_STAMP_NUM
        CMP     BX, 3                   ;if within 3 then good time stamp
        JBE     GOOD_STAMP_NUM
BAD_STAMP_NUM:        
        STC                             ;bad time stamp, set carry flag
        JMP     EXIT_CHK_STAMP
GOOD_STAMP_NUM:
        CLC                             ;good time stamp, clear carry flag
EXIT_CHK_STAMP:        
        RET
CHK_STAMP_NUM ENDP

;T*******  This procedure converts a DWORD into a time-stamp number   *******
;time-stamp number is passed back in AX:DX
GET_STAMP_NUM PROC NEAR
        
        MOV     AX, DX                  ;DX is SECOND_WORD
        AND     AX, 7                   ;get low 3 bits
        MOV     AH, AL                  ;and move them to the top 3 bits
        XOR     AL, AL
        SHL     AX, 1                   
        SHL     AX, 1                   
        SHL     AX, 1
        SHL     AX, 1
        SHL     AX, 1
;BX is FIRST_WORD
        OR      AX, BX                  ;add to low 13 bits  (16 bits)
        SHR     DX, 1                   ;shift off low 3 bits
        SHR     DX, 1
        SHR     DX, 1
        RET
GET_STAMP_NUM   ENDP


;T**********  Send a digital output during time-stamp scanning   ***********

SEND_DIGOUT PROC NEAR        
        PUSH    BX                      ;save used registers
        PUSH    CX
        PUSH    DX
        MOV     AX, DIGOUTWORD
        MOV     OUTB1, AL
        AND     OUTB1, 0Fh              ;first nibble of command
        SHR     AL,1                    ;DIVIDE BY 16
        SHR     AL,1
        SHR     AL,1
        SHR     AL,1
        AND     AL, 0Fh
        MOV     OUTB2, AL               ;second nibble of command
        AND     AH, 0Fh
        MOV     OUTB3, AH               ;third nibble of command
        CALL    SENDIT
        POP     DX                      ;restore used registers
        POP     CX
        POP     BX
        MOV     DO_DIG_CMDS, 0          ;finished sending digital output
        RET
SEND_DIGOUT ENDP

;T*********  Request a digital input during time-stamp scanning   **********

SEND_DIGIN PROC NEAR       
        MOV     DIGIN_STATUS, 1         ;flag waiting for digin, do not send
                                        ;another command while DIGIN_STATUS
                                        ;is set
        MOV     OUTB1, 5
        MOV     OUTB2, 0                ;second nibble of command
        MOV     OUTB3, 0                ;third nibble of command
        PUSH    BX                      ;save used registers
        PUSH    CX
        PUSH    DX
        CALL    SENDIT
        POP     DX                      ;restore used registers
        POP     CX
        POP     BX
        MOV     DO_DIG_CMDS, 0          ;finished requesting digital input
        RET
SEND_DIGIN ENDP

;T***************   This routine sets the 8254 mode 2 timer  ****************

SETIMER     PROC NEAR
        
        CLI                             ;disable interrupts
        MOV     DX, 43H                 ;timer 2 address
        MOV     AL, 36H                 ;set to read lsb, msb count and set
        OUT     DX, AL                  ;to square wave function
        MOV     DX, 40H                 ;counter latch
        MOV     AL, LOCOUNT
        OUT     DX, AL                  ;set lsb count
        MOV     AL, HICOUNT
        OUT     DX, AL                  ;set msb count
        STI                             ;enable interrupts
        RET
SETIMER ENDP

;T***************   This routine gets one time-stamp scan    ****************

GET_TIMER_SCAN    PROC NEAR               ;get one scan
        CALL    GET_NUM_WORDS           ;get number of words
        CMP     BX, 0
        JE      NOT_OVERFLOW3
        MOV     ERRORCODE, 1BH
        JMP     EXIT_SCANS1
NOT_OVERFLOW3:
        CMP     START_ISR, 0
        JNE     LESS_F000
        CMP     AX, 0F000H
        JB      LESS_F000
        MOV     LESS_FLAG, 1
        MOV     START_ISR, 1
LESS_F000:
        CMP     DIGIN_STATUS, 1
        JNE     NO_DIGIN_PENDING
        CALL    CHK_FOR_DIGIN           ;check for digital input word
NO_DIGIN_PENDING:
        MOV     ERRORCODE, 0
        CMP     NUMSCANS, 0             ;first time through ?
        JE      FIRST_SCAN              
        MOV     AX, TOT_SCANS           ;check to see if last scan
        CMP     AX, NUMSCANS
        JNE     START_TIMER_SCAN        ;not last scan, don't reset buffer
        MOV     NUMSCANS, 0
FIRST_SCAN:        
        MOV     CURRENT_DI, 0           ;start at zero offset
        MOV     AX, START_SEG
        MOV     CURRENT_ES, AX          ;reset seg to save data
        MOV     NUMSCANS, 0
START_TIMER_SCAN:
        MOV     ES, CURRENT_ES          ;restore current ES
        MOV     DI, CURRENT_DI          ;restore current DI
        MOV     ERRORCODE, 0            ;clear ERRORCODE
        CALL    PACK_STAMP              ;parse stamp scan number

        CMP     ERRORCODE, 0            ;check for error
        JE      START_T_SCAN            ;no error, good stamp number
        JMP     EXIT_SCANS1             ;exit

START_T_SCAN:        
        XOR     CX, CX                  ;clear counter
        MOV     CL, NUM_DATAPOINTS      ;setup loop counter for scan
T_DATAPOINTS:        
        PUSH    CX                      ;store loop counter on stack
        JMP     GET_TIMER_OFFSET        ;get stamp scan offset correction
RET_GET_TIMER_OFFSET:
        
        CALL    CHKBUFFER               ;get one 12 bit word from buffer
        CMP     ERRORCODE, 0            ;check for error
        JNE     POP_CX_START            ;exit scan and wait for next scan
        
        MOV     LOW14, 0                ;clear LOW14
        CLC                             ;clear carry
        RCR     CX, 1                   ;carry contains msb of LOW14
        
        MOV     HIGH12, CX              ;save HIGH12
        
        JNC     DC11
        MOV     LOW14, 2000h            ;carry set, set msb in LOW14
DC11:
        CALL    CHKBUFFER               ;get one 12 bit word from buffer
        CMP     ERRORCODE, 0            ;check for error
        JNE     POP_CX_START            ;exit scan and wait for next scan
        
        ADD     CX, LOW14               ;LOW14 in CX, add msb
        PUSH    SI                      ;store current array location
        CALL    GETCOUNTS               ;returns low word of count in CX,
                                        ;high word in AX
        POP     SI                      ;restore current array location
        
        SUB     CX, ZCOR_LO             ;subtract offset + stamp scan
        SBB     AX, ZCOR_HI             ;interaction correction
        
        MOV     [SI], CX                ;put dword in temp stamp array
        ADD     SI, 2
        MOV     [SI], AX
        ADD     SI, 2                   ;increment pointer to next location
        
        POP     CX                      ;pop loop counter
        LOOP    T_DATAPOINTS
        
        MOV     AL, STAMP_CHKSUM
        MOV     STORE_CHKSUM, AL
        CALL    CHKBUFFER               ;get M250 stamp checksum
        MOV     STAMP_CHKSUM, 0
        CMP     ERRORCODE, 0            ;check for error
        JNE     T_STAMP_BAD             ;exit scan and wait for next scan
        
        CMP     CL, STORE_CHKSUM        ;checksums agree?
        JE      NOT_END_TIMER           ;yes, continue scan
        JMP     BAD_CHKSMS
POP_CX_START:
        POP     CX                      ;pop loop counter off stack
T_STAMP_BAD:        
        JMP     EXIT_SCANS1             ;exit scan

BAD_CHKSMS:
        INC     S_1ST_WORD              ;no, bad checksum. inc to next 
        CMP     S_1ST_WORD, 8192        ;time stamp number in storage vars
        JNE     NOT_WORD_WRAP
        MOV     S_1ST_WORD, 0
        INC     S_2ND_WORD
        CMP     S_2ND_WORD, 8192
        JNE     NOT_WORD_WRAP
        MOV     S_2ND_WORD, 0
NOT_WORD_WRAP:
        MOV     ERRORCODE, 1Eh          ;bad scan
        JMP     EXIT_SCANS1

NOT_END_TIMER:
        MOV     CX, S_1ST_WORD
        MOV     STORE_1ST_WORD, CX
        MOV     CX, S_2ND_WORD
        MOV     STORE_2ND_WORD, CX
        CMP     DIGIN_STATUS, 1         ;if digin pending, check for digin
        JNE     NOT_END_TIMER1
        CALL    CHK_FOR_DIGIN           ;check for digital input word
        MOV     ERRORCODE, 0
NOT_END_TIMER1:
        INC     NUMSCANS                ;did a scan inc numscans
        MOV     SI, OFFSET STAMP_ARRAY  ;start at temp stamp array offset
        XOR     CX, CX                  
        MOV     CL, NUM_DATAPOINTS      ;setup loop counter for # of 
        INC     CL                      ;datapoints + stamp scan number
        SHL     CX, 1
        
        INC     NSCANLO
        CMP     NSCANLO, 0
        JNE     T_STUFF_LOOP
        INC     NSCANHI

T_STUFF_LOOP:
        MOV     AX, [SI]                ;take word from stamp_array
        MOV     ES:[DI], AX             ;and put it in scan buffer array
        ADD     SI, 2                   ;next word
        ADD     DI, 2                   ;next word
        
        MOV     AX, END_OFFSET      
        CMP     DI, AX
        JNE     NOT_END_BUFFER1         ;is DI = end?
        MOV     AX, END_SEGMENT
        MOV     BX, ES
        CMP     AX, BX
        JNE     NOT_END_BUFFER1         ;is ES = end?
        MOV     DI, 0                   ;wrap buffer
        MOV     AX, START_SEG
        MOV     ES, AX
        JMP     SAME_TIMER_SEG
NOT_END_BUFFER1:
        
        CMP     DI, 0
        JNZ     SAME_TIMER_SEG          ;data is in the same segment
        MOV     AX, ES                  ;set ES to next segment
        ADD     AX, 1000H
        MOV     ES, AX
SAME_TIMER_SEG:
        LOOP    T_STUFF_LOOP              ;loop until done
        JMP     EXIT_SCANS1
GET_TIMER_SCAN ENDP

;get offset channel in time stamp scanning
;this is a jump procedure rather than a called procedure so that when an error
;occurs, it can jump directly to the correct error label

GET_TIMER_OFFSET:
        
        CALL    CHKBUFFER               ;get one 12 bit word from buffer
        CMP     ERRORCODE, 0            ;check for error
        JE      DO_NEXT_TIMER1
        JMP     POP_CX_START            ;exit scan and wait for next scan
DO_NEXT_TIMER1:        
        
        MOV     LOW14, 0                ;initialize LOW14
        CLC                             ;clear carry
        RCR     CX, 1                   ;carry contains msb of LOW14
        
        MOV     HIGH12, CX              ;save HIGH12
        
        JNC     DC12
        MOV     LOW14, 2000h            ;carry set, set msb in LOW14
DC12:
        CALL    CHKBUFFER               ;get one 12 bit word from buffer
        CMP     ERRORCODE, 0            ;check for error
        JE      T_NOERR1
        JMP     POP_CX_START            ;exit scan and wait for next scan
T_NOERR1:        
        
        ADD     CX, LOW14               ;LOW14 in CX, add msb
        PUSH    SI                      ;store current array location
        CALL    GETCOUNTS               ;returns low word of count in CX,
                                        ;high word in AX
        POP     SI                      ;restore current array location
        
        SUB     CX, OFS_LO              ;subtract offset calibration dword
        SBB     AX, OFS_HI
        CLC                             ;divide difference by 4
        SAR     AX, 1
        RCR     CX, 1
        CLC
        SAR     AX, 1
        RCR     CX, 1                   
        MOV     ZCOR_LO, CX             ;store zero correction dword
        MOV     ZCOR_HI, AX
        MOV     AX, OFS_LO              ;add scan offset dword +
        MOV     BX, OFS_HI              ;zero correction dword
        ADD     ZCOR_LO, AX             ;save scan offset + 
        ADC     ZCOR_HI, BX             ;zero corretion dword
        JMP     RET_GET_TIMER_OFFSET

;T***************   This routine ends time-stamp scanning    ****************

END_TIMER_SCAN  PROC NEAR
        
        AND     STATUS, 3Bh             ;clear bit #2, "stop scan" bit
        MOV     AL, STATUS
        CALL    WRITENIBBLE
        MOV     DIGIN_STATUS, 0
        CALL    DO_END_TIMER
        
EXIT_SCANS1:
        CMP     LESS_FLAG, 1
        JNE     WAS_LESS
        MOV     ERRORCODE, 1FH
        MOV     LESS_FLAG, 0
WAS_LESS:        
        MOV     CURRENT_DI, DI
        MOV     CURRENT_ES, ES
        MOV     AX, NSCANLO
        MOV     ES, L_RESPONSE_SEG      ;segment and offset to response
        MOV     BX, L_RESPONSE_OFS
        MOV     ES:[BX], AX             ;number of scans completed
        MOV     AX, NSCANHI
        MOV     ES:[BX+2], AX
        JMP     T_EXIT_SCANS            ;else exit scan with error
END_TIMER_SCAN ENDP

;T*  This routine restores INT 8 and puts the M250 back to polled or reset  *

DO_END_TIMER PROC NEAR
        
        CMP     START_ISR, 2
        JE      ABORT_TIMER_SCAN
        MOV     START_ISR, 1
        CALL    ADD_REGS
        CMP     AL, 27                  ;ESC KEY PRESSED ?
        JNE     ADD_REGS_OK             
        MOV     ERRORCODE, 0            ;add regs was aborted
ABORT_TIMER_SCAN:        
        CALL    DO_INITIALIZATION       ;initialize M250
        JMP     EXIT_SCANS2             ;exit scan
ADD_REGS_OK:
        OR      STATUS, 8               ;raise host error to signal exit
        MOV     AL, STATUS
        CALL    WRITENIBBLE             ;SIGNAL EXIT TIME STAMP SCAN
        MOV     ERRORCODE, 0            ;initialize ERRORCODE
        MOV     STATUS, 30h             ;initialize STATUS
        MOV     AL, STATUS
        CALL    WRITENIBBLE             ;CLEAR ERROR BIT
        MOV     NEWDATA, 0              ;initialize NEWDATA
        MOV     M250ACK, 10h            ;initialize M250ACK
        MOV     CX, 10                  ;# of chk1's to call before timeout
FLUSH_BUFFER:                           ;flush M250 buffer
        PUSH    CX                      ;store loop counter on stack
        MOV     ERRORCODE, 0            ;clear ERRORCODE
        CALL    CHK1                    ;get one 12 bit word from the M250
        POP     CX                      ;pop loop counter off stack
        CMP     ERRORCODE, 11h          ;timeout error ?
        JE      FLUSHED_BUFFER          ;timeout, exit loop
        LOOP    FLUSH_BUFFER
FLUSHED_BUFFER:
        MOV     ERRORCODE, 0
        MOV     OUTB1, 4                ;get checksum command
        MOV     OUTB2, 0
        MOV     OUTB3, 0                ;get checksum
        CALL    SENDIT
        MOV     CHECKSUM, 0             ;clear checksum
        CALL    CHK1                    ;read M250 checksum
EXIT_SCANS2:        
        CALL    RESTORE_INT8            ;stop ISR
        RET
DO_END_TIMER ENDP

;T******************   pack the time stamp scan number   ********************

PACK_STAMP PROC NEAR                    
        
        MOV     ERRORCODE, 0
        MOV     SI, OFFSET STAMP_ARRAY  ;get offset to storage array
        
        CALL    CHKBUFFER
        CMP     ERRORCODE, 0            ;check for error
        JNE     PACK_T_LO_SKIP          ;exit with error
        MOV     S_1ST_WORD, CX
        MOV     [SI], CX                ;put word in stamp_array
        ADD     SI, 2                   ;increment pointer to next array location
        
        CALL    CHKBUFFER               ;get one 12 bit word from buffer
        CMP     ERRORCODE, 0            ;check for error
        JNE     PACK_T_LO_SKIP          ;exit with error
        MOV     S_2ND_WORD, CX
        MOV     AX, CX
        AND     AX, 7                   ;get low 3 bits
        MOV     AH, AL                  ;and move them to the top 3 bits
        XOR     AL, AL
        SHL     AX, 1                   
        SHL     AX, 1                   
        SHL     AX, 1
        SHL     AX, 1
        SHL     AX, 1
        SUB     SI, 2                   ;index to low 13 bits
        OR      [SI], AX                ;add to low 13 bits  (16 bits)
        ADD     SI, 2                   ;index to high 10 bits
        SHR     CX, 1                   ;shift off low 3 bits
        SHR     CX, 1
        SHR     CX, 1
        MOV     [SI], CX                ;put word in scan buffer array
        ADD     SI, 2                   ;increment pointer to next location
PACK_T_LO_SKIP:
        RET
PACK_STAMP ENDP

;T************   This routine gets one word from the buffer    **************

CHKBUFFER  PROC NEAR
        
        PUSH    SI                      ;store SI
        PUSH    ES
        MOV     AX, @FARDATA?
        MOV     ES, AX                  ;set seg to save data
        MOV     SI, TAIL_POINTER        ;get tail pointer into array
        
        MOV     CX, 0FFFFh
CHK_BUFFER_STAT:
        CMP     SI, HEAD_POINTER        ;check to see if new data
        JNE     NEW_DATA                ;if equal, no new data
        CMP     START_ISR, 0            ;gathering bytes?
        JE      NOT_STOPPED1
        MOV     ERRORCODE, 1Ch          ;signal buffer empty
        JMP     EXIT_CHKBUFFER
NOT_STOPPED1:
        LOOP    CHK_BUFFER_STAT
        MOV     ERRORCODE, 1Ch          ;no data was ready, buffer empty
        JMP     EXIT_CHKBUFFER

NEW_DATA:        
        MOV     CX, ES:[SI]             ;put word in CX
        ADD     SI, 2                   ;increment to next location
        MOV     TAIL_POINTER, SI
        JNZ     NOT_WRAP1
        DEC     WRAP_POINTER            ;decrement WRAP_POINTER if SI = 0
NOT_WRAP1:        
        CMP     CX, 4000h               ;M250 error raised during scanning ?
        JNE     NOT_PIC_ERROR
        MOV     ERRORCODE, 1Eh          ;bad scan
        JMP     EXIT_CHKBUFFER
NOT_PIC_ERROR:        
        MOV     BL, CL
        AND     BL, 0Fh
        ADD     STAMP_CHKSUM, BL
        MOV     BL, CL
        SHR     BL, 1
        SHR     BL, 1
        SHR     BL, 1
        SHR     BL, 1
        ADD     STAMP_CHKSUM, BL
        MOV     BH, CH
        AND     BH, 0Fh
        ADD     STAMP_CHKSUM, BH
EXIT_CHKBUFFER:        
        POP     ES
        POP     SI                      ;restore SI
        RET
CHKBUFFER ENDP

;T*  This routine resets the M250 and sets track/hold, average, and channel *

DO_INITIALIZATION PROC NEAR
        CALL    DO_INIT                 ;re-initialize M250
        CMP     ERRORCODE, 0            ;check for error
        JNE     EXIT_DO_INIT
        
        MOV     OUTB3, 3                ;track/hold command
        MOV     OUTB2, 0
        MOV     AL, HOLD                ;reset current track/hold status
        MOV     OUTB1, AL
        CALL    SENDIT                  ;send command to M250
        CMP     ERRORCODE, 0            ;check for error
        JNE     EXIT_DO_INIT

        MOV     OUTB3, 5                ;remote average command
        MOV     AL, AVERAGE2            ;reset current average
        MOV     OUTB2, AL
        MOV     AL, AVERAGE1
        MOV     OUTB1, AL
        CALL    SENDIT                  ;send command to M250
        CMP     ERRORCODE, 0            ;check for error
        JNE     EXIT_DO_INIT

        MOV     OUTB3, 1                ;set control code
        MOV     AL, CHANNEL             ;reset current channel
        MOV     OUTB2 , AL
        MOV     AL, EXTCODE             ;and current external control code
        MOV     OUTB1, AL
        CALL    SENDIT                  ;send command to M250
EXIT_DO_INIT:        
        RET
DO_INITIALIZATION ENDP

;T********   This routine checks for end of scan by adding up regs    *******
ADD_REGS PROC NEAR
ADD_REGS_AGAIN:
        MOV     AX, 100h                ;check for keystroke
        INT     16H
        JZ      NO_ADD_KEY
        XOR     AX, AX                  ;get key
        INT     16H                     
        CMP     AL, 27                  ;esc key ?
        JNE     NO_ADD_KEY
        MOV     ERRORCODE, 0            ;clear ERRORCODE
        JMP     EXIT_ADD_REGS
NO_ADD_KEY:        
        MOV     BX, 0
        CALL    READ_REGS               ;read and accumulate 4 regs
        CMP     BX, 3C0h                ;all regs Fh ?
        JNE     ADD_REGS_AGAIN
EXIT_ADD_REGS:        
        RET        
ADD_REGS ENDP

;T**********   This routine returns the number of words in AX    ************

GET_NUM_WORDS   PROC NEAR
        MOV     AX, HEAD_POINTER        ;calculate number of bytes in buffer
        MOV     BX, WRAP_POINTER
        MOV     CX, TAIL_POINTER
        MOV     DX, 0
        SUB     AX, CX
        SBB     BX, DX
        RET
GET_NUM_WORDS ENDP

;T************   This routine checks for a digital input word   ************

CHK_FOR_DIGIN   PROC NEAR
        MOV     DX, 60                  ;wait 50 usec
        XOR     CX, CX
        MOV     AX, 8600h
        INT     15h
        CALL    GET_NUM_WORDS
        CMP     AX, 0
        JE      NO_BYTES
        CMP     BX, 0
        JE      NOT_OVERFLOW    
        MOV     ERRORCODE, 1BH
        JMP     EXIT_CHK_DIGIN
NOT_OVERFLOW:        
        PUSH    SI                      ;store SI
        PUSH    ES
        MOV     AX, @FARDATA?
        MOV     ES, AX                  ;set seg to save data
        MOV     SI, TAIL_POINTER        ;get tail pointer into array
        MOV     CX, ES:[SI]
        MOV     FIRST_WORD, CX
        POP     ES
        POP     SI
        JMP     CHK_FIRST
NO_BYTES:        
        MOV     ERRORCODE, 1Dh          ;flag digin pending
        MOV     RESPONSE, 0
        JMP     EXIT_CHK_DIGIN
CHK_FIRST:        
        ;CHECK FOR A BYTE IN THE BUFFER
        CALL    GET_NUM_WORDS
        CMP     AX, 2
        JA      WORDS_IN_BUFFER
        CMP     BX, 0
        JE      NOT_OVERFLOW1
        MOV     ERRORCODE, 1BH
        JMP     EXIT_CHK_DIGIN
NOT_OVERFLOW1:        
        MOV     DL, LOCOUNT             ;setup wait baised on TICKCOUNT
        MOV     DH, HICOUNT
        SHL     DX, 1                   ;MULT BY 4
        SHL     DX, 1
        XOR     CX, CX
        MOV     AX, 8600h               ;wait tickcount * 4 usec
        INT     15h
        CALL    GET_NUM_WORDS
        CMP     AX, 2
        JA      WORDS_IN_BUFFER
        CMP     BX, 0
        JE      NOT_OVERFLOW2
        MOV     ERRORCODE, 1BH
        JMP     EXIT_CHK_DIGIN
NOT_OVERFLOW2:        
        CALL    CHKBUFFER
        MOV     DIGIN_WORD, CX
        MOV     DIGIN_STATUS, 2
        JMP     EXIT_CHK_DIGIN
WORDS_IN_BUFFER:
        PUSH    SI                      ;store SI
        PUSH    ES
        MOV     AX, @FARDATA?
        MOV     ES, AX                  ;set seg to save data
        MOV     SI, TAIL_POINTER        ;get tail pointer into array
        MOV     CX, ES:[SI + 2]
        POP     ES
        POP     SI
        MOV     BX, FIRST_WORD
        MOV     DX, CX                  ;SECOND WORD IN DX
        CALL    CHK_STAMP_NUM           ;EXPECTS 1ST IN BX, 2ND IN DX
        JNC     GOOD_TIMESTAMP          ;IF NO CARRY, GOOD TIME STAMP
        CALL    CHKBUFFER
        MOV     DIGIN_WORD, CX
        MOV     DIGIN_STATUS, 2
        MOV     STAMP_CHKSUM, 0
        JMP     EXIT_CHK_DIGIN
GOOD_TIMESTAMP:
        MOV     ERRORCODE, 1Dh          ;flag digin pending
EXIT_CHK_DIGIN:
        RET
CHK_FOR_DIGIN ENDP

END
