'  Lawson Labs, Inc.
'  15 Phoenixville Pike
'  Malvern, PA  19355
'  800 321-5355 or 406 725-8800                          September 12, 1996
'                                                        Rev E
'  Demonstration Software for
'  Model 250 23-Bit Fast Data Acquisition System

' for Model 250 w/microcontroller Version 6
' To run this code from the environment QuickBASIC must be started with
' these settings:  /ah/l DRIVR250 (date September 12, 1996)
' /ah enables huge arrays and the /L option loads the DRIVR250.QLB library

 DECLARE SUB M250 (SEG COMMAND%, SEG RESPONSE&, SEG ERRORCODE%)

 DECLARE SUB DOTGRAPH ()
 DECLARE SUB ETRAP ()
 DECLARE SUB LOADCALIBRATIONS ()
 DECLARE SUB SELFCAL ()
 COMMON SHARED COMMAND%, RESPONSE&, ERRORCODE%, GAINFACTOR#, HOLD%
 COMMON SHARED CalErr%, FactErr%, CHAN%(), NUMCCS%, DWORDSINDARRAY
 COMMON SHARED DARRAY&(), TINTERVAL#, STNUM&, NUMWORDS%, TSCAN%
 COMMON SHARED VIDCHANS%(), POSFS, NEGFS, PFS, NFS, VCONST, VDIV, HDIV
 COMMON SHARED NEWCAL$, CURRENTCAL$, LETTER%, SERIALNUM%
 COMMON SHARED MINUSGAIN#, PLUSGAIN#, A$
 CLS
 CLEAR , , 4000              ' reserve extra stack space

' All Model 250 Commands are handled by the M250.QLB library
' The library communicates with the Model 250 via a parallel port
' The library processes 16-bit COMMAND% words
' When a response is expected, the library returns a long integer, RESPONSE&
' Commands that pass data to the Model 250 include the data being sent
'    in the least significant 8 bits of the COMMAND word
' Some commands pass data to the library in RESPONSE&
' After any COMMAND is issued, the host should check that the integer
'  ERRORCODE% is equal to zero. A non-zero value is an error code which
'  will indicate the nature of the problem.
'
'  The following commands require 8-Bit data to be added to the command word:
'

 ChannelCodeCMD% = &H100               'select input channel (control code)

'   valid data:   0 to 127
'                 The channel selected is defined by the data as follows:
'                 CHANNEL% * 16 + SUBCHAN%
'                 CHANNEL% is the A/D input channel number:   0 to 7
'                 SUBCHAN% is the external expansion code:    0 to 15
'                          external multiplexers or multiplexed amplifiers are
'                          required in order to use the external control code
'
' Example:  COMMAND% = ChannelCodeCMD% + CHANNEL% * 16 + SUBCHAN%
'           CALL M250(COMMAND%, RESPONSE&, ERRORCODE%)
'           IF ERRORCODE% <> 0 THEN GOTO ERRORTRAP
'           RESPONSE& is not changed by this command
'

 DigitalOutCMD% = &H200                'set digital output

'   valid data:   0 to 255   data word appears at 25-pin connector pins 6-13
'
' Example:  COMMAND% = DigitalOutCMD% + DIGOUTWORD%
'           CALL M250 (COMMAND%, RESPONSE&, ERRORCODE%)
'           IF ERRORCODE% <> 0 THEN GOTO ERRORTRAP
'           RESPONSE& is not changed by this command
'

 TrackHoldCMD% = &H300                'set sampling mode

'   valid data:   0 or 1            1 is track then hold, 0 is track only
'                           track only is usually refered to as "DC mode"
'
' Example:   COMMAND% = TrackHoldCMD% + HOLD%
'            CALL M250(COMMAND%, RESPONSE&, ERRORCODE%)
'            IF ERRORCODE% <> 0 THEN GOTO ERRORTRAP
'            RESPONSE& is not changed by this command
'

 RemoteAvgCMD% = &H500      'set number of samples to average 1,2,4,8,16..128

' valid data   1 to 128      The number to average must be a power of 2
'
' Example:   COMMAND% = RemoteAvgCMD% + AVERAGE%
'            CALL M250(COMMAND%, RESPONSE&, ERRORCODE%)
'            IF ERRORCODE% <> 0 THEN GOTO ERRORTRAP
'            RESPONSE& is not changed by this command
'

 ConvertCMD% = &H600                  'get conversion command

'   valid data:  0 to 127    data is channel to convert, same as ChannelCode
'   The conversion result is passed back as RESPONSE&
'
' Example:  COMMAND% = ConvertCMD% + CHANNEL% * 16 + SUBCHAN%
'           CALL M250 (COMMAND%, RESPONSE&, ERRORCODE%)
'           IF ERRORCODE% <> 0 THEN GOTO ERRORTRAP
'           COUNT& = RESPONSE&

 CalOffsetCMD% = &H7000             'offset calibration

'   valid data:  0 to 127    data is channel to convert, same as ChannelCode

'  on entry, RESPONSE& is set based on the power of 2 of the number of
'  samples to average. The valid range for POWER2% is 0 - 9 (1 to 512 samples)
'  After this command is issued, the selected input channel will read 0 volts
'
' Example:  COMMAND% = CalOffsetCMD% + CHANNEL% * 16 + SUBCHAN%
'           RESPONSE& = 2 ^ POWER2% + (POWER2% * 65536)
'           CALL M250(COMMAND%, RESPONSE&, ERRORCODE%)
'           IF ERRORCODE% <> 0 THEN GOTO ERRORTRAP
'

  CalFullScaleCMD% = &H6000               'full scale calibration command

'  The CalFullScale command is identical to the CalOffset above except that
'  the average count is passed back in RESPONSE&. The actual scaling is
'  done by the CALLing routine.
'
' Example:  COMMAND% = CalFullScaleCMD% + CHANNEL% * 16 + SUBCHAN%
'           RESPONSE& = 2 ^ POWER2% + (POWER2% * 65536)
'           CALL M250(COMMAND%, RESPONSE&, ERRORCODE%)
'           A$ = INKEY$                           'absorb keystroke, if any
'           IF ERRORCODE% <> 0 THEN GOTO ERRORTRAP
'           GAINFACTOR# = 1: GOSUB countsTOvolts  'convert RESPONSE& to volts
'           GAINFACTOR# = 5000 / ANSWER


' The following commands request data from the Model 250, but do not require
' data to be added to the command word.
' After the CALL, RESPONSE& will contain the result.
'

 VersionCMD% = 3                     'request version number from Model 250

'   The version number is passed back as RESPONSE&
'
' Example:  COMMAND% = VersionCMD%
'           CALL M250(COMMAND%, RESPONSE&, ERRORCODE%)
'           IF ERRORCODE% <> 0 THEN GOTO ERRORTRAP
'           VERSION% = RESPONSE&

 DigitalInCMD% = 5                   'digital input

'   The digital input status from the Digital I/O connector pins 18-25
'   is passed back as RESPONSE&
'
' Example:  COMMAND% = DigitalInCMD%
'           CALL M250(COMMAND%, RESPONSE&, ERRORCODE%)
'           IF ERRORCODE% <> 0 THEN GOTO ERRORTRAP
'           LASTDIGINP% = RESPONSE&
'

 ChecksumCMD% = 4                     'verify checksum

' The Checksum command has no effect except on ERRORCODE%
' If ERRORCODE% = 0 then the data received since the last Checksum command
'    are in all probability correct. Note that if checksum commands are
'    not issued frequently, there is a small chance that multiple
'    errors will cancel and no checksum error will be detected.
'
' Example:  COMMAND% = ChecksumCMD%
'           CALL M250(COMMAND%, RESPONSE&, ERRORCODE%)
'           IF ERRORCODE% <> 0 THEN GOTO ERRORTRAP

 DoCalIntScanCMD% = 2   'start interrupt time-stamped scanning w/ calibration
 DoNormalScanCMD% = 6   'start normal scan
 DoIntScanCMD% = 7      'start interrupt time-stamped scanning
 DoFastScanCMD% = 8     'start fast scan

' All DoScan commands are issued in the same way. The only difference
' is the Command word itself.
'
' Example:  COMMAND% = DoNormalScanCMD%
'           RESPONSE& = DSTARTSEG%     'send starting segment for scan array
'           CALL M250(COMMAND%, RESPONSE&, ERRORCODE%)
'           COMPLETESCANS = RESPONSE&             'number of scans completed
'           A$ = INKEY$                           'absorb keystroke, if any
'           IF ERRORCODE% <> 0 THEN GOTO ERRORTRAP
'

 ScanCalCMD% = 11               'Pre-scan calibration command

'  on entry, RESPONSE& is set based on the power of 2 of the number of
'  samples to average. The valid range for POWER2% is 0 - 9 (1 to 512 samples)
'  more averaging means better reproducibility, but longer calibration time
'
' Example:  COMMAND% = ScanCalCMD%
'           RESPONSE& = 2 ^ POWER2% + (POWER2% * 65536)
'           CALL M250(COMMAND%, RESPONSE&, ERRORCODE%)
'           A$ = INKEY$                           'absorb keystroke if any
'           IF ERRORCODE% <> 0 THEN GOTO ERRORTRAP


 PortTestCMD% = 10                   'parallel port test

' The PortTest command resets the Model 250 and checks the function of
' the parallel port. It sends back the read/write delay in RESPONSE&
' The delay will be zero unless errors were encountered.
' You can manually increase the delay for extra timing margins.
' The read/write delay is initialized in the SetPort command.
'
' Example:  COMMAND% = PortTestCMD%
'           CALL M250(COMMAND%, RESPONSE&, ERRORCODE%)
'           IF ERRORCODE% <> 0 THEN GOTO ERRORTRAP
'           RWDELAY% = RESPONSE&
'
                   
' The Set commands pass scanning parameters to the M250 in DATA1&
' In some cases, additional data is sent in the low byte of COMMAND%

 SetRateCMD% = &HC000

' The SetRate command establishes the scanning rate for fast and normal
' scanning. See the SetStampRate command for time-stamped scanning.
' DATA1& is set to reflect scan the interval
' the interchannel delay is added to the command word
'
' Example:  COMMAND% = SetRateCMD% + INTERCHAN%
'           CALL M250(COMMAND%, DATA1&, ERRORCODE%)
'           IF ERRORCODE% <> 0 THEN GOTO ERRORTRAP
'

 SetScanChannelsCMD% = &HD000
                 
' This command specifies the active channels for multi-channel scanning.
' A single 8-Bit byte is written to each of 8 locations.
' The 8 locations are addressed by adding an index to the command word.
' The valid index range is 0 through 7
' The first position is reserved for expansion. It must be set to zero.
' The next 6 positions define the channel selection sequence during the scan.
' DATA1& is derived from the channel code sequence and is put at the
' location for that channel.
' The 8th position is the total number of active data points.
' The valid range for the total is 1 to 96. The 8 positions must all
' be set before the scan configuration is sent to the Model 250.
'
' Example:   FOR INDEX% = 0 to 5
'              COMMAND% = SetScanChannelsCMD% + INDEX% + 1
'              DATA1& = CHAN%(INDEX%)
'              CALL M250(COMMAND%, DATA1&, ERRORCODE%)
'              IF ERRORCODE% <> 0 THEN GOTO ERRORTRAP
'            NEXT INDEX%
'


 SetCalCMD% = &HC00             ' setup long-term calibration process
 StartCalCMD% = &HD00           'start long term self-calibration

' These commands are not normally needed in application code because
' long-term calibration need be done only monthly (or even annually).
' We recommend using DRIVR250 for calibration or using these subroutines
' essentially intact.
'

 SendSetupCMD% = &HE000
                   
' The scanning information previously defined with Set commands
' is sent in a block to the Model 250. Two other parameters are also
' passed when the SendSetup command is issued.
' The SendSetup command must be issued before a DoScan command is sent.
'
' DATA1& = number of scans to complete                (1 to 32766)
' For Time-stamped scanning DATA1& is the buffer size (1 to 32766)
' The trigger type (TRIGGERTYPE%) is added to the command word
' valid types are:          0 = no trigger
'                           1 = positive
'                           2 = negative
'
' Example:  COMMAND% = SendSetupCMD% + TRIGGERTYPE%
'           DATA1& = NUMNORMALSCANS                 'NUMBER OF SCANS
'           CALL M250(COMMAND%, DATA1&, ERRORCODE%)
'           IF ERRORCODE% <> 0 THEN GOTO ERRORTRAP
'

 SetPortCMD% = &HF000                   'set parallel port address

' This command initializes the Read/write delay and the port address
' and resets the Model 250.
' The read/write delay is passed to the library in RESPONSE&
' The port base address (PORTADDR%) is added to the command word
'
' Example:  COMMAND% = SetPortCMD% + PORTADDR%
'           RESPONSE& = RWDELAY%
'           CALL M250(COMMAND%, RESPONSE&, ERRORCODE%)
'           IF ERRORCODE% <> 0 THEN
'             CLS : PRINT "Parallel Port not found at "; HEX$(PORTADDR%);
'             PRINT ". Press any key to continue. "
'             A$ = INPUT$(1)
'           END IF
'

 SetStampRateCMD% = &HE00          'set time-stamp scan rate

' The rate is determined by the interval between scans.
' That interval is held in three bytes STAMPLO%, STAMPMID%, & STAMPHI%
' The low byte is added to the command word. The other bytes are passed in
' DATA1&
'
' Example:  COMMAND% = SetStampRateCMD% + STAMPLO%
'           DATA1& = STAMPMID% + STAMPHI% * &H100
'           CALL M250(COMMAND%, DATA1&, ERRORCODE%)
'           IF ERRORCODE% <> 0 THEN GOTO ERRORTRAP
'

 GetCalFilesCMD% = &HA000   'get long-term calibration and factory cal arrays
                              
' The library needs to load the correct .CAL files before the Model 250
' can perform as specified. Letter% is is an ascii code available for
' housekeeping purposes. The original calibration provided has the letter "A"
' SERIALCODE% is derived from the last 4 digits of the M250's serial number
'
' Example:  COMMAND% = GetCalFilesCMD% + Letter%
'           RESPONSE& = SERIALCODE%
'           CALL M250(COMMAND%, RESPONSE&, ERRORCODE%)
'           IF ERRORCODE% <> 0 THEN GOSUB FileError  'different error handler
'

 SetInterruptCMD% = &H3000      'set interrupt rate and timer frequency
 
'This command sets up the 8254 timer mask for number of interrupts per second
'
' Example:  COMMAND% = SetInterruptCMD%
'           RESPONSE& = TickCount + (DOOLDINT8 * &H10000)
'           CALL M250(COMMAND%, RESPONSE&, ERRORCODE%)
'           IF ERRORCODE% <> 0 THEN GOTO ERRORTRAP

 StartInterruptCMD% = &H3001    'start interrupt

'This command latches the timer service interrupt and sets the 8254 mode 2
'timer for interrupt scanning.
'
' Example:  COMMAND% = StartInterruptCMD%
'           CALL M250(COMMAND%, RESPONSE&, ERRORCODE%)
'           IF ERRORCODE% <> 0 THEN GOTO ERRORTRAP

 RestoreInterruptCMD% = &H3002  'restore interrupt

'This command restores the timer service interrupt and resets the 8254 mode 2
'timer for normal operation
'
' Example:  COMMAND% = RestoreInterruptCMD%
'           CALL M250(COMMAND%, RESPONSE&, ERRORCODE%)
'           IF ERRORCODE% <> 0 THEN GOTO ERRORTRAP

 CancelIntScanCMD% = &H4000     'stop gathering data

'This command ends the gathering of data from the timer interrupt routine
' If CH% = 1 then stop at end of current scan in progress
' If CH% = 2 then cancel scan in progress and exit
'
' Example:  COMMAND% = CancelIntScanCMD% + CH%
'           CALL M250(COMMAND%, RESPONSE&, ERRORCODE%)
'           IF ERRORCODE% = &H10 THEN GOTO ENDTIMERSCAN

 SendDigOutCMD% = &H4003      'request digital output during time-stamped scans

'This command issues digital output commands during time stamped scanning
'if ERRORCODE% = &H1D then there was a digital input pending and no digital
'output command was issued to the 250.
'
' Example:  COMMAND% = SendDigOutCMD%: ERRORCODE% = 0
'           DATA1& = DigitalOutCMD% + (DIGOUTWORD% MOD 256)
'           CALL M250(COMMAND%, DATA1&, ERRORCODE%)
'           IF ERRORCODE% <> &H1D THEN
'              PRINT USING "###"; DIGOUTWORD%;
'           END IF

 ReqDigInCMD% = &H4004        'request digital input during time-stamped scans

'This command requests a digital input during time-stamped scanning.
'if ERRORCODE% = &H1D then there was a digital input pending.
'When a digital input comes in from the 250, the next digital input command
'will have the digital input placed in DATA1&
'
' Example:  COMMAND% = ReqDigInCMD%: ERRORCODE% = 0
'           CALL M250(COMMAND%, DATA1&, ERRORCODE%)
'           IF ERRORCODE% = &H1D THEN
'             LOCATE 24, 28: PRINT "PENDING";
'          ELSE
'            IF ERRORCODE% = 0 THEN
'              LASTDIGINP% = DATA1&
'              LOCATE 24, 28: PRINT USING "###    "; LASTDIGINP%;
'            END IF
'          END IF

 GetNumWordsCMD% = &H5000       'get number of bytes in buffer

'Command to get # of bytes in the buffer, returned in RESPONSE&
'if ERRORCODE% = &H1B  then the buffer overflowed and the data in the buffer
'is lost.
'
' Example:  COMMAND% = GetNumWordsCMD%
'           CALL M250(COMMAND%, RESPONSE&, ERRORCODE%)
'           IF ERRORCODE% = &H1B THEN RETURN ENDTIMERSCAN

 GetIntScanCMD% = &H9000      'get one scan in interrupt time-stamped scanning

' After a DoIntScanCMD% or DoIntCalScanCMD% has been issued,
' Data is retrieved with the GetIntScanCMD% command
' On exit, RESPONSE& is a pointer used to locate data in the scan buffer
'
' Example:  COMMAND% = GetIntScanCMD%
'           CALL M250(COMMAND%, RESPONSE&, ERRORCODE%)
'       error codes     &H1B  buffer overflowed
'                       &H1C  parallel port buffer is empty
'                       &H1E  host missed model 250 scans


 StopIntScanCMD% = &H8000       'stop interrupt time-stamped scanning

' This command ends interrupt scans, resets the timer interrupt, and resets
' the 8254 mode 2 timer.
' On exit,  RESPONSE& is the current scan buffer location
'
' Example:  COMMAND% = StopIntScanCMD%
'           CALL M250(COMMAND%, RESPONSE&, ERRORCODE%)
'           IF ERRORCODE% <> 0 THEN GOTO ERRORTRAP
'           RESPONSE& is the current scan buffer location

'***** misc. arrays
 SDGRP% = 40                        'sample for standard deviation
 REDIM SD#(SDGRP%)                  'standard deviation array
 DIM VIDCHANS%(6)                   'array for displaying video channels
 DIM CHAN%(6)                       'array for channel start and end codes
 GOSUB CHECKFORVGA
 ON ERROR GOTO NOVGA

GOSUB READCONFIG
  NEEDRESET% = -1                    'Flag% for initialization
MENU:
  CLS
  LOCATE 4, 18: PRINT "MODEL 250 23-Bit Data Acquisition System"
  LOCATE 6, 34: PRINT "Main Menu"
  LOCATE 10, 20: PRINT "1) Parallel Port          "; HEX$(PORTADDR%)
  LOCATE 12, 20: PRINT "2) Check Port"
  LOCATE 14, 20: PRINT "3) Select Calibration File"
  LOCATE 16, 20: PRINT "4) Polled Mode Operation";
  LOCATE 18, 20: PRINT "5) Scan Mode Operation"
  LOCATE 20, 18: PRINT "Esc) Exit program";
  A$ = UCASE$(INPUT$(1))
  IF A$ = CHR$(27) THEN GOSUB EXITPROGRAM
  IF A$ = "1" THEN GOSUB CYCLEPORT: GOTO MENU
  GOSUB INITIALIZE
  IF A$ = "2" THEN GOSUB PORTCHECK
  IF A$ = "3" THEN CALL SELFCAL
  IF A$ = "4" THEN GOSUB POLLED
  IF A$ = "5" THEN GOSUB SCANMENU
  GOTO MENU

POLLED:
  CLS
  GOSUB GETVERSION
  GOSUB PRINTCOMMANDMENU
  MAX = -12000: MIN = 12000
NEXTCONVERSION:
  A$ = UCASE$(INKEY$)           'get a keystroke
  IF A$ = CHR$(27) THEN RETURN
  IF A$ <> "" THEN GOSUB DOCOMMAND
  GOSUB GETAVERAGE                      'get an A/D conversion
  VOLTS = ANSWER
  PRINT USING "######### "; RESPONSE&;
  PRINT USING "          #####.###     "; VOLTS    'display voltage
  IF MAX < VOLTS THEN MAX = VOLTS                  'update maximum
  IF MIN > VOLTS THEN MIN = VOLTS                  'update minimum
  GOSUB SD                                      'calculate standard deviation
  IF CSRLIN < 23 THEN GOTO NEXTCONVERSION
  LOCATE 14, 40
  PRINT USING "Maximum     #####.### mv       "; MAX;
  LOCATE 15, 40
  PRINT USING "Minimum     #####.### mv       "; MIN;
  LOCATE 16, 40
  PRINT USING "Delta       #####.### mv       "; ABS(MAX - MIN);
  LOCATE 17, 40
  PRINT USING "Average     #####.### mv       "; RUNNINGAVG;
  LOCATE 19, 40
  PRINT USING "Standard deviation  ####.### mv    "; STDV;
  LOCATE 20, 40
  PRINT USING "Effective Resolution  ##.#   bits"; RMSBITS;
  LOCATE 14, 1
  GOTO NEXTCONVERSION

PRINTCOMMANDMENU:
  LOCATE 1, 20
  PRINT "COMMAND  MENU"; SPACE$(15); "Model 250   Version="; VERSION%
  LOCATE 4, 5: PRINT "C) A/D input Channel    "; CHANNEL%;
  LOCATE 5, 5: PRINT "E) External sub-channel "; SUBCHAN%;
  LOCATE 6, 5: PRINT "A) remote Averaging     "; AVERAGE%; "  ";
  LOCATE 7, 5: PRINT "L) Local averaging      "; LOCALAVG; "      ";
  LOCATE 8, 5: PRINT "T) Track/Hold mode      ";
  IF HOLD% = 1 THEN PRINT "Track/Hold";  ELSE PRINT "DC Mode   ";
  LOCATE 4, 45: PRINT "D) Digital inputs   ";
  IF A$ = "D" THEN
    PRINT LASTDIGINP%; "  ";
    LOCATE 3, 64: PRINT "Decimal  76543210"
    LOCATE 4, 73:
    FOR NN% = 7 TO 0 STEP -1
      IF LASTDIGINP% AND 2 ^ NN% THEN PRINT "1";  ELSE PRINT "0";
    NEXT NN%
  END IF
  LOCATE 5, 45: PRINT "U) digital oUtput   ";
  IF A$ = "U" THEN PRINT DIGOUTWORD%; "  ";
  LOCATE 6, 45: PRINT "K) checKsum         ";
  IF A$ = "K" THEN PRINT " OK";  ELSE PRINT "   ";
  LOCATE 7, 45: PRINT "S) System calibration";
  LOCATE 8, 45: PRINT "O) Offset calibration";
  LOCATE 13, 1: PRINT "  Counts             Milivolts"
  LOCATE 25, 1
  PRINT " Press ESC for menu or space to reset max/min"; SPACE$(35);
  LOCATE 14, 1
RETURN

'**************************************************************************
'**************************   SUBROUTINES   *******************************
'**************************************************************************

CYCLEPORT:
     IF PORTADDR% = &H278 THEN PORTADDR% = &H378: GOTO CYCLEXIT
     IF PORTADDR% = &H378 THEN PORTADDR% = &H3BC: GOTO CYCLEXIT
     PORTADDR% = &H278
CYCLEXIT:
     RESPONSE& = RWDELAY%
     COMMAND% = SetPortCMD% + PORTADDR%
     CALL M250(COMMAND%, RESPONSE&, ERRORCODE%)
     IF ERRORCODE% <> 0 THEN
       CLS : PRINT "Parallel Port not found at "; HEX$(PORTADDR%);
       PRINT ". Press any key to continue. ": A$ = INPUT$(1)
     END IF
     A$ = "": NEEDRESET% = -1
RETURN

DOCOMMAND:
  TEMPCSR = CSRLIN
  LOCATE 25, 1: PRINT SPACE$(80);
  IF A$ = "T" THEN GOSUB TRACKHOLD
  IF A$ = "L" THEN GOSUB SETLOCALAVG
  IF A$ = "A" THEN GOSUB NEXTAVG        'select averaging
  IF A$ = "C" THEN GOSUB CONTROLCODE    'change to selected channel
  IF A$ = "D" THEN GOSUB GETDIGIN       'read the digital inputs
  IF A$ = "E" THEN GOSUB SUBCHANCODE    'set the isolated outputs
  IF A$ = "K" THEN GOSUB GETCHKSUM
  IF A$ = "S" THEN GOSUB SYSCAL         'system calibration
  IF A$ = "O" THEN GOSUB OFFSETCAL      'offset calibration
  IF A$ = "F" THEN GOSUB FULLSCALECAL   'full scale calibration
  IF A$ = "U" THEN GOSUB SETDIGOUT      'set the digital outputs
  IF A$ = " " THEN MAX = -12000: MIN = 12000   'reset max-min
  GOSUB PRINTCOMMANDMENU
  LOCATE TEMPCSR, 1
RETURN

'****  Do a system calibration, restore the original channel
SYSCAL:
 LOCATE 25, 33: PRINT "Calibrating...";
 ORGCHAN% = CHANNEL%
 CHANNEL% = 7: GOSUB OFFSETCAL
 CHANNEL% = 6: GOSUB FULLSCALECAL
 LOCATE 25, 33: PRINT SPACE$(14);
 CHANNEL% = ORGCHAN%: GOSUB SETCCODE
RETURN

OFFSETCAL:
  POWER2% = (9 - INT(LOG(AVERAGE%) / LOG(4)))
  COMMAND% = CalOffsetCMD% + CHANNEL% * 16 + SUBCHAN%
  RESPONSE& = 2 ^ POWER2% + (POWER2% * 65536)
  CALL M250(COMMAND%, RESPONSE&, ERRORCODE%)
  IF ERRORCODE% <> 0 THEN GOTO ERRORTRAP
RETURN

FULLSCALECAL:
  POWER2% = (9 - INT(LOG(AVERAGE%) / LOG(4)))
  COMMAND% = CalFullScaleCMD% + CHANNEL% * 16 + SUBCHAN%
  RESPONSE& = 2 ^ POWER2% + (POWER2% * 65536)
  CALL M250(COMMAND%, RESPONSE&, ERRORCODE%)
  IF ERRORCODE% <> 0 THEN GOTO ERRORTRAP
  GAINFACTOR# = 1: GOSUB countsTOvolts
  IF ANSWER <> 0 THEN GAINFACTOR# = 5000 / ANSWER
RETURN

GETAVERAGE:                'obtain and average LOCALAVG samples
  tot# = 0
  FOR NB% = 1 TO LOCALAVG
    GOSUB GETCONV: GOSUB countsTOvolts
    tot# = tot# + ANSWER
  NEXT NB%
  ANSWER = tot# / LOCALAVG
RETURN

'****              Calculate volts using count in RESPONSE&
countsTOvolts:
  ANSWER = RESPONSE& * .001192093# * GAINFACTOR#
RETURN

'****              Get a conversion
GETCONV:
  COMMAND% = ConvertCMD% + CHANNEL% * 16 + SUBCHAN%
  CALL M250(COMMAND%, RESPONSE&, ERRORCODE%)
  IF ERRORCODE% <> 0 THEN GOTO ERRORTRAP
RETURN

GETVERSION:
  COMMAND% = VersionCMD%
  CALL M250(COMMAND%, RESPONSE&, ERRORCODE%)
  IF ERRORCODE% <> 0 THEN GOTO ERRORTRAP
  VERSION% = RESPONSE&
RETURN

'****             Test checksum from M250
GETCHKSUM:
  COMMAND% = ChecksumCMD%
  CALL M250(COMMAND%, RESPONSE&, ERRORCODE%)
  IF ERRORCODE% <> 0 THEN GOTO ERRORTRAP
RETURN

'****  Get digital inputs from the M250
GETDIGIN:
  COMMAND% = DigitalInCMD%
  CALL M250(COMMAND%, RESPONSE&, ERRORCODE%)
  IF ERRORCODE% <> 0 THEN GOTO ERRORTRAP
  LASTDIGINP% = RESPONSE&
RETURN

'****  Select the averaging factor for m250
NEXTAVG:
  LOCATE 25, 1
  PRINT "Press + increase, - decrease averaging, Enter to continue ";
  PRINT USING "#####              "; AVERAGE%;
  A$ = INKEY$
  IF A$ = "+" THEN AVERAGE% = AVERAGE% * 2
  IF A$ = "-" THEN AVERAGE% = AVERAGE% / 2
  IF AVERAGE% < 1 THEN AVERAGE% = 128
  IF AVERAGE% > 128 THEN AVERAGE% = 1
  IF A$ <> CHR$(13) THEN GOTO NEXTAVG
NEXTAVG1:
  COMMAND% = RemoteAvgCMD% + AVERAGE%
  CALL M250(COMMAND%, RESPONSE&, ERRORCODE%)
  IF ERRORCODE% <> 0 THEN GOTO ERRORTRAP
RETURN

TRACKHOLD:                                  'Toggle Track/hold mode
  HOLD% = (HOLD% + 1) MOD 2
TRACKHOLD1:
  COMMAND% = TrackHoldCMD% + HOLD%          'Set Track/hold mode to HOLD%
  CALL M250(COMMAND%, RESPONSE&, ERRORCODE%)
  IF ERRORCODE% <> 0 THEN GOTO ERRORTRAP
RETURN

'****  Prompt for sub-channel and send it to the M250
SUBCHANCODE:
  LOCATE 25, 1
  INPUT ; "Sub-channel "; SUBCHAN$
  SUBCHAN% = VAL(SUBCHAN$)
  IF SUBCHAN% < 0 OR SUBCHAN% > 15 THEN GOTO SUBCHANCODE
  GOTO SETCCODE

'****  Prompt for channel and send it to the M250
CONTROLCODE:
  LOCATE 25, 1
  INPUT ; "Enter Channel"; CHANNEL$
  CHANNEL% = VAL(CHANNEL$)
  IF CHANNEL% < 0 OR CHANNEL% > 7 THEN GOTO CONTROLCODE

'****   send channel and sub-channel code to the M250
SETCCODE:
  COMMAND% = ChannelCodeCMD% + CHANNEL% * 16 + SUBCHAN%
  CALL M250(COMMAND%, RESPONSE&, ERRORCODE%)
  IF ERRORCODE% <> 0 THEN GOTO ERRORTRAP
  GOSUB CLEARSDCOUNT
RETURN

'****  Prompt for digital output code, send data to the M250
SETDIGOUT:
  LOCATE 25, 1
  INPUT ; "Enter Output Code"; DIGOUTWORD$
  DIGOUTWORD% = VAL(DIGOUTWORD$)
  IF DIGOUTWORD% < 0 OR DIGOUTWORD% > 255 THEN GOTO SETDIGOUT

'****  Send digital output code to the M250
SETDIGOUT1:
  COMMAND% = DigitalOutCMD% + DIGOUTWORD%
  CALL M250(COMMAND%, RESPONSE&, ERRORCODE%)
  IF ERRORCODE% <> 0 THEN GOTO ERRORTRAP
RETURN

SETLOCALAVG:
  LOCATE 25, 1
  INPUT ; "Enter Local Averaging Factor "; LOCALAVG$
  LOCALAVG = VAL(LOCALAVG$)
  IF LOCALAVG < 1 THEN LOCALAVG = 1
RETURN

'****  Write the setup variables to the file DRIVR250.CFG
SAVECONFIG:
  ON ERROR GOTO WRITERR
  CLOSE #2
  OPEN "DRIVR250.CFG" FOR OUTPUT AS #2
  WRITE #2, PORTADDR%, CHANNEL%, SUBCHAN%, HOLD%, AVGTIME, SERIALNUM%, LETTER%
  WRITE #2, GAINFACTOR#, LOCALAVG, AVERAGE%, TSCAN%, INTERCHAN%
  WRITE #2, FASTCHAN%, FASTSUB%, FASTRATE#, FASTDELAY, NUMFASTSCANS
  WRITE #2, NORMALRATE#, NORMALDELAY, NUMNORMALSCANS
  WRITE #2, STAMPINTVL#, STAMPBUFFER
  WRITE #2, CALINTVL#, CALBUFFER
  WRITE #2, CHAN%(0), CHAN%(1), CHAN%(2), CHAN%(3), CHAN%(4), CHAN%(5)
  WRITE #2, POSFS, NEGFS, PFS, NFS, VCONST, VDIV, HDIV
  CLOSE #2: ON ERROR GOTO 0
RETURN

WRITERR:
   CLOSE #2
   CLS : PRINT "Unable to save DRIVR250.CFG file, press any key to continue"
   A$ = INPUT$(1)
RESUME 3000
3000 ON ERROR GOTO 0: RETURN

'****  Read the setup variables from the DRIVR250.CFG file
READCONFIG:
  ON ERROR GOTO READERR
  OPEN "DRIVR250.CFG" FOR INPUT AS #2
  INPUT #2, PORTADDR%, CHANNEL%, SUBCHAN%, HOLD%, AVGTIME, SERIALNUM%, LETTER%
  INPUT #2, GAINFACTOR#, LOCALAVG, AVERAGE%, TSCAN%, INTERCHAN%
  INPUT #2, FASTCHAN%, FASTSUB%, FASTRATE#, FASTDELAY, NUMFASTSCANS
  INPUT #2, NORMALRATE#, NORMALDELAY, NUMNORMALSCANS
  INPUT #2, STAMPINTVL#, STAMPBUFFER
  INPUT #2, CALINTVL#, CALBUFFER
  INPUT #2, CHAN%(0), CHAN%(1), CHAN%(2), CHAN%(3), CHAN%(4), CHAN%(5)
  INPUT #2, POSFS, NEGFS, PFS, NFS, VCONST, VDIV, HDIV
  CLOSE #2
4000
  ON ERROR GOTO 0
  IF LOCALAVG < 1 THEN LOCALAVG = 1
'ST = Sample Time  ICD = Inter-Channel Delay  ISDM = Inter-Scan Delay Minimum
  ST = .0002076: ICD = .000017: ISDM = .0000212
  INTERCHANDELAY% = INTERCHAN% * 17
  SINTVL# = STAMPINTVL#: CINTVL# = CALINTVL#
  FASTGAINFLAG% = 0: NORMALGAINFLAG% = 0
  STAMPGAINFLAG% = 0: CALGAINFLAG% = 0
  S$ = RIGHT$(STR$(SERIALNUM%), LEN(STR$(SERIALNUM%)) - 1)
  TEMPLEN = LEN(S$): FOR N = 1 TO (4 - TEMPLEN): S$ = "0" + S$: NEXT N
  NEWCAL$ = "250" + CHR$(LETTER%) + S$ + ".CAL"
  CALL LOADCALIBRATIONS
RETURN

READERR:   '    default values if no .CFG file found
  CLOSE #2
  CLS : PRINT "Unable to read DRIVR250.CFG file, press any key to continue"
  A$ = INPUT$(1)
  CHANNEL% = 7: AVERAGE% = 1: INTERCHAN% = 1
  PORTADDR% = &H3BC              'default parallel port address
  AVGTIME = 0                    'NO AVERAGING INTERVAL
  RWDELAY% = 0                   'no read/write delay
  FASTRATE# = 5470.5: NUMFASTSCANS = 1000
  NORMALRATE# = 100: NUMNORMALSCANS = 1000
  STAMPINTVL# = .01: STAMPBUFFER = 1000
  CALINTVL# = .01: CALBUFFER = 1000
  GAINFACTOR# = 1
  POSFS = 5: NEGFS = -5: PFS = POSFS * 1000: NFS = NEGFS * 1000
  VCONST = (399000 / (NFS - PFS)): VDIV = 1: HDIV = 0
RESUME 4000

EXITPROGRAM:                                             'save or just exit
  CLS : LOCATE 12, 20: PRINT "SAVE CONFIGURATION DATA ? (Y/N) "
  A$ = UCASE$(INPUT$(1))
  IF A$ = "Y" THEN GOSUB SAVECONFIG: GOTO EXITPROGRAM1
  IF A$ <> "N" THEN GOTO EXITPROGRAM
EXITPROGRAM1:
  CLS : LOCATE 12, 20: PRINT "EXIT PROGRAM ? (Y/N) "
  A$ = UCASE$(INPUT$(1))
  IF A$ = "N" THEN RETURN
  IF A$ <> "Y" THEN GOTO EXITPROGRAM1
  CLEAR : CLOSE
END

'****  Send reset command to the M250
INITIALIZE:
  IF NEEDRESET% OR A$ = "2" THEN
    CLS : LOCATE 10, 32: PRINT "INITIALIZING..."
    RESPONSE& = RWDELAY%
    COMMAND% = SetPortCMD% + PORTADDR%
    CALL M250(COMMAND%, RESPONSE&, ERRORCODE%)
    IF ERRORCODE% <> 0 THEN GOTO ERRORTRAP
    GOSUB TRACKHOLD1
    GOSUB NEXTAVG1
    GOSUB SETCCODE
    NEEDRESET% = 0
    CLS
    IF A$ <> "2" THEN GOSUB SYSCAL     'no syscal if checking port
  END IF
RETURN

SD:                      'standard deviation and RMS resolution calculation
  SD#(SDCOUNT%) = ANSWER: SDTOT# = SDTOT# + ANSWER
  IF SDCOUNT% <> SDGRP% THEN SDCOUNT% = SDCOUNT% + 1: GOTO EXITSD
  STDV = 0: SDTOT# = SDTOT# / SDGRP%
  FOR NNN% = 1 TO SDGRP%: STDV = STDV + ((SDTOT# - SD#(NNN%)) ^ 2): NEXT NNN%
  STDV = SQR(STDV / SDGRP%)
  STDVRMS = STDV / .001192093#
  I% = 0
  TEMPSTDV = STDVRMS
CALCNEXT:
  IF TEMPSTDV >= 1 THEN I% = I% + 1: TEMPSTDV = TEMPSTDV / 2: GOTO CALCNEXT
  RMSBITS = 23 - I%
  IF (TEMPSTDV < .75) AND (I% <> 0) THEN RMSBITS = RMSBITS + .5
  RUNNINGAVG = SDTOT#
CLEARSDCOUNT:
  SDTOT# = 0: SDCOUNT% = 1
EXITSD:
RETURN

PORTCHECK:
  CLS : PRINT "Initial port settling delay = "; RWDELAY%
  LOCATE 3, 1: PRINT "This should take under a minute. "
  PRINT "Press any key to cut test short."
  COMMAND% = PortTestCMD%
  CALL M250(COMMAND%, RESPONSE&, ERRORCODE%)
  A$ = INKEY$
  IF ERRORCODE% <> 0 THEN GOTO ERRORTRAP
  RWDELAY% = RESPONSE&
  IF RESPONSE& = 0 THEN PRINT "             PORT OK"
  IF RESPONSE& = 100 THEN
     PRINT "Model 250 not found. Press space to continue";
     RWDELAY% = 0
     GOTO EXITCHECK
  END IF
  PRINT "Current port settling delay = "; RWDELAY%
  PRINT "Type new delay or Enter only to leave unchanged";
EXITCHECK:
  INPUT A$
  IF A$ <> "" THEN RWDELAY% = VAL(A$)
  NEEDRESET% = -1
RETURN

'***********************************************************************
'****  Model 250 error handler
ERRORTRAP:
  CALL ETRAP                            'parse and display error message
  NEEDRESET% = -1
  COMMAND% = RestoreInterruptCMD%               'restore interrupts
  CALL M250(COMMAND%, RESPONSE&, ERRORCODE%)
  IF ERRORCODE% <> 0 THEN GOTO ERRORTRAP
  GOTO MENU

'!****************************************************************************
'*************************  SCANNING ROUTINES  *******************************
'*****************************************************************************

'**** setup scan array and starting segment address
SETUPDARRAY:
   REDIM DARRAY&(DWORDSINDARRAY)
   DSTARTSEG% = VARSEG(DARRAY&(0))
RETURN

'****   get count from scan array and convert to voltage
GETNEXTDATAPOINT:
  RESPONSE& = DARRAY&(II - 1)
  GOSUB countsTOvolts
RETURN

'****  Change scanning type
SCANMENU:
  TRIGGER$(0) = "None         "
  TRIGGER$(1) = "Positive edge"
  TRIGGER$(2) = "Negative edge"
  SCANTYPE$(0) = "  Fast Single-Channel"
  SCANTYPE$(1) = "  Normal"
  SCANTYPE$(2) = "  Time-stamped"
  SCANTYPE$(3) = "  Time-stamped w/cal"
SMENU:
  IF TSCAN% = 0 THEN GOSUB FASTSCAN
  IF TSCAN% = 1 THEN GOSUB NORMALSCAN
  IF TSCAN% = 2 THEN GOSUB STAMPSCAN
  IF TSCAN% = 3 THEN GOSUB CALSCAN      'do time-stamp scan with calibration
  IF A$ = "T" THEN TSCAN% = (TSCAN% + 1) MOD 4  'select next scan type
  IF A$ <> CHR$(27) THEN GOTO SMENU
  A$ = ""
RETURN

'!************************************************************************
'****************   start of fast scan routines    ***********************
'*************************************************************************

FASTSCAN:
  CLS
  NUMCCS% = 1                                  'number of active channels
  REDIM DARRAY&(10)                                'deallocate DARRAY&
  IF FRE(-1) < &H20200 THEN MAXLEN = FRE(-1) \ 4 - 512 ELSE MAXLEN = 32766
  MAXSCANS = MAXLEN                      'maximum allowable number of scans
  LOCATE 1, 29: PRINT "SCANNING CONFIGURATION"
  LOCATE 7, 26: PRINT "Channel     Sub-Channel"
  LOCATE 10, 30: PRINT USING "#            ##"; FASTCHAN%; FASTSUB%;
  LOCATE 20, 17
  PRINT USING "7) Scanning rate                   ######.# Hz"; FASTRATE#;
  LOCATE 21, 17
  PRINT USING "8) Number of scans to perform    ########"; NUMFASTSCANS;
  LOCATE 22, 17
  PRINT "9) Trigger Type                         "; TRIGGER$(TRIGGERTYPE%);
  LOCATE 24, 13: PRINT "Enter) begin scanning";
  LOCATE 24, 40: PRINT USING "Total points per scan ##"; NUMCCS%;
  LOCATE 25, 40: PRINT "T) Toggle scan type:"; SCANTYPE$(TSCAN%);
  LOCATE 25, 13: PRINT "Esc) Return to Main Menu";
  A$ = UCASE$(INPUT$(1)): VALA% = VAL(A$)
  IF (VALA% = 0) AND (A$ <> "0") THEN VALA% = 15  'flag invalid keystroke
  IF (VALA% >= 0) AND (VALA% <= 5) THEN GOSUB GETFASTCHANNEL
  IF A$ = "T" THEN GOTO EXITFAST
  IF A$ = "7" THEN GOSUB GETFASTRATE
  IF A$ = "8" THEN GOSUB SETNUMFASTSCANS
  IF A$ = "9" THEN TRIGGERTYPE% = (TRIGGERTYPE% + 1) MOD 3
  IF A$ = CHR$(13) THEN GOSUB DOFASTSCAN
  IF A$ = CHR$(27) THEN GOTO EXITFAST
  GOTO FASTSCAN
EXITFAST:
RETURN

'****   perform fast scanning
DOFASTSCAN:
  CLS : COMPLETESCANS = 0
  LOCATE 6, 20: PRINT "Begin Scanning Menu"
  LOCATE 10, 20: PRINT "1) Scan Calibration"
  LOCATE 12, 20: PRINT "2) Begin Scanning"
  LOCATE 14, 18: PRINT "Esc) Return to Scan Configuration Menu";

  A$ = UCASE$(INPUT$(1))
  CLS
  IF A$ = CHR$(27) THEN GOTO EXITDOFAST
  IF VAL(A$) < 1 OR VAL(A$) > 2 THEN GOTO DOFASTSCAN
  IF FASTRATE# = 0 THEN GOSUB GETFASTRATE
  GOSUB SETFASTRATE
  IF A$ = "1" THEN
    CLS
    GOSUB GETFASTGAINCORRECTION
    IF FANS# = 0 THEN GOTO EXITDOFAST
    FASTGAINFLAG% = 1        'successful gain correction for fast scans
    GOTO DOFASTSCAN
  END IF

'if there was no gain correction scan, do gain correcton scan
  IF FASTGAINFLAG% = 0 THEN
    CLS
    PRINT "No calibration scan found, performing calibration scan"
    GOSUB GETFASTGAINCORRECTION
    IF FANS# = 0 THEN GOTO EXITDOFAST
    FASTGAINFLAG% = 1        'successful gain correction for fast scans
    LOCATE 1, 1: PRINT SPACE$(79)
    LOCATE 10, 1: PRINT "Press space to start scanning, ESC to quit";
    A$ = UCASE$(INPUT$(1))
    IF A$ = CHR$(27) THEN GOTO EXITDOFAST
  END IF

  GOSUB SETFASTCHANNEL                  'set fast scan channel
'send scan configuration block, type of trigger, and number of scans
  COMMAND% = SendSetupCMD% + TRIGGERTYPE%
  DATA1& = NUMFASTSCANS                 'number of scans
  CALL M250(COMMAND%, DATA1&, ERRORCODE%)
  IF ERRORCODE% <> 0 THEN GOTO ERRORTRAP
 
  LOCATE 25, 10: PRINT "Press any key to cancel";
  IF A$ = CHR$(27) THEN GOTO EXITDOFAST
  LOCATE 10, 1: PRINT "Scanning started at "; TIME$; "  on  "; DATE$
  PRINT "Data collection will take ";
  PRINT USING "#####.##### seconds"; (NUMFASTSCANS / FASTRATE#)
  DWORDSINDARRAY = NUMFASTSCANS                   'allocate scan array
  GOSUB SETUPDARRAY
  IF A$ = CHR$(27) THEN RETURN
  COMMAND% = DoFastScanCMD%
  RESPONSE& = DSTARTSEG%                                'scan array segment
  CALL M250(COMMAND%, RESPONSE&, ERRORCODE%)            'start the scan
  IF INKEY$ = CHR$(27) THEN GOTO EXITDOFAST
  PRINT "Scanning complete"
  COMPLETESCANS = RESPONSE&
  IF ERRORCODE% <> 0 THEN GOTO ERRORTRAP
  IF COMPLETESCANS = 0 THEN GOTO EXITDOFAST     'if no scans complete, exit
  GOSUB DISPFAST                                'display scans
EXITDOFAST:
  IF A$ <> CHR$(27) THEN GOTO DOFASTSCAN
  A$ = ""
RETURN

'****   display fast scans
DISPFAST:
  CLS
  SCANMAX = -12000: SCANMIN = 12000
  LOCATE 25, 10: PRINT "Press space to continue, R to re-run, ESC to quit";
  LOCATE 1, 1
  FOR II = 1 TO COMPLETESCANS
    GOSUB GETNEXTDATAPOINT
    ANSWER = ANSWER * FASTGAIN#
    PRINT USING "Scan ######    "; II;
    PRINT USING "#####.### mv"; ANSWER
    IF II MOD 22 = 0 THEN                     ' pause at bottom of page
      A$ = UCASE$(INPUT$(1))
      IF A$ = "R" THEN GOTO DISPFAST
      IF A$ = CHR$(27) THEN GOTO EXITDISPFAST
    END IF
  NEXT II
FASTSCANSDONE:
  LOCATE 25, 10: PRINT "Press R to re-run, ESC to quit                     ";
  A$ = UCASE$(INKEY$)
  IF A$ = "R" THEN GOTO DISPFAST
  IF A$ <> CHR$(27) THEN GOTO FASTSCANSDONE
EXITDISPFAST:
RETURN

'****   get the fast scanning rate
GETFASTRATE:
  CLS
  LOCATE 3, 15: PRINT "220.6  -  5470.5"
  LOCATE 1, 1: INPUT "Enter rate in Hertz"; FASTRATE$
  FASTRATE# = VAL(FASTRATE$)
  IF FASTRATE# = 0 THEN FASTRATE# = 5470.5
  IF FASTRATE# > 5470.5 THEN FASTRATE# = 5470.5
  IF FASTRATE# < 220.6 THEN FASTRATE# = 220.6
  FASTDELAY = INT(((1 / FASTRATE# - .0001828) * 1000000) + .5)
  IF FASTDELAY < 0 THEN FASTDELAY = 0: FASTRATE# = 5470.5
  IF FASTDELAY > 4351 THEN FASTDELAY = 4351: FASTRATE# = 220.6
  DATALO% = INT(FASTDELAY / 17)
  SINGLETICK% = FASTDELAY - (DATALO% * 17)
  FASTRATE# = 1 / ((FASTDELAY / 1000000) + .0001828)
RETURN

'****   set the fast scanning rate
SETFASTRATE:
  COMMAND% = SetRateCMD% + DATALO%
  DATA1& = SINGLETICK% * &H100
  CALL M250(COMMAND%, DATA1&, ERRORCODE%)
  IF ERRORCODE% <> 0 THEN GOTO ERRORTRAP
RETURN

'****   get fast channel and sub channel
GETFASTCHANNEL:
  CLS : LOCATE 23, 4
  LOCATE 12, 1: PRINT "Channel number:    "; VALA%
  PRINT : INPUT "Enter sub channel 0 - 15"; FASTSUB$
  FASTSUB% = VAL(FASTSUB$)
  IF (FASTSUB% < 0) OR (FASTSUB% > 15) THEN GOTO GETFASTCHANNEL
  FASTCHAN% = VALA%
RETURN

'****   set fast scanning channel and sub channel
SETFASTCHANNEL:
  COMMAND% = ChannelCodeCMD% + FASTCHAN% * 16 + FASTSUB%
  CALL M250(COMMAND%, RESPONSE&, ERRORCODE%)
  IF ERRORCODE% <> 0 THEN GOTO ERRORTRAP
  GOSUB CLEARSDCOUNT
RETURN

'****   get the gain factor for fast scanning
GETFASTGAINCORRECTION:
  LOCATE 3, 1
  PRINT USING "Calibration will take    #####.### seconds"; (1000 / FASTRATE#)
  LOCATE 10, 10: PRINT "Press any key to cancel";
  SAVEFCHAN% = FASTCHAN%: SAVEFSUB% = FASTSUB%
  COMMAND% = SendSetupCMD%
  DATA1& = 1000
  CALL M250(COMMAND%, DATA1&, ERRORCODE%)  'DATA1& = number of scans
  IF ERRORCODE% <> 0 THEN GOTO ERRORTRAP
  DWORDSINDARRAY = 1000              'allocate scan array
  GOSUB SETUPDARRAY
  GOSUB GETFASTFULL
  FASTGAIN# = 5000 / FFULL#
getfastgaincorrectionexit:
  FASTCHAN% = SAVEFCHAN%: FASTSUB% = SAVEFSUB%
  LOCATE 3, 1: PRINT SPACE$(50)
RETURN

'****   perform a fast scan on channel 6
GETFASTFULL:
  FASTCHAN% = 6: FASTSUB% = 0
  GOSUB SETFASTCHANNEL
  COMMAND% = DoFastScanCMD%: RESPONSE& = DSTARTSEG%    'segment for scan array
  CALL M250(COMMAND%, RESPONSE&, ERRORCODE%)  'start the scan
  COMPLETESCANS = RESPONSE&
  IF ERRORCODE% <> 0 THEN GOTO ERRORTRAP
  II = 0: FANS# = 0 'initialize vars
  IF INKEY$ <> "" THEN GOTO getfastgaincorrectionexit
  WHILE II <> COMPLETESCANS
    II = II + 1: GOSUB GETNEXTDATAPOINT: FANS# = FANS# + ANSWER
  WEND
  FFULL# = (FANS# / COMPLETESCANS)
RETURN

'****   get number of scans for fast scanning
SETNUMFASTSCANS:
  CLS
  LOCATE 3, 1: PRINT "Current scans: "; NUMFASTSCANS
  LOCATE 5, 1: PRINT "Number of scans to perform (1 to"; MAXSCANS; "), "
  INPUT "or enter to leave unchanged ", A$
  A = VAL(A$)
  IF A > MAXSCANS THEN A = MAXSCANS
  IF A < 1 THEN A = 1
  IF A$ <> "" THEN NUMFASTSCANS = A
  IF NUMFASTSCANS > MAXSCANS THEN NUMFASTSCANS = MAXSCANS
RETURN

'!************************************************************************
'***************   start of normal scan routines    **********************
'*************************************************************************

NORMALSCAN:
  CLS
  LOCATE 1, 29: PRINT "SCANNING CONFIGURATION"
  LOCATE 3, 14: PRINT "Channel      External control code"
  LOCATE 3, 54: PRINT "Number of data"
  LOCATE 4, 29: PRINT "from           to        points this channel"
  NUMCCS% = 0                   'calculate total number of active channels
  FOR N% = 0 TO 5
    CC%(N%) = CHAN%(N%) MOD 16 - CHAN%(N%) \ 16 + 1
    IF CC%(N%) = -7 THEN CC%(N%) = 0
    NUMCCS% = NUMCCS% + CC%(N%)                      '# of datapoints
    LOCATE 6 + N% * 2, 17: PRINT USING "#)"; N%;
    PRINT , CHAN%(N%) \ 16, CHAN%(N%) MOD 16, CC%(N%); ""
  NEXT N%
 
  REDIM DARRAY&(10)                           'deallocate DARRAY&
  IF FRE(-1) < &H20200 THEN MAXLEN = FRE(-1) \ 4 - 512 ELSE MAXLEN = 32766
  IF NUMCCS% > 0 THEN
    MAXSCANS = INT(MAXLEN / NUMCCS%)          'calc max number of scans
  END IF
 
  LOCATE 19, 17
  PRINT "6) Interchannel delay               ";
  PRINT USING " ####   microseconds"; INTERCHANDELAY%
  LOCATE 20, 17
  PRINT USING "7) Scanning rate                   ######.# Hz"; NORMALRATE#;
  LOCATE 21, 17
  PRINT USING "8) Number of scans to perform    ########"; NUMNORMALSCANS;
  LOCATE 22, 17
  PRINT "9) Trigger Type                         "; TRIGGER$(TRIGGERTYPE%);
  LOCATE 24, 13: PRINT "Enter) begin scanning";
  LOCATE 24, 40: PRINT USING "Total points per scan ##"; NUMCCS%;
  LOCATE 25, 40: PRINT "T) Toggle scan type:"; SCANTYPE$(TSCAN%);
  LOCATE 25, 13: PRINT "Esc) Return to Main Menu";
  A$ = UCASE$(INPUT$(1)): VALA% = VAL(A$)
  IF (VALA% = 0) AND (A$ <> "0") THEN VALA% = 15  'flag invalid keystroke
  IF (VALA% >= 0) AND (VALA% <= 5) THEN GOSUB GETCHANNEL
  IF A$ = "T" THEN GOTO EXITNORMAL              'cycle scan type
  IF A$ = "6" THEN GOSUB GETINTERDELAY
  IF A$ = "7" THEN GOSUB GETNORMALRATE
  IF A$ = "8" THEN GOSUB SETNUMNORMALSCANS
  IF A$ = "9" THEN TRIGGERTYPE% = (TRIGGERTYPE% + 1) MOD 3
  IF A$ = CHR$(13) AND NUMCCS% > 0 THEN GOSUB DONORMALSCAN
  IF A$ <> CHR$(27) THEN GOTO NORMALSCAN
EXITNORMAL:
RETURN

'****   perform a normal scan
DONORMALSCAN:
  CLS : COMPLETESCANS = 0
  LOCATE 6, 20: PRINT "Begin Scanning Menu"
  LOCATE 10, 20: PRINT "1) Scan Calibration"
  LOCATE 12, 20: PRINT "2) Begin Scanning"
  LOCATE 14, 18: PRINT "Esc) Return to Scan Configuration Menu";

  A$ = UCASE$(INPUT$(1))
  CLS
  IF A$ = CHR$(27) THEN GOTO EXITDOSTAMP
  IF VAL(A$) < 1 OR VAL(A$) > 2 THEN GOTO DONORMALSCAN
  IF NORMALRATE# = 0 THEN GOSUB GETNORMALRATE
  GOSUB SETNORMALRATE
  IF NUMNORMALSCANS > MAXSCANS THEN GOSUB SETNUMNORMALSCANS
  IF A$ = "1" THEN
    CLS
    GOSUB GETGAINCORRECTION     'get correction factor for scan
    IF FANS# = 0 THEN GOTO EXITDONORMAL
    NORMALGAINFLAG% = 1        'successful gain correction for normal scans
    GOTO DONORMALSCAN
  END IF

'if there was no gain correction scan, do gain correcton scan
  IF NORMALGAINFLAG% = 0 THEN
    CLS
    PRINT "No calibration scan found, performing calibration scan"
    GOSUB GETGAINCORRECTION     'get correction factor for scan
    IF FANS# = 0 THEN GOTO EXITDONORMAL
    NORMALGAINFLAG% = 1        'successful gain correction for normal scans
    LOCATE 1, 1: PRINT SPACE$(79)
    LOCATE 10, 1: PRINT "Press space to start scanning, ESC to quit";
    A$ = UCASE$(INPUT$(1))
    IF A$ = CHR$(27) THEN GOTO EXITDONORMAL
    GOSUB SETNORMALRATE
  END IF

  GOSUB SETSCANCHANNELS           'setup scan control codes

'send scan configuration block, type of trigger, and number of scans
  COMMAND% = SendSetupCMD% + TRIGGERTYPE%
  DATA1& = NUMNORMALSCANS
  CALL M250(COMMAND%, DATA1&, ERRORCODE%)
  IF ERRORCODE% <> 0 THEN GOTO ERRORTRAP

  LOCATE 25, 10: PRINT "Press any key to cancel";
  LOCATE 10, 1: PRINT "Scanning started at "; TIME$; "  on  "; DATE$
  PRINT "Data collection will take ";
  PRINT USING "#####.### seconds"; (NUMNORMALSCANS / NORMALRATE#)
  DWORDSINDARRAY = NUMCCS% * NUMNORMALSCANS          'allocate scan array
  GOSUB SETUPDARRAY
  COMMAND% = DoNormalScanCMD%
  RESPONSE& = DSTARTSEG%                'send starting segment for scan array
  CALL M250(COMMAND%, RESPONSE&, ERRORCODE%)
  IF INKEY$ = CHR$(27) THEN GOTO EXITDONORMAL
  COMPLETESCANS = RESPONSE&
  IF ERRORCODE% <> 0 THEN GOTO ERRORTRAP
  IF COMPLETESCANS = 0 THEN GOTO EXITDONORMAL             'if no scans, exit
  PRINT "Scanning complete";
  GOSUB DISPNORMAL
EXITDONORMAL:
  GOTO DONORMALSCAN
RETURN

'****   display normal scans
DISPNORMAL:
  CLS : NUMSCAN = 1: II = 0
  PRINT " CC     chan 0       chan 1       chan 2      ";
  PRINT " chan 3       chan 4       chan 5": PRINT
  FOR I% = 0 TO 15: PRINT I%: NEXT
NORMALSCANLOOP:
  FOR N% = 0 TO 5                   'get datapoints for each channel
    STARTCC% = CHAN%(N%) \ 16: ENDCC% = CC%(N%) + STARTCC% - 1
    FOR CC% = STARTCC% TO ENDCC%
      II = II + 1: GOSUB CONVDISP
    NEXT CC%
  NEXT N%
  LOCATE 25, 1: PRINT USING "scan ######  of ######"; NUMSCAN; COMPLETESCANS;
  NUMSCAN = NUMSCAN + 1
  IF NUMSCAN > COMPLETESCANS THEN GOTO NORMALSCANSDONE
  PRINT "  Press space to continue, R to re-run, ESC to quit";
  A$ = UCASE$(INPUT$(1))
  IF A$ = "R" THEN GOTO DISPNORMAL
  IF A$ = CHR$(27) THEN GOTO EXITDISPNORMAL
  GOTO NORMALSCANLOOP
NORMALSCANSDONE:
  LOCATE 25, 23: PRINT "  Press R to re-run, ESC to quit"; SPACE$(25);
  A$ = UCASE$(INKEY$)
  IF A$ = "R" THEN GOTO DISPNORMAL
  IF A$ <> CHR$(27) THEN GOTO NORMALSCANSDONE
EXITDISPNORMAL:
RETURN

'****   display voltage from scan array indexed with II
CONVDISP:
  GOSUB GETNEXTDATAPOINT
  IF HOLD% = 1 THEN      'secondary gain correction for Track/hold mode only
    IF ANSWER < 0 THEN
      ANSWER = ANSWER * MINUSGAIN#
    ELSE
      ANSWER = ANSWER * PLUSGAIN#
    END IF
  END IF
  LOCATE (CC% + 3), (N% * 13) + 7
  PRINT USING "#####.###"; ANSWER;
RETURN

'****   Get scan rate for normal scans
GETNORMALRATE:
  CLS
  GOSUB NORMALMAXMINRATE
  LOCATE 5, 15: PRINT USING "###.#  -  ####.#"; MINRATE#; MAXRATE#
  LOCATE 7, 1: INPUT "Enter rate in Hertz"; NORMALRATE$
  NORMALRATE# = VAL(NORMALRATE$)
  IF NORMALRATE# = 0 THEN NORMALRATE# = 5000
  GOSUB CALCNORMALRATE                          'calc normal rate
  IF NORMALDELAY < 0 THEN NORMALDELAY = 0: NORMALRATE# = MAXRATE#
  IF NORMALDELAY > 1109776 THEN NORMALDELAY = 1109776: NORMALRATE# = MINRATE#
RETURN

'****   Set scan rate for normal scans
SETNORMALRATE:
  GOSUB NORMALMAXMINRATE
  GOSUB CALCNORMALRATE                          'calc the normal rate
  IF NORMALRATE# < MINRATE# OR NORMALRATE# > MAXRATE# THEN
    CLS : PRINT USING "   ####.#         "; NORMALRATE#;
    PRINT USING "MIN ###.#    MAX ####.#"; MINRATE#; MAXRATE#
    LOCATE 4, 1: PRINT "Scan rate out of limits.  Press any key to continue"
    A$ = INPUT$(1): A$ = "": GOSUB GETNORMALRATE
  END IF
  COMMAND% = SetRateCMD% + INTERCHAN%
  IF NORMALDELAY = -1 THEN NORMALDELAY = 0
  DATAHI% = INT(NORMALDELAY / 4335)                     'get hi delay
  IF DATAHI% > 255 THEN DATAHI% = 255
  DATALO% = INT((NORMALDELAY - DATAHI% * 4335&) / 17)   'get lo delay
  SINGLETICK% = NORMALDELAY - DATAHI% * 4335& - DATALO% * 17  'get singletick
  DATA1& = DATALO% + SINGLETICK% * &H100 + DATAHI% * &H10000 'send delays
  CALL M250(COMMAND%, DATA1&, ERRORCODE%)
  IF ERRORCODE% <> 0 THEN GOTO ERRORTRAP
RETURN

'****   calculate the maximum and minimum rate for normal scans
NORMALMAXMINRATE:
  GOSUB GETAVGTIME
  SCANTIME# = (ST + AVGTIME + INTERCHAN% * ICD) * 2 * NUMCCS% + ISDM
  MAXRATE# = 1 / SCANTIME#
  MINRATE# = 1 / (1.109776# + SCANTIME#)
RETURN

'****   calculate the rate for normal scans
CALCNORMALRATE:
  SCANTIME# = (ST + AVGTIME + INTERCHAN% * ICD) * 2 * NUMCCS% + ISDM
  NORMALDELAY = INT((((1 / NORMALRATE#) - (SCANTIME#)) * 1000000) + .5)
  NORMALRATE# = 1 / ((NORMALDELAY / 1000000) + SCANTIME#)
RETURN

'****   Get starting & ending control codes for a scanning channel
GETCHANNEL:
  CLS : LOCATE 23, 4
  PRINT "Enter an end code smaller than the start code to skip this channel"
  LOCATE 12, 1: PRINT "Channel number:    "; VALA%
  PRINT : INPUT "Enter starting control code"; STARTCC$
  STARTCC% = VAL(STARTCC$)
  IF (STARTCC% < 0) OR (STARTCC% > 15) THEN GOTO GETCHANNEL
GETEND:
  PRINT : INPUT "Enter ending control code"; ENDCC$
  ENDCC% = VAL(ENDCC$)
  IF (ENDCC% < 0) OR (ENDCC% > 15) THEN GOTO GETEND
  IF STARTCC% > ENDCC% THEN STARTCC% = 8: ENDCC% = 0
  CHAN%(VALA%) = STARTCC% * 16 + ENDCC%
RETURN

'****   Send starting & ending channel control codes and number of datapoints
SETSCANCHANNELS:
  COMMAND% = SetScanChannelsCMD%: DATA1& = 0    '0 = normal scan setup
  CALL M250(COMMAND%, DATA1&, ERRORCODE%)
  IF ERRORCODE% <> 0 THEN GOTO ERRORTRAP
  FOR N% = 0 TO 5                'put control codes for each channel in array
    COMMAND% = SetScanChannelsCMD% + (N% + 1): DATA1& = CHAN%(N%)
    CALL M250(COMMAND%, DATA1&, ERRORCODE%)
    IF ERRORCODE% <> 0 THEN GOTO ERRORTRAP
  NEXT N%
  COMMAND% = SetScanChannelsCMD% + 7: DATA1& = NUMCCS%  'put # of datapoints
  CALL M250(COMMAND%, DATA1&, ERRORCODE%)                'in array
  IF ERRORCODE% <> 0 THEN GOTO ERRORTRAP
RETURN

'****   get interchannel delay
GETINTERDELAY:
  CLS
  LOCATE 20, 5
  PRINT "Delay will be rounded to a multiple of 17 microseconds."
  LOCATE 3, 1
  PRINT "Current interchannel delay = "; INTERCHANDELAY%
  LOCATE 5, 1
  INPUT "Enter interchannel delay 0-4335 ", INTERCHANDELAY$
  IF INTERCHANDELAY$ <> "" THEN
    INTERCHAN% = (VAL(INTERCHANDELAY$) + 8) \ 17
    IF INTERCHAN% > 255 THEN INTERCHAN% = 255
    INTERCHANDELAY% = INTERCHAN% * 17
  END IF
RETURN

'****   get number of scans for normal scanning
SETNUMNORMALSCANS:
  CLS
  IF NUMNORMALSCANS > MAXSCANS THEN PRINT "TOO MANY SCANS"
  LOCATE 3, 1: PRINT "Current scans: "; NUMNORMALSCANS
  LOCATE 5, 1: PRINT "Number of scans to perform (1 to"; MAXSCANS; "), "
  INPUT "or enter to leave unchanged ", A$
  A = VAL(A$)
  IF A > MAXSCANS THEN A = MAXSCANS
  IF A < 1 THEN A = 1
  IF A$ <> "" THEN NUMNORMALSCANS = A
  IF NUMNORMALSCANS > MAXSCANS THEN NUMNORMALSCANS = MAXSCANS
RETURN

'****   get correction factor for scan
GETGAINCORRECTION:
  POWER2% = 9 - INT(LOG(AVERAGE%) / LOG(4))
  II = 0: FANS# = 0 'initialize vars
  LOCATE 10, 10: PRINT "Press any key to cancel";
  LOCATE 3, 1
  GOSUB CALCGAINRATE                  'find gainrate#
  PRINT "Calibration will take   ";
  PRINT USING "#####.### seconds"; ((2 ^ (POWER2% + 1)) / GAINRATE#)
  COMMAND% = SetScanChannelsCMD%: DATA1& = 0    '0 = normal scan setup
  CALL M250(COMMAND%, DATA1&, ERRORCODE%)
  IF ERRORCODE% <> 0 THEN GOTO ERRORTRAP

'********** SELECT SPECIAL OFFSET CHANNEL SCAN
  COMMAND% = SetScanChannelsCMD% + 1: DATA1& = &HA0
  CALL M250(COMMAND%, DATA1&, ERRORCODE%)
  IF ERRORCODE% <> 0 THEN GOTO ERRORTRAP
  FOR N% = 2 TO 6
    COMMAND% = SetScanChannelsCMD% + N%: DATA1& = &H80
    CALL M250(COMMAND%, DATA1&, ERRORCODE%)
    IF ERRORCODE% <> 0 THEN GOTO ERRORTRAP
  NEXT N%
  COMMAND% = SetScanChannelsCMD% + 7: DATA1& = 2
  CALL M250(COMMAND%, DATA1&, ERRORCODE%)
  IF ERRORCODE% <> 0 THEN GOTO ERRORTRAP

  COMMAND% = SendSetupCMD%: DATA1& = 0
  CALL M250(COMMAND%, DATA1&, ERRORCODE%)
  IF ERRORCODE% <> 0 THEN GOTO ERRORTRAP
 
  COMMAND% = ScanCalCMD%
  RESPONSE& = 2 ^ POWER2% + (POWER2% * 65536)
  CALL M250(COMMAND%, RESPONSE&, ERRORCODE%)
  IF ERRORCODE% <> 0 THEN GOTO ERRORTRAP
  IF INKEY$ <> "" THEN GOTO GETGAINEXIT
 
  COMMAND% = SetScanChannelsCMD%: DATA1& = 0    '0 = normal scan setup
  CALL M250(COMMAND%, DATA1&, ERRORCODE%)
  IF ERRORCODE% <> 0 THEN GOTO ERRORTRAP

'********** SELECT SPECIAL FULL SCALE SCAN
  COMMAND% = SetScanChannelsCMD% + 1: DATA1& = &H90
  CALL M250(COMMAND%, DATA1&, ERRORCODE%)
  IF ERRORCODE% <> 0 THEN GOTO ERRORTRAP
  FOR N% = 2 TO 6
    COMMAND% = SetScanChannelsCMD% + N%: DATA1& = &H80
    CALL M250(COMMAND%, DATA1&, ERRORCODE%)
    IF ERRORCODE% <> 0 THEN GOTO ERRORTRAP
  NEXT N%
  COMMAND% = SetScanChannelsCMD% + 7: DATA1& = 1
  CALL M250(COMMAND%, DATA1&, ERRORCODE%)
  IF ERRORCODE% <> 0 THEN GOTO ERRORTRAP
'**********
  COMMAND% = SendSetupCMD%: DATA1& = 2 ^ POWER2%
  CALL M250(COMMAND%, DATA1&, ERRORCODE%)
  IF ERRORCODE% <> 0 THEN GOTO ERRORTRAP
  DWORDSINDARRAY = 2 ^ POWER2%                   'allocate scan array
  GOSUB SETUPDARRAY
  COMMAND% = DoNormalScanCMD%: RESPONSE& = DSTARTSEG%
  CALL M250(COMMAND%, RESPONSE&, ERRORCODE%)
  IF ERRORCODE% <> 0 THEN GOTO ERRORTRAP
  COMPLETESCANS = RESPONSE&
  IF INKEY$ <> "" THEN GOTO GETGAINEXIT
  WHILE II <> COMPLETESCANS
    II = II + 1: GOSUB GETNEXTDATAPOINT: FANS# = FANS# + ANSWER
  WEND
  TFULL# = (FANS# / COMPLETESCANS)
  MINUSGAIN# = 5000 / (5000 + (TFULL# - 5000) * .5)
  PLUSGAIN# = 5000 / (5000 + (TFULL# - 5000) * .75)
GETGAINEXIT:
  LOCATE 3, 1: PRINT SPACE$(50)
RETURN

CALCGAINRATE:   'calculate the gain correction scan rate
  GOSUB NORMALMAXMINRATE
  IF TSCAN% = 1 THEN GAINRATE# = NORMALRATE# * NUMCCS%
  IF TSCAN% = 2 THEN GAINRATE# = (1 / STAMPINTVL#) * (NUMCCS% + 1)
  IF TSCAN% = 3 THEN GAINRATE# = (1 / CALINTVL#) * (NUMCCS% + 1)
  SCANTIME# = (ST + AVGTIME + INTERCHAN% * ICD) * 2 + ISDM
  GAINDELAY = INT(((1 / GAINRATE#) - SCANTIME#) * 1000000 + .5)
  GAINRATE# = 1 / ((GAINDELAY / 1000000) + SCANTIME#)
  IF GAINDELAY < 0 THEN GAINDELAY = 0: GAINRATE# = MAXRATE#
  IF GAINDELAY > 1109776 THEN GAINDELAY = 1109776: GAINLRATE# = MINRATE#
  GAINRATE# = 1 / ((GAINDELAY / 1000000) + SCANTIME#)
  IF GAINRATE# < 204.8 THEN
    GAINRATE# = 204.8
    GAINDELAY = INT((((1 / GAINRATE#) - SCANTIME#) * 1000000) + .5)
    IF GAINDELAY < 0 THEN GAINDELAY = 0: GAINRATE# = MAXGAINRATE#
    IF GAINDELAY > 1109776 THEN GAINDELAY = 1109776: GAINLRATE# = MINGAINRATE#
  END IF
  COMMAND% = SetRateCMD% + INTERCHAN%
  DATAHI% = INT(GAINDELAY / 4335)                     'get hi delay
  IF DATAHI% > 255 THEN DATAHI% = 255
  DATALO% = INT((GAINDELAY - DATAHI% * 4335&) / 17)   'get lo delay
  SINGLETICK% = GAINDELAY - DATAHI% * 4335& - DATALO% * 17  'get singletick
  DATA1& = DATALO% + SINGLETICK% * &H100 + DATAHI% * &H10000  'send delays
  CALL M250(COMMAND%, DATA1&, ERRORCODE%)
  IF ERRORCODE% <> 0 THEN GOTO ERRORTRAP
  GAINRATE# = 1 / ((GAINDELAY / 1000000) + SCANTIME#)
RETURN

'****   if averaging, calculate time to average
GETAVGTIME:
  AVGTIME = 0
  IF AVERAGE% > 1 THEN
    AVGTIME = 176.6 * AVERAGE% - 164 + 2.2 * LOG(AVERAGE%) / LOG(2#)
    AVGTIME = AVGTIME / 1000000
  END IF
RETURN

'!************************************************************************
'************    start of time-stamp scan routines     *******************
'*************************************************************************

STAMPSCAN:
  CLS
  LOCATE 1, 29: PRINT "SCANNING CONFIGURATION"
  LOCATE 3, 14: PRINT "Channel      External control code"
  LOCATE 3, 54: PRINT "Number of data"
  LOCATE 4, 29: PRINT "from           to        points this channel"
  NUMCCS% = 0                   'calculate total number of active channels
  FOR N% = 0 TO 5
    CC%(N%) = CHAN%(N%) MOD 16 - CHAN%(N%) \ 16 + 1
    IF CC%(N%) = -7 THEN CC%(N%) = 0
    NUMCCS% = NUMCCS% + CC%(N%)                         '# of datapoints
    LOCATE 6 + N% * 2, 17: PRINT USING "#)"; N%;
    PRINT , CHAN%(N%) \ 16, CHAN%(N%) MOD 16, CC%(N%); ""
    VIDCHANS%(N%) = 0
  NEXT N%
  FOR N% = 0 TO (NUMCCS% - 1)
    IF N% < 6 THEN VIDCHANS%(N%) = 1
  NEXT N%
  REDIM DARRAY&(10)                       'deallocate DARRAY&
  IF FRE(-1) < &H20200 THEN MAXLEN = FRE(-1) \ 4 - 512 ELSE MAXLEN = 32766
  IF NUMCCS% > 0 THEN
    MAXBUFFER = INT(MAXLEN / (NUMCCS% + 1))  'calc max buffer size
  END IF

  LOCATE 19, 17
    PRINT "6) Interchannel delay               ";
    PRINT USING " ####   microseconds"; INTERCHANDELAY%
  LOCATE 20, 17
  IF STAMPINTVL# > 1 THEN
    PRINT "7) Scanning interval, seconds      ";
    PRINT USING "######.######"; STAMPINTVL#;
  ELSE
    PRINT "7) Scanning rate                   ";
    PRINT USING "######.# Hz"; 1 / (STAMPINTVL#);
  END IF
  LOCATE 21, 17
  PRINT USING "8) Size of circular buffer       ########"; STAMPBUFFER;
  LOCATE 22, 17
  PRINT "9) Trigger Type                         "; TRIGGER$(TRIGGERTYPE%);
  LOCATE 24, 13: PRINT "Enter) begin scanning";
  LOCATE 24, 40: PRINT USING "Total points per scan ##"; NUMCCS%;
  LOCATE 25, 40: PRINT "T) Toggle scan type:"; SCANTYPE$(TSCAN%);
  LOCATE 25, 13: PRINT "Esc) Return to Main Menu";
  A$ = UCASE$(INPUT$(1)): VALA% = VAL(A$)

  IF (VALA% = 0) AND (A$ <> "0") THEN VALA% = 15  'flag invalid keystroke
  IF (VALA% >= 0) AND (VALA% <= 5) THEN GOSUB GETCHANNEL

  IF A$ = "T" THEN GOTO EXITSTAMP              'cycle to next scan type
  IF A$ = "6" THEN GOSUB GETINTERDELAY
  IF A$ = "7" THEN GOSUB GETSTAMPINTERVAL
  IF A$ = "8" THEN GOSUB SETSTAMPBUFFER
  IF A$ = "9" THEN TRIGGERTYPE% = (TRIGGERTYPE% + 1) MOD 3
  IF A$ = CHR$(13) AND NUMCCS% > 0 THEN GOSUB DOSTAMPSCAN
  IF A$ <> CHR$(27) THEN GOTO STAMPSCAN
EXITSTAMP:
RETURN

DOSTAMPSCAN:
  CLS : COMPLETESCANS = 0
  LOCATE 6, 20: PRINT "Begin Scanning Menu"
  LOCATE 10, 20: PRINT "1) Scan Calibration"
  LOCATE 12, 20: PRINT "2) Begin Scanning"
  LOCATE 14, 20: PRINT "3) Begin Scanning and Graphing"
  LOCATE 16, 18: PRINT "Esc) Return to Scan Configuration Menu";
 
  A$ = UCASE$(INPUT$(1))
  CLS
  IF A$ = CHR$(27) THEN GOTO EXITDOSTAMP
  IF VAL(A$) < 1 OR VAL(A$) > 3 THEN GOTO DOSTAMPSCAN
  IF STAMPINTVL# = 0 THEN GOSUB GETSTAMPINTERVAL
  GOSUB SETSTAMPINTERVAL
  IF STAMPBUFFER > MAXBUFFER THEN GOSUB SETSTAMPBUFFER
  IF A$ = "1" THEN
    CLS
    GOSUB GETGAINCORRECTION     'get correction factor for scan
    IF FANS# = 0 THEN GOTO EXITDOSTAMP
    STAMPGAINFLAG% = 1          'successful gain correction for stamp scans
    GOTO DOSTAMPSCAN
  END IF
  IF A$ = "2" THEN NOVGA% = 1           'take data
  IF A$ = "3" THEN NOVGA% = 0           'display data on graph

'if there was no gain correction scan, do gain correcton scan
  IF STAMPGAINFLAG% = 0 THEN
    CLS
    PRINT "No calibration scan found, performing calibration scan"
    GOSUB GETGAINCORRECTION     'get correction factor for scan
    IF FANS# = 0 THEN GOTO EXITDOSTAMP
    STAMPGAINFLAG% = 1          'successful gain correction for stamp scans
    LOCATE 1, 1: PRINT SPACE$(79)
    LOCATE 10, 1: PRINT "Press space to start scanning, ESC to quit";
    A$ = UCASE$(INPUT$(1))
    IF A$ = CHR$(27) THEN GOTO EXITDOSTAMP
    GOSUB SETSTAMPINTERVAL
  END IF

STARTSCANS:
  GOSUB SETSCANCHANNELS      'setup scan control codes

'send scan configuration block, type of trigger, and size of buffer
  COMMAND% = SendSetupCMD% + TRIGGERTYPE%
  DATA1& = STAMPBUFFER                             'size of circular buffer
  CALL M250(COMMAND%, DATA1&, ERRORCODE%)
  IF ERRORCODE% <> 0 THEN GOTO ERRORTRAP
 
  LOCATE 10, 1: PRINT SPACE$(70);
  LOCATE 6, 1: PRINT "Scanning started at "; TIME$; "  on  "; DATE$
  DWORDSINDARRAY = (NUMCCS% + 1) * STAMPBUFFER     'allocate circular buffer
  GOSUB SETUPDARRAY

  TINTERVAL# = STAMPINTVL#: STNUM& = STAMPS&: COUNTER& = 0: MISSEDSCAN = 0
  IF NOVGA% = 0 THEN
    SCREEN 12
    PALETTE 3, (0 * 65536 + 35 * 256 + 63)        'create orange color,
                                                  'assign to color 3
  END IF
  GOSUB SETUPINTERRUPT          'setup interrupt rate and init new ISR
  
  COMMAND% = DoIntScanCMD%      'sends start scan to M250 and sets
                                'the timer interrupt to start gathering data
 
  RESPONSE& = DSTARTSEG%        'send starting segment for buffer
  CALL M250(COMMAND%, RESPONSE&, ERRORCODE%)
  IF ERRORCODE% <> 0 THEN GOTO ERRORTRAP
  IF NOVGA% = 0 THEN
    CALL DOTGRAPH
    SCREEN 0: COLOR 7
    IF A$ = CHR$(27) THEN GOTO DOSTAMPSCAN
  ELSE
    GOSUB GETTIMERSCANS
  END IF
  'pointer to current location
  BUFFERPOINTER% = ((RESPONSE& * (NUMCCS% + 1)) + 1) MOD DWORDSINDARRAY
  IF DARRAY&(BUFFERPOINTER%) = 0 THEN
    SCANSTART% = 0: COMPLETESCANS = RESPONSE&     'next location is zero
  ELSE
  SCANSTART% = ((RESPONSE& * (NUMCCS% + 1))) MOD DWORDSINDARRAY
  COMPLETESCANS = STAMPBUFFER
  END IF
  GOSUB DISPSTAMP
EXITDOSTAMP:
  IF A$ <> CHR$(27) THEN GOTO DOSTAMPSCAN
  A$ = ""
RETURN

'****   display the time-stamp scan data
DISPSTAMP:
  CLS : NUMSCAN = 1: II = SCANSTART%: NONSTOP% = 0
  PRINT " CC     chan 0       chan 1       chan 2      ";
  PRINT " chan 3       chan 4       chan 5": PRINT
  FOR I% = 0 TO 15: PRINT I%: NEXT
STAMPSCANLOOP:
  II = II + 1: IF II > DWORDSINDARRAY THEN II = 1
  SCANCOUNTER# = DARRAY&(II - 1)                'get time-stamp number
  LOCATE 24, 1: PRINT "elapsed time:  ";
  PRINT USING "########.####"; (SCANCOUNTER# - 1) * (STAMPINTVL#);
  FOR N% = 0 TO 5                   'get # of datapoints for each channel
    STARTCC% = CHAN%(N%) \ 16: ENDCC% = CC%(N%) + STARTCC% - 1
    FOR CC% = STARTCC% TO ENDCC%
      II = II + 1: IF II > DWORDSINDARRAY THEN II = 1
      GOSUB CONVDISP
    NEXT CC%
  NEXT N%

  LOCATE 25, 1: PRINT USING "scan ######  of ######"; NUMSCAN; COMPLETESCANS;
  NUMSCAN = NUMSCAN + 1
  IF NUMSCAN > COMPLETESCANS THEN GOTO STAMPSCANSDONE
  PRINT "  Press space to continue, R to re-run, ESC to quit";
  A$ = UCASE$(INKEY$)
  IF NONSTOP% = 0 THEN A$ = UCASE$(INPUT$(1))
  IF A$ = "N" THEN
    IF NONSTOP% = 1 THEN NONSTOP% = 0 ELSE NONSTOP% = 1
  END IF
  IF A$ = "R" THEN GOTO DISPSTAMP
  IF A$ = CHR$(27) THEN GOTO EXITSTAMPSCAN
  GOTO STAMPSCANLOOP
STAMPSCANSDONE:
  LOCATE 25, 23: PRINT "  Press R to re-run, ESC to quit"; SPACE$(25);
  A$ = UCASE$(INKEY$)
  IF A$ = "N" THEN
    IF NONSTOP% = 1 THEN NONSTOP% = 0 ELSE NONSTOP% = 1
    A$ = "R"
  END IF
  IF A$ = "R" THEN GOTO DISPSTAMP
  IF A$ <> CHR$(27) THEN GOTO STAMPSCANSDONE
EXITSTAMPSCAN:
  A$ = ""
RETURN

'****   calculate the maximum and minimum interval for time-stamp scanning
CALCSTAMPMAXMIN:
  GOSUB GETAVGTIME
  SCANTIME# = (ST + .000008 + AVGTIME + INTERCHAN% * ICD) * 2 * NUMCCS%
  STAMPMIN# = SCANTIME# + .0001052 + .000062
  STAMPMAX = .000062 * 16777215
RETURN

'****   get time-stamp scan interval
GETSTAMPINTERVAL:
  CLS
  GOSUB CALCSTAMPMAXMIN
  IF INT(STAMPINTVL#) >= 1 THEN
    PRINT USING "   MIN #####.###### sec    MAX #####.###### sec"; STAMPMIN#; STAMPMAX
    INPUT "Enter time-stamp scan interval in seconds "; STAMPINTERVAL$
    SINTVL# = VAL(STAMPINTERVAL$)
  ELSE
    PRINT USING "   MIN #####.###### Hz     MAX #####.###### Hz"; 1 / STAMPMAX; 1 / STAMPMIN#
    INPUT "Enter time-stamp scan rate in hertz "; STAMPINTERVAL$
    IF VAL(STAMPINTERVAL$) = 0 THEN
      SINTVL# = STAMPMIN#
    ELSE
      SINTVL# = 1 / VAL(STAMPINTERVAL$)
    END IF
  END IF
  IF SINTVL# < STAMPMIN# THEN SINTVL# = STAMPMIN#
  IF SINTVL# > STAMPMAX THEN SINTVL# = STAMPMAX
  STMIN# = STAMPMIN#
  GOSUB CALCSTAMPINTERVAL
RETURN

'****   set time-stamp scan interval
SETSTAMPINTERVAL:
  GOSUB CALCSTAMPMAXMIN
  GOSUB CALCSTAMPINTERVAL
  IF STAMPINTVL# < STAMPMIN# OR STAMPINTVL# > STAMPMAX THEN
    CLS : PRINT USING "   #####.######       "; STAMPINTVL#;
    PRINT USING "MIN #####.######     MAX #####.######"; STAMPMIN#; STAMPMAX
    LOCATE 4, 1
    PRINT "Scan interval out of limits.  Press any key to continue"
    A$ = INPUT$(1): A$ = "": CLS : GOSUB GETSTAMPINTERVAL
  END IF
SENDSTAMPINTVL:
  'send inter-channel delay
  COMMAND% = SetRateCMD% + INTERCHAN%
  DATA1& = &H10101                'inter-scan delay counters not used
  CALL M250(COMMAND%, DATA1&, ERRORCODE%)
  IF ERRORCODE% <> 0 THEN GOTO ERRORTRAP
  COMMAND% = SetStampRateCMD% + STAMPLO%
  DATA1& = STAMPMID% + STAMPHI% * &H100&
  CALL M250(COMMAND%, DATA1&, ERRORCODE%)
  IF ERRORCODE% <> 0 THEN GOTO ERRORTRAP
RETURN

'****   calculate time-stamp scan interval
CALCSTAMPINTERVAL:
  STAMPINTVL# = INT(((SINTVL# - (STAMPMIN# - .000062)) / .000062) + .0000005#)
  SS1# = STAMPINTVL#
  STAMPHI% = INT(STAMPINTVL# / 65536)
  STAMPMID% = INT((STAMPINTVL# - STAMPHI% * 65536) / 256)
  STAMPLO% = INT(STAMPINTVL# - (STAMPHI% * 65536 + STAMPMID% * 256&))
  'calc stamp scan interval based on # of whole rollovers
  STAMPS& = ((STAMPHI% * 65536 + STAMPMID% * 256& + STAMPLO%))
  STAMPINTVL# = STAMPMIN# + (STAMPS& - 1) * .000062
RETURN

'****   get size of circular buffer for time-stamp scanning
SETSTAMPBUFFER:
  CLS
  IF STAMPBUFFER > MAXBUFFER THEN PRINT "BUFFER TOO LARGE"
  LOCATE 3, 1: PRINT "Current buffer size: "; STAMPBUFFER
  LOCATE 5, 1: PRINT "Enter size of buffer (1 to"; MAXBUFFER; "), "
  INPUT "or enter to leave unchanged ", A$
  A = VAL(A$)
  IF A > MAXBUFFER THEN A = MAXBUFFER
  IF A < 1 THEN A = 1
  IF A$ <> "" THEN STAMPBUFFER = A
  IF STAMPBUFFER > MAXBUFFER THEN STAMPBUFFER = MAXBUFFER
RETURN

'!************************************************************************
'*********    start of calibration time-stamp scan routines     **********
'*************************************************************************

CALSCAN:
  CLS
  LOCATE 1, 29: PRINT "SCANNING CONFIGURATION"
  LOCATE 3, 14: PRINT "Channel      External control code"
  LOCATE 3, 54: PRINT "Number of data"
  LOCATE 4, 29: PRINT "from           to        points this channel"
  NUMCCS% = 0                   'calculate total number of active channels
  FOR N% = 0 TO 5
    CC%(N%) = CHAN%(N%) MOD 16 - CHAN%(N%) \ 16 + 1
    IF CC%(N%) = -7 THEN CC%(N%) = 0
    NUMCCS% = NUMCCS% + CC%(N%)                         '# of datapoints
    LOCATE 6 + N% * 2, 17: PRINT USING "#)"; N%;
    PRINT , CHAN%(N%) \ 16, CHAN%(N%) MOD 16, CC%(N%); ""
    VIDCHANS%(N%) = 0
  NEXT N%
  NUMCCS% = NUMCCS% + 1
  FOR N% = 0 TO (NUMCCS% - 1)
    IF N% < 6 THEN VIDCHANS%(N%) = 1
  NEXT N%
 
  REDIM DARRAY&(10)                     'deallocate DARRAY&
  IF FRE(-1) < &H20200 THEN MAXLEN = FRE(-1) \ 4 - 512 ELSE MAXLEN = 32766
  IF NUMCCS% > 0 THEN
    MAXBUFFER = INT(MAXLEN / (NUMCCS% + 1))      'calc max buffer size
  END IF

  LOCATE 19, 17
  PRINT "6) Interchannel delay               ";
  PRINT USING " ####   microseconds"; INTERCHANDELAY%
  LOCATE 20, 17
  IF CALINTVL# > 1 THEN
    PRINT "7) Scanning interval, seconds      ";
    PRINT USING "######.######"; CALINTVL#;
  ELSE
    PRINT "7) Scanning rate                   ";
    PRINT USING "######.# Hz"; 1 / CALINTVL#;
  END IF
  LOCATE 21, 17
  PRINT USING "8) Size of circular buffer       ########"; CALBUFFER;
  LOCATE 22, 17
  PRINT "9) Trigger Type                         "; TRIGGER$(TRIGGERTYPE%);
  LOCATE 24, 13: PRINT "Enter) begin scanning";
  LOCATE 24, 40: PRINT USING "Total points per scan ##"; NUMCCS%;
  LOCATE 25, 40: PRINT "T) Toggle scan type:"; SCANTYPE$(TSCAN%);
  LOCATE 25, 13: PRINT "Esc) Return to Main Menu";
  A$ = UCASE$(INPUT$(1)): VALA% = VAL(A$)

  IF (VALA% = 0) AND (A$ <> "0") THEN VALA% = 15  'flag invalid keystroke
  IF (VALA% >= 0) AND (VALA% <= 5) THEN GOSUB GETCHANNEL

  IF A$ = "T" THEN GOTO EXITCAL              'cycle to next scan type
  IF A$ = "6" THEN GOSUB GETINTERDELAY
  IF A$ = "7" THEN GOSUB GETCALRATE
  IF A$ = "8" THEN GOSUB SETCALBUFFER
  IF A$ = "9" THEN TRIGGERTYPE% = (TRIGGERTYPE% + 1) MOD 3
  IF A$ = CHR$(13) AND NUMCCS% > 0 THEN GOSUB DOCALSCAN
  IF A$ <> CHR$(27) THEN GOTO CALSCAN
EXITCAL:
RETURN

'****   Perform a time-stamp scan with calibrations
DOCALSCAN:
  CLS : COMPLETESCANS = 0
  LOCATE 6, 20: PRINT "Begin Scanning Menu"
  LOCATE 10, 20: PRINT "1) Scan Calibration"
  LOCATE 12, 20: PRINT "2) Begin Scanning"
  LOCATE 14, 20: PRINT "3) Begin Scanning and Graphing"
  LOCATE 16, 18: PRINT "Esc) Return to Scan Configuration Menu";

  A$ = UCASE$(INPUT$(1))
  CLS
  IF A$ = CHR$(27) THEN GOTO EXITDOSTAMP
  IF VAL(A$) < 1 OR VAL(A$) > 3 THEN GOTO DOCALSCAN
  IF CALINTVL# = 0 THEN GOSUB GETCALRATE
  GOSUB SETCALRATE
  IF CALBUFFER > MAXBUFFER THEN GOSUB SETCALBUFFER
  IF A$ = "1" THEN
    CLS
    GOSUB GETGAINCORRECTION                   'get correction factor for scan
    IF FANS# = 0 THEN GOTO EXITDOCAL
    CALGAINFLAG% = 1          'successful gain correction for calibrate scans
    GOTO DOCALSCAN
  END IF
  IF A$ = "2" THEN NOVGA% = 1           'take data
  IF A$ = "3" THEN NOVGA% = 0           'display data on graph
 
  IF CALGAINFLAG% = 0 THEN
    CLS
    PRINT "No calibration scan found, performing calibration scan"
    GOSUB GETGAINCORRECTION     'get correction factor for scan
    IF FANS# = 0 THEN GOTO EXITDOCAL
    CALGAINFLAG% = 1          'successful gain correction for calibrate scans
    LOCATE 1, 1: PRINT SPACE$(79)
    LOCATE 10, 1: PRINT "Press space to start scanning, ESC to quit";
    A$ = UCASE$(INPUT$(1))
    IF A$ = CHR$(27) THEN GOTO EXITDOCAL
    GOSUB SETCALRATE
  END IF

STARTSCANS1:
  GOSUB SETCALRATE
  GOSUB SETSCANCHANNELS      'setup scan control codes

 'send scan configuration block, type of trigger, and size of circular buffer
  COMMAND% = SendSetupCMD% + TRIGGERTYPE%
  DATA1& = CALBUFFER                             'size of buffer
  CALL M250(COMMAND%, DATA1&, ERRORCODE%)
  IF ERRORCODE% <> 0 THEN GOTO ERRORTRAP
  IF A$ = CHR$(27) THEN GOTO EXITDOCAL
  LOCATE 10, 1: PRINT SPACE$(70);
  LOCATE 6, 1: PRINT "Scanning started at "; TIME$; "  on  "; DATE$
  DWORDSINDARRAY = (NUMCCS% + 1) * CALBUFFER       'allocate circular buffer
  GOSUB SETUPDARRAY

  TINTERVAL# = CALINTVL#: STNUM& = STAMPSCAL&: COUNTER& = 0: MISSEDSCAN = 0
  CALFULL# = 0: CALCOUNT& = 0
  IF NOVGA% = 0 THEN
    SCREEN 12
    PALETTE 3, (0 * 65536 + 35 * 256 + 63)        'create orange color,
                                                  'assign to color 3
  END IF
  GOSUB SETUPINTERRUPT          'setup interrupt rate and init new ISR

  COMMAND% = DoCalIntScanCMD%   'sends start scan to M250 and sets
                                'the timer interrupt to start gathering data

  RESPONSE& = DSTARTSEG%        'send starting segment for buffer
  CALL M250(COMMAND%, RESPONSE&, ERRORCODE%)
  IF ERRORCODE% <> 0 THEN GOTO ERRORTRAP
  IF NOVGA% = 0 THEN
    CALL DOTGRAPH
    SCREEN 0: COLOR 7
    IF A$ = CHR$(27) THEN GOTO DOCALSCAN
  ELSE
    GOSUB GETTIMERSCANS
  END IF

  BUFFERPOINTER% = ((RESPONSE& * (NUMCCS% + 1)) + 1) MOD DWORDSINDARRAY
  IF DARRAY&((BUFFERPOINTER% + 1) MOD DWORDSINDARRAY) = 0 THEN
    CALSCANSTART% = 0: COMPLETESCANS = RESPONSE&     'next location is zero
  ELSE
    CALSCANSTART% = ((RESPONSE& * (NUMCCS% + 1))) MOD DWORDSINDARRAY
    COMPLETESCANS = CALBUFFER
  END IF
  GOSUB DISPCAL
EXITDOCAL:
  IF A$ <> CHR$(27) THEN GOTO DOCALSCAN
  A$ = ""
RETURN

'****   display the time-stamp scan data and gain channel
DISPCAL:
  CLS : NUMSCAN = 1: II = CALSCANSTART%: CALFULL# = 0: NONSTOP% = 0
  TSAVE# = TFULL#: MINUSGAINSAVE# = MINUSGAIN#: PLUSGAINSAVE# = PLUSGAIN#
  PRINT " CC     chan 0       chan 1       chan 2      ";    'display header
  PRINT " chan 3       chan 4       chan 5": PRINT
  FOR I% = 0 TO 15: PRINT I%: NEXT
CALSCANLOOP:
  II = II + 1: IF II > DWORDSINDARRAY THEN II = 1
  SCANCOUNTER# = DARRAY&(II - 1)
  LOCATE 24, 1: PRINT "elapsed time:  ";
  PRINT USING "########.####"; (SCANCOUNTER# - 1) * CALINTVL#;
  II = II + 1: GOSUB GETNEXTDATAPOINT
  IF (NUMSCAN - 1) MOD 256 = 0 AND (NUMSCAN - 1 <> 0) THEN
    TFULL# = CALFULL# / 256
    MINUSGAIN# = 5000 / (5000 + (TFULL# - 5000) * .5)
    PLUSGAIN# = 5000 / (5000 + (TFULL# - 5000) * .75)
    CALFULL# = 0
  END IF
  CALFULL# = CALFULL# + ANSWER
  IF HOLD% = 1 THEN      'secondary gain correction for Track/hold mode only
    IF ANSWER < 0 THEN
      ANSWER = ANSWER * MINUSGAIN#
    ELSE
      ANSWER = ANSWER * PLUSGAIN#
    END IF
  END IF
  LOCATE 22, 1
  PRINT USING "CHANNEL 6  #####.###"; ANSWER;

  FOR N% = 0 TO 5              'get # of datapoints for each channel
    STARTCC% = CHAN%(N%) \ 16: ENDCC% = CC%(N%) + STARTCC% - 1
    FOR CC% = STARTCC% TO ENDCC%
      II = II + 1: IF II > DWORDSINDARRAY THEN II = 1
      GOSUB CONVDISP
    NEXT CC%
  NEXT N%

  LOCATE 25, 1: PRINT USING "scan ######  of ######"; NUMSCAN; COMPLETESCANS;
  NUMSCAN = NUMSCAN + 1
  IF NUMSCAN > COMPLETESCANS THEN GOTO CALSCANSDONE
  PRINT "  Press space to continue, R to re-run, ESC to quit";
  A$ = INKEY$
  IF NONSTOP% = 0 THEN A$ = UCASE$(INPUT$(1))
  IF A$ = "N" THEN
    IF NONSTOP% = 1 THEN NONSTOP% = 0 ELSE NONSTOP% = 1
  END IF
  IF A$ = "R" THEN
    TFULL# = TSAVE#: MINUSGAIN# = MINUSGAINSAVE#: PLUSGAIN# = PLUSGAINSAVE#
    GOTO DISPCAL
  END IF
  IF A$ = CHR$(27) THEN GOTO EXITCALSCAN
  GOTO CALSCANLOOP
CALSCANSDONE:
  LOCATE 25, 23: PRINT "  Press R to re-run, ESC to quit"; SPACE$(25);
  A$ = UCASE$(INKEY$)
  IF A$ = "N" THEN
    IF NONSTOP% = 1 THEN NONSTOP% = 0 ELSE NONSTOP% = 1
    A$ = "R"
  END IF
  IF A$ = "R" THEN TFULL# = TSAVE#: SCANGAIN# = SGAINSAVE#: GOTO DISPCAL
  IF A$ <> CHR$(27) THEN GOTO CALSCANSDONE
EXITCALSCAN:
  A$ = ""
RETURN

CALCCALMAXMIN:
  GOSUB GETAVGTIME
  SCANTIME# = (ST + .000008 + AVGTIME + INTERCHAN% * ICD) * 2 * NUMCCS%
  CALMIN# = SCANTIME# + .0001052 + .000062
  CALMAX = .000062 * 16777215
RETURN

'****   get time-stamp with calibration interval
GETCALRATE:
  CLS
  GOSUB CALCCALMAXMIN
  IF INT(CALINTVL#) >= 1 THEN
    PRINT USING "   MIN #####.###### sec    MAX #####.###### sec"; CALMIN#; CALMAX
    INPUT "Enter time-stamp scan interval in seconds "; CALINTERVAL$
    CINTVL# = VAL(CALINTERVAL$)
  ELSE
    PRINT USING "   MIN #####.###### Hz     MAX #####.###### Hz"; 1 / CALMAX; 1 / CALMIN#
    INPUT "Enter time-stamp scan rate in hertz "; CALINTERVAL$
    IF VAL(CALINTERVAL$) = 0 THEN
      CINTVL# = CALMIN#
    ELSE
      CINTVL# = 1 / VAL(CALINTERVAL$)
    END IF
  END IF
  IF CINTVL# < CALMIN# THEN CINTVL# = CALMIN#   'minimum interval
  IF CINTVL# > CALMAX THEN CINTVL# = CALMAX   'minimum interval
  GOSUB CALCCALINTERCAL
RETURN

'****   set interval for time-stamp with calibration scanning
SETCALRATE:
  GOSUB CALCCALMAXMIN
  GOSUB CALCCALINTERCAL
  IF CALINTVL# < CALMIN# OR CALINTVL# > CALMAX THEN
    CLS : PRINT USING "   #####.######       "; CALINTVL#;
    PRINT USING "MIN #####.######     MAX #####.######"; CALMIN#; CALMAX
    LOCATE 4, 1
    PRINT "Scan interval out of limits.  Press any key to continue"
    A$ = INPUT$(1): A$ = "": CLS : GOSUB GETCALRATE
  END IF

  'send inter-channel delay
  COMMAND% = SetRateCMD% + INTERCHAN%
  DATA1& = &H10101    'inter-scan delay counters not used
  CALL M250(COMMAND%, DATA1&, ERRORCODE%)
  IF ERRORCODE% <> 0 THEN GOTO ERRORTRAP

  COMMAND% = SetStampRateCMD% + CALLO%
  DATA1& = CALMID% + CALHI% * &H100
  CALL M250(COMMAND%, DATA1&, ERRORCODE%)
  IF ERRORCODE% <> 0 THEN GOTO ERRORTRAP
RETURN

'****   calculate interval for time-stamp scan with calibration
CALCCALINTERCAL:
  CALINTVL# = INT(((CINTVL# - (CALMIN# - .000062)) / .000062) + .0000005#)
  CALHI% = INT(CALINTVL# / 65536)
  CALMID% = INT((CALINTVL# - CALHI% * 65536) / 256)
  CALLO% = INT(CALINTVL# - (CALHI% * 65536 + CALMID% * 256&))
  'calc stamp scan interval based on # of whole rollovers
  STAMPSCAL& = ((CALHI% * 65536 + CALMID% * 256& + CALLO%))
  CALINTVL# = CALMIN# + (STAMPSCAL& - 1) * .000062
RETURN

'****   get size of circular buffer for time-stamp scanning with calibration
SETCALBUFFER:
  CLS
  IF CALBUFFER > MAXBUFFER THEN PRINT "BUFFER TOO LARGE"
  LOCATE 3, 1: PRINT "Current buffer size: "; CALBUFFER
  LOCATE 5, 1: PRINT "Enter size of buffer (1 to"; MAXBUFFER; "), "
  INPUT "or enter to leave unchanged ", A$
  A = VAL(A$)
  IF A > MAXBUFFER THEN A = MAXBUFFER
  IF A < 1 THEN A = 1
  IF A$ <> "" THEN CALBUFFER = A
  IF CALBUFFER > MAXBUFFER THEN CALBUFFER = MAXBUFFER
RETURN

'******* setup interrupt rate and initialize new ISR timer interrupt
SETUPINTERRUPT:
  NUMWORDS% = NUMCCS% * 4 + 3
  DOOLDINT8 = &H80: TickCount = &H200
  IF NUMCCS% = 1 THEN
    TT# = NUMWORDS% / TINTERVAL# / 18.2
    WHILE DOOLDINT8 < TT#
      DOOLDINT8 = DOOLDINT8 * 2: TickCount = TickCount \ 2
    WEND
  ELSE
    IF INTERCHAN% <= 1 THEN DOOLDINT8 = &H400: TickCount = &H40
    IF INTERCHAN% >= 2 THEN DOOLDINT8 = &H200: TickCount = &H80
    IF INTERCHAN% > 15 THEN DOOLDINT8 = &H100: TickCount = &H100
  END IF
  IF STNUM& <= 6 AND INTERCHAN% <= 10 THEN
    DOOLDINT8 = DOOLDINT8 * 2: TickCount = TickCount \ 2
  END IF
 
  COMMAND% = SetInterruptCMD%            'command to setup 8254 timer mask
  RESPONSE& = TickCount + (DOOLDINT8 * &H10000) '# of interrupts * 18.2 per second
  CALL M250(COMMAND%, RESPONSE&, ERRORCODE%)
  IF ERRORCODE% <> 0 THEN GOTO ERRORTRAP

  COMMAND% = StartInterruptCMD%
  CALL M250(COMMAND%, RESPONSE&, ERRORCODE%)
  IF ERRORCODE% <> 0 THEN GOTO ERRORTRAP
RETURN

'******* check for digital commands during scanning
CHKDIGCMDS:
  IF DiginPending% <> 0 THEN
    COMMAND% = ReqDigInCMD%: ERRORCODE% = 0
    CALL M250(COMMAND%, DATA1&, ERRORCODE%)
    IF ERRORCODE% = &H1D THEN
      LOCATE 24, 28: PRINT "PENDING";
    ELSE
      IF ERRORCODE% = 0 THEN
        NDIGIN = NDIGIN + 1: LASTDIGINP% = DATA1&: DiginPending% = 0
        LOCATE 24, 28: PRINT USING "###    "; LASTDIGINP%;
      END IF
    END IF
  END IF
  IF Digout% <> 0 THEN
    LOCATE 24, 52: PRINT SPACE$(27);
    LOCATE 24, 38: PRINT "Enter output word  ";
    GOSUB GETINPUT1
    LOCATE 24, 38: PRINT "(O) digital output ";
    'protect against overflow from large number entries
    DIGOUTWORD% = VAL(K$) - INT(VAL(K$) / 256) * 256
    IF ERRORCODE% <> &H1B THEN
      Digout% = 0
      COMMAND% = SendDigOutCMD%: ERRORCODE% = 0
      DATA1& = DigitalOutCMD% + (DIGOUTWORD% MOD 256)
      CALL M250(COMMAND%, DATA1&, ERRORCODE%)
      IF ERRORCODE% <> &H1D THEN
        NDIGOUT = NDIGOUT + 1
        PRINT USING "###"; DIGOUTWORD%;
      END IF
    END IF
  END IF
  IF ERRORCODE% = &H1B THEN RETURN ENDTIMERSCAN
  IF ERRORCODE% = &H1F THEN RETURN DONETIMER
RETURN

GETINPUT1:       '.NEW Input a single character at a time
  K$ = ""
GETNEXTKEY1:
  S$ = INKEY$
  COMMAND% = GetNumWordsCMD%  'get # of words in RESPONSE&
  CALL M250(COMMAND%, RESPONSE&, ERRORCODE%)
  'if errorcode% = &H1B, abort input
  IF RESPONSE& > &HF000& THEN ERRORCODE% = &H1F
  IF ERRORCODE% = &H1B OR ERRORCODE% = &H1F THEN S$ = CHR$(13)
  TEMPROW = CSRLIN: TEMPCOL = POS(0)
  LOCATE 16, 41: PRINT USING " #####"; RESPONSE&;
  IF RESPONSE& > NUMWORDS% THEN
    COMMAND% = GetIntScanCMD%
    CALL M250(COMMAND%, RESPONSE&, ERRORCODE%)
    ERRNUM% = 0
    IF ERRORCODE% = &H1E THEN ERRNUM% = 1 'data lost
    IF ERRORCODE% <> &H1C AND ERRNUM% = 0 THEN SCANCOUNT& = SCANCOUNT& + 1
    IF ERRORCODE% = &H1B OR ERRORCODE% = &H1F THEN S$ = CHR$(13)
  END IF
  LOCATE TEMPROW, TEMPCOL
  IF S$ = "" THEN GOTO GETNEXTKEY1       'if no key, wait for next key
  IF ASC(S$) = 8 THEN                   'if backspace key
    IF LEN(K$) > 0 THEN
      K$ = LEFT$(K$, LEN(K$) - 1)   'one less character
      LOCATE CSRLIN, (POS(0) - 1): PRINT " "; : LOCATE CSRLIN, (POS(0) - 1)
    END IF
  ELSE
    IF (ASC(S$) >= 48 AND ASC(S$) <= 57) THEN   'char between "0" and "9"
      K$ = K$ + S$: PRINT S$;
    END IF
  END IF
  IF S$ <> CHR$(13) THEN GOTO GETNEXTKEY1
RETURN

'***** get time-stamped scans and store in DARRAY&
GETTIMERSCANS:
  LOCATE 25, 10: PRINT "(ESC) cancel scan           any other key to stop";
  LOCATE 24, 10: PRINT "(D) digital input";
  LOCATE 24, 38: PRINT "(O) digital output ";
  STARTTIME$ = TIME$: STARTDAY$ = DATE$: DAYS = 0
  SCANSTART = TIMER                     'the first scan begins now
  ERRNUM% = 0                           'flag for lost data clear
  SCANCOUNT& = 0: NDIGIN = 0: NDIGOUT = 0: DiginPending% = 0: Digout% = 0
  LOCATE 16, 10: PRINT "Bytes in buffer"
  LOCATE 8, 10
  IF TINTERVAL# > 1 THEN
    PRINT "Scanning interval, seconds     ";
    PRINT USING "######.######"; TINTERVAL#;
  ELSE
    PRINT "Scanning rate                  ";
    PRINT USING "######.# Hz"; 1 / (TINTERVAL#);
  END IF
  LOCATE 10, 10: PRINT USING "Words expected per scan           ###"; NUMCCS% + 1
  LOCATE 12, 10: PRINT "Elapsed time, seconds"
  LOCATE 14, 10: PRINT "Scans completed"

GETONETIMER:
  A$ = INKEY$
  IF UCASE$(A$) = "D" THEN DiginPending% = 1: A$ = ""
  IF UCASE$(A$) = "O" THEN Digout% = 1: A$ = ""
  IF A$ = CHR$(27) THEN GOTO DONETIMER
  GOSUB CHECKWORDS
  IF RESPONSE& > &HF000& THEN GOTO DONETIMER
  IF RESPONSE& >= NUMWORDS% THEN
    COMMAND% = GetIntScanCMD%
    CALL M250(COMMAND%, RESPONSE&, ERRORCODE%)
    IF ERRORCODE% = &H1B THEN GOTO ENDTIMERSCAN
    IF ERRORCODE% = &H1E THEN ERRNUM% = 1 ELSE ERRNUM% = 0  'data lost
    IF ERRORDOCE% <> &H1C AND ERRNUM% = 0 THEN SCANCOUNT& = SCANCOUNT& + 1
    IF ERRORCODE% = &H1F THEN GOTO FULLBUFFER
    GOSUB CHKDIGCMDS
    IF ERRORCODE% = &H1B THEN GOTO ENDTIMERSCAN
    IF ERRORCODE% = &H1F THEN GOTO FULLBUFFER
  END IF
  GOSUB CHKDIGCMDS
  IF ERRORCODE% = &H1B THEN GOTO ENDTIMERSCAN
  IF ERRORCODE% = &H1F THEN GOTO FULLBUFFER
  IF A$ = "" THEN GOTO GETONETIMER

DONETIMER:
  IF A$ = CHR$(27) THEN CH% = 2 ELSE CH% = 1
  COMMAND% = CancelIntScanCMD% + CH%            'command to stop filling buffer
                                                'at end of current scan
  CALL M250(COMMAND%, RESPONSE&, ERRORCODE%)
  IF ERRORCODE% = &H10 THEN GOTO ENDTIMERSCAN

FULLBUFFER:
'time to empty the buffer
  ELAPSED = TIMER + DAYS - SCANSTART
  LOCATE 12, 38: PRINT USING "#########.#"; ELAPSED;
  LOCATE 25, 1: PRINT SPACE$(79);
  LOCATE 25, 10: PRINT "Empting buffer. Press ESC to cancel";
GETONEMORE:
  IF INKEY$ = CHR$(27) THEN GOTO ENDTIMERSCAN
  COMMAND% = GetNumWordsCMD%                    'command to get # of bytes in
                                                'the buffer
  CALL M250(COMMAND%, RESPONSE&, ERRORCODE%)    '# of bytes returned in RESPONSE&
  IF ERRORCODE% = &H1B THEN GOTO ENDTIMERSCAN
  LOCATE 16, 41: PRINT USING " #####"; RESPONSE&;
  LOCATE 14, 38: PRINT USING "#########"; SCANCOUNT&
  LOCATE 4, 65: IF ERRNUM% = 1 THEN PRINT "DATA LOST" ELSE PRINT "         ";
  IF RESPONSE& > NUMWORDS% THEN
    COMMAND% = GetIntScanCMD%                 'get one scan
    CALL M250(COMMAND%, RESPONSE&, ERRORCODE%)
    IF ERRORCODE% = &H1E THEN ERRNUM% = 1 ELSE ERRNUM% = 0  'data lost
    IF ERRORDOCE% <> &H1C AND ERRNUM% = 0 THEN SCANCOUNT& = SCANCOUNT& + 1
    GOTO GETONEMORE
  END IF
ENDTIMERSCAN:
  LOCATE 25, 1: PRINT SPACE$(79);
  LOCATE 25, 28: PRINT "Scanning Complete     ";
  IF ERRORCODE% = &H1B THEN LOCATE 25, 28: PRINT "BUFFER OVERFLOW       ";
  COMMAND% = StopIntScanCMD%                    'ends the scan, resets the
                                                'timer interrupt, and resets
                                                'the 8254 mode 2 timer
  CALL M250(COMMAND%, RESPONSE&, ERRORCODE%)
  IF ERRORCODE% <> 0 THEN GOTO ERRORTRAP
  LOCATE 16, 38: PRINT USING "#########"; SCANCOUNT&
RETURN

CHECKWORDS:
  COMMAND% = GetNumWordsCMD%    'command to get # of bytes in the buffer
  CALL M250(COMMAND%, RESPONSE&, ERRORCODE%)    '# of bytes returned in RESPONSE&
  IF ERRORCODE% = &H1B THEN RETURN ENDTIMERSCAN
  LOCATE 16, 41: PRINT USING " #####"; RESPONSE&;
  IF RESPONSE& < 10240 THEN
    ELAPSED = TIMER + DAYS - SCANSTART
    LOCATE 12, 38: PRINT USING "#########.#"; ELAPSED;
    LOCATE 14, 38: PRINT USING "#########"; SCANCOUNT&;
    LOCATE 4, 65: IF ERRNUM% = 1 THEN PRINT "DATA LOST";  ELSE PRINT "         ";
  END IF
RETURN

'check for vga card
CHECKFORVGA:
ON ERROR GOTO NOVGA
  SCREEN 12      '640 x 400  = horizontal, vertical 1,1 = upper left
ON ERROR GOTO 0
  SCREEN 0
NOVGARET:
  ON ERROR GOTO 0
RETURN

'********* VGA graphics not found
NOVGA:
  CLS : PRINT " VGA Video required for graphing routines"
  PRINT "press any key to continue"; A$ = INPUT$(1)
  NOVGA% = 1
RESUME NOVGARET
END

SUB DOTGRAPH
 GetIntScanCMD% = &H9000     'get one scan in interrupt time-stamped scanning
 StopIntScanCMD% = &H8000       'stop interrupt time-stamped scanning
 CancelIntScanCMD% = &H4000     'stop gathering data
                                '     + 1 = stop at end of current scan
                                '     + 2 = cancel scan in progress and exit
 GetNumWordsCMD% = &H5000       'get number of words in buffer
  DIM SVOLTY%(6, 650)
  II = 0: COUNTER& = 0: STOPPED% = 0: MISSEDSCAN = 0
  DISPCNT = 1 / TINTERVAL#: IF TINTERVAL# >= 1 THEN DISPCNT = 1
  GOSUB SETUPSCREEN
GETTSCAN:
  COMMAND% = GetNumWordsCMD%            'get # of words in RESPONSE&
  CALL M250(COMMAND%, RESPONSE&, ERRORCODE%)
  IF ERRORCODE% = &H1B THEN GOTO ENDTSCAN     'buffer overflowed
  IF RESPONSE& > &HF000& AND STOPPED% = 0 THEN GOTO STOPTSCAN
  IF RESPONSE& < NUMWORDS% AND STOPPED% = 1 THEN GOTO ENDTSCAN
  IF RESPONSE& >= NUMWORDS% THEN
    COMMAND% = GetIntScanCMD%
    CALL M250(COMMAND%, RESPONSE&, ERRORCODE%)
    IF ERRORCODE% = 0 THEN GOSUB DISPLAYVOLTS ELSE GOTO GrfErr
  END IF
MISSEDRET:
  A$ = INKEY$
  IF A$ <> "" THEN
    A$ = UCASE$(A$): IF A$ = "E" OR A$ = CHR$(27) THEN GOTO STOPTSCAN
    GOSUB CheckAString
  END IF
  IF COUNTER& MOD 640 = 0 THEN
    LINE (1, 0)-(1, 399), 0: LINE (2, 0)-(2, 399), 0
  END IF
  GOTO GETTSCAN                         'continue scan
STOPTSCAN:
  IF A$ = CHR$(27) THEN CH% = 2 ELSE CH% = 1
  COMMAND% = CancelIntScanCMD% + CH%            'command to stop filling buffer
                                                'at end of current scan
  CALL M250(COMMAND%, RESPONSE&, ERRORCODE%)
  IF A$ = CHR$(27) THEN GOTO ENDTSCAN
  STOPPED% = 1
  LOCATE 30, 25: COLOR 15: PRINT "EMPTING BUFFER";
  GOTO GETTSCAN                                 'empty buffer to screen
ENDTSCAN:
  COLOR 15: LOCATE 30, 20:  PRINT SPACE$(59); : LOCATE 30, 31:
  IF ERRORCODE% = &H1B THEN
    PRINT "BUFFER OVERFLOW";
  ELSE
    LOCATE 30, 1: PRINT SPACE$(10);
    LOCATE 30, 10: PRINT "Scanning Complete.  Press ESC to quit, space to continue";
  END IF
  COMMAND% = StopIntScanCMD%                    'ends the scan, resets the
                                                'timer interrupt, and resets
                                                'the 8254 mode 2 timer
  CALL M250(COMMAND%, RESPONSE&, ERRORCODE%)
  IF ERRORCODE% <> 0 THEN GOTO EXITGRAPHSCAN
  A$ = INPUT$(1)
  GOTO EXITGRAPHSCAN

'******** error handling routine for graphics routines
GrfErr:
  IF ERRORCODE% = &H1B THEN GOTO ENDTSCAN     'buffer overflowed
  IF ERRORCODE% = &H1C AND STOPPED% = 0 THEN GOTO GETTSCAN
  IF ERRORCODE% = &H1C AND STOPPED% = 1 THEN GOTO ENDTSCAN
  IF ERRORCODE% = &H1E THEN DIDMISS& = COUNTER& + 1: MISSEDFLAG% = 1
  IF ERRORCODE% = &H1F THEN GOTO STOPTSCAN
  GOTO MISSEDRET

'******** process keypress from graphics routines
CheckAString:
  IF A$ = "P" THEN
    LOCATE 30, 1: PRINT "Enter positive full scale (mv) "; SPACE$(8);
    LOCATE 30, 32: GOSUB GETINPUT
    IF K$ <> "" AND VAL(K$) > NFS THEN
      PFS = VAL(K$): GOSUB SETUPSCREEN     'setup graphic screen
    END IF
    GOSUB UPDATELINE30
  END IF
  IF A$ = "M" THEN
    LOCATE 30, 1: PRINT "Enter negative full scale (mv) "; SPACE$(8);
    LOCATE 30, 32: GOSUB GETINPUT
    IF K$ <> "" AND VAL(K$) < PFS THEN
      NFS = VAL(K$): GOSUB SETUPSCREEN     'setup graphic screen
    END IF
    GOSUB UPDATELINE30
  END IF
  IF A$ = "H" THEN
    LOCATE 30, 1: PRINT "Enter horizontal division (seconds) "; SPACE$(5);
    LOCATE 30, 37: GOSUB GETINPUT
    IF K$ <> "" AND VAL(K$) > 0 THEN
      HDIV = VAL(K$): GOSUB SETUPSCREEN     'setup graphic screen
    END IF
    GOSUB UPDATELINE30
  END IF
  IF A$ = "V" THEN
    LOCATE 30, 1: PRINT "Enter vertical division (mv) "; SPACE$(10);
    LOCATE 30, 30: GOSUB GETINPUT
    IF K$ <> "" AND VAL(K$) > 0 THEN
      VDIV = VAL(K$) / 1000: GOSUB SETUPSCREEN   'setup graphic screen
    END IF
    GOSUB UPDATELINE30
  END IF
  A$ = ""
RETURN

'********* rewrite line 30 of graphics display
UPDATELINE30:
  LOCATE 30, 1: PRINT SPACE$(65);
  IF MISSEDSCAN > 0 THEN
    LOCATE 30, 1: COLOR 7: PRINT "MISSED SCANS   "; MISSEDSCAN;
  END IF
  LOCATE 30, 50: PRINT "bytes in buffer";
  IF STOPPED% = 1 THEN LOCATE 30, 25: COLOR 15: PRINT "EMPTING BUFFER";
RETURN

'********* display graph of active channel voltages
DISPLAYVOLTS:
  GOSUB UPDATEPOSITION
  GOSUB DISPCHANS               'display channel voltages on graph
  IF MISSEDFLAG% = 1 THEN GOSUB MISSEDSCANS
RETURN

'***** missed scans, clean up graph
MISSEDSCANS:
  IF MISSEDSCAN = 0 THEN
    LOCATE 30, 1: COLOR 7: PRINT "MISSED SCANS   ";
  END IF
  MISSEDSCAN = MISSEDSCAN + COUNTER& - DIDMISS&
  IF COUNTER& - DIDMISS& > DMAX THEN DMAX = COUNTER& - DIDMISS&
  STARTN% = (DIDMISS& - STARTSCREEN&) MOD 640 + 2
  ENDN% = (COUNTER& - STARTSCREEN&) MOD 640 + 2
  FOR N% = STARTN% TO ENDN%
    LINE (N%, 0)-(N%, 399), 0 ' draw vertical black line
    IF (N%) MOD HINDEX% = 0 THEN    ' redraw ticks
       FOR NK% = 0 TO 399 STEP VINDEX&: PSET (N%, NK%), 15: NEXT NK%
    END IF
  NEXT N%
  MISSEDFLAG% = 0
  LOCATE 30, 15: COLOR 7: PRINT MISSEDSCAN;
RETURN

'********* move to next X point on graph
UPDATEPOSITION:
  S% = (COUNTER& - STARTSCREEN&) MOD 640 + 2
  FOR N% = 0 TO (NUMCCS% - 1) ' display data
    IF N% > 5 THEN GOTO EXITUPDATE
    LINE (S%, SVOLTY%(N%, S%))-(S% + 1, SVOLTY%(N%, S% + 1)), 0
  NEXT N%
EXITUPDATE:
  IF (S% + 1) MOD HINDEX% = 0 THEN      ' redraw ticks
    FOR NK% = 0 TO 399 STEP VINDEX&: PSET (S% + 1, NK%), 15: NEXT NK%
  END IF
RETURN

'********* get and display channel voltages on graph
DISPCHANS:
  II = II + 1: IF II > DWORDSINDARRAY THEN II = 1
  COUNTER& = DARRAY&(II - 1)
  FOR N% = 0 TO (NUMCCS% - 1) ' display data
    IF N% > 5 THEN
      II = II + 1: GOSUB GETNEXTGRAPHPOINT
      GOTO NEXTDOT
    END IF
    IF VIDCHANS%(N%) = 0 THEN GOTO NEXTDOT
    II = II + 1
    GOSUB GETNEXTGRAPHPOINT
    V = ANSWER

    IF TSCAN% = 3 AND N% = 0 THEN     'if time-stamped scan with calibration
      CALCOUNT& = CALCOUNT& + 1
      IF (CALCOUNT& - 1) MOD 256 = 0 AND (CALCOUNT& - 1 <> 0) THEN
        TFULL# = CALFULL# / 256: CALFULL# = 0
        MINUSGAIN# = 5000 / (5000 + (TFULL# - 5000) * .5)
        PLUSGAIN# = 5000 / (5000 + (TFULL# - 5000) * .75)
      END IF
      CALFULL# = CALFULL# + ANSWER
    END IF

    IF HOLD% = 1 THEN      'secondary gain correction for Track/hold mode only
      IF ANSWER < 0 THEN V = V * MINUSGAIN# ELSE V = V * PLUSGAIN#
    END IF
    TEMPVOLTS = V
    IF V > PFS THEN V = PFS ELSE IF V < NFS THEN V = NFS
    SVOLTS% = ((V - PFS) / 1000) * VCONST  'calculate y location
    IF S% = 2 THEN SVOLTY%(N%, S% - 1) = SVOLTS% 'if same y for 1st point on screen
    IF MISSEDFLAG% = 1 THEN SVOLTY%(N%, S% - 1) = SVOLTS% 'same y if missed scan
    SCOLOR% = 10 + N%                         'select color
    'line from last x,y to current x,y
    LINE (S% - 1, SVOLTY%(N%, S% - 1))-(S%, SVOLTS%), SCOLOR%
    SVOLTY%(N%, S%) = SVOLTS%                'save current y location
    IF COUNTER& MOD DISPCNT = 0 THEN GOSUB UPDATEVOLTS
NEXTDOT:
  NEXT N%
  COLOR 7
RETURN

'********* update values and display voltages in fields below graph
UPDATEVOLTS:
  COLOR SCOLOR%: LOCATE 27, ((N% * 13) + 5)
  ELAPSED = (COUNTER& - 1) * TINTERVAL#
  PRINT USING "######.##"; TEMPVOLTS;     'display current voltage
  LOCATE 28, 15: COLOR 7: PRINT USING "########"; ELAPSED;
  COMMAND% = GetNumWordsCMD%            'get # of words in RESPONSE&
  CALL M250(COMMAND%, RESPONSE&, ERRORCODE%)
  LOCATE 30, 66: COLOR 7: PRINT USING " ######"; RESPONSE&;
RETURN

'********* setup graphics screen
SETUPSCREEN:
  CLS
'  TINTERVAL#, PFS, NFS, VDIV CANNOT BE 0
  IF PFS = 0 AND NFS = 0 THEN
    PFS = 5000: NFS = -5000
  END IF
 
  IF VDIV = 0 THEN VDIV = 1
  STARTSCREEN& = COUNTER&        'reset screen position counter
  VCONST = (399000 / (NFS - PFS))      'used to get Y location from voltage
  NOVGA% = 0     'there was a VGA screen, flag graphics display
  TSCREEN = 639 * TINTERVAL#       'TSCREEN = time for one screen
  LINE (0, 0)-(0, 400), 2 ' vertical axis
  POSFS = PFS / 1000: NEGFS = NFS / 1000
  VINDEX& = (399 / (ABS(POSFS - NEGFS) / VDIV))
  IF VINDEX& > 399 THEN VINDEX& = 399              'check bounds on VINDEX&
  IF VINDEX& < 1 THEN VINDEX& = 1
  'tick marks on vertical axis
  FOR N% = 0 TO 399 STEP VINDEX&: LINE (0, N% - 1)-(0, N% + 1), 15: NEXT N%
  LINE (1, 400)-(640, 400), 2                   'horizontal axis
  IF HDIV = 0 THEN                              'initialize HINDEX% and HDIV
    GOSUB GETHINDEX
  ELSE
    HINDEX% = 639 / (TSCREEN / HDIV): HSEC = HDIV
  END IF
  IF HINDEX% > 639 THEN HINDEX% = 639           'check bounds on HINDEX%
  IF HINDEX% < 1 THEN HINDEX% = 1
  FOR N% = 0 TO 639 STEP HINDEX%                'tick marks on screen
    FOR NK% = 0 TO 399 STEP VINDEX&: PSET (N%, NK%), 15: NEXT NK%
    LINE (N% - 1, 400)-(N% + 1, 400), 15        'tick marks on horizontal axis
  NEXT N%

'display screen information
 LOCATE 28, 1: COLOR 7: PRINT "elapsed time:";
 LOCATE 28, 32: COLOR 3: PRINT "P";
 COLOR 7: PRINT USING "lus  FS ######.## mv  "; PFS;
 LOCATE 28, 55: COLOR 3: PRINT "H";
 COLOR 7: PRINT USING "oriz #####.## sec/div"; HSEC;
 LOCATE 29, 1
 IF TSCREEN < 100 THEN PRINT USING "###.## sec/screen"; TSCREEN;
 IF TSCREEN >= 100 THEN PRINT USING "##### sec/screen"; TSCREEN;
 LOCATE 29, 22: COLOR 3: PRINT "E";
 COLOR 7: PRINT "xit";
 LOCATE 29, 32: COLOR 3: PRINT "M";
 COLOR 7: PRINT USING "inus FS ######.## mv"; NFS;
 LOCATE 29, 55: COLOR 3: PRINT "V";
 COLOR 7: PRINT USING "ert ######.## mv/div"; VDIV * 1000;
 FOR N% = 0 TO 5
   COLOR 10 + N%: LOCATE 27, ((N% * 13) + 1): PRINT N%; "=";
 NEXT N%
 IF MISSEDSCAN > 0 THEN
   LOCATE 30, 1: COLOR 7: PRINT "MISSED SCANS   "; MISSEDSCAN;
 END IF
 COLOR 7: LOCATE 30, 50: PRINT "BYTES IN BUFFER";
RETURN

GETHINDEX:
  HSEC = .1
RETRY: HSEC = HSEC * 10: IF TSCREEN / HSEC > 40 THEN GOTO RETRY
DONEHINDEX:
  HINDEX% = 639 / (TSCREEN / HSEC)
RETURN

'********* Input a single character at a time
GETINPUT:
  K$ = ""
GETNEXTKEY:
  S$ = INKEY$
  IF S$ = "" THEN GOTO GETNEXTKEY       'if no key, wait for next key
  IF ASC(S$) = 8 THEN                   'if backspace key
    IF LEN(K$) > 0 THEN
      K$ = LEFT$(K$, LEN(K$) - 1)   'one less character
      LOCATE CSRLIN, (POS(0) - 1): PRINT " "; : LOCATE CSRLIN, (POS(0) - 1)
    END IF
  ELSE
    IF (ASC(S$) >= 43 AND ASC(S$) <= 57) THEN   'char between "." and "9"
      IF ASC(S$) <> 44 AND ASC(S$) <> 47 THEN   'not "," or "/"
        K$ = K$ + S$: PRINT S$;
      END IF
    END IF
  END IF
  IF S$ <> CHR$(13) THEN GOTO GETNEXTKEY
RETURN

'****   get count from scan array and convert to voltage
GETNEXTGRAPHPOINT:
  ANSWER = DARRAY&(II - 1) * .001192093# * GAINFACTOR#
RETURN

EXITGRAPHSCAN:
END SUB

SUB ETRAP
  DIM ER$(30)                        'Error code identification
  DIM ERCODE$(10)
  CLS
  IF COMMAND% < 2 THEN CMD$ = "INVALID"
  IF COMMAND% = 2 THEN CMD$ = "TIME-STAMP SCAN WITH CALIBRATION"
  IF COMMAND% = 3 THEN CMD$ = "VERSION"
  IF COMMAND% = 4 THEN CMD$ = "CHECKSUM"
  IF COMMAND% = 5 THEN CMD$ = "DIGITAL INPUT"
  IF COMMAND% = 6 THEN CMD$ = "NORMAL SCAN"
  IF COMMAND% = 7 THEN CMD$ = "STAMP SCAN"
  IF COMMAND% = 8 THEN CMD$ = "FAST SCAN"
  IF COMMAND% = 9 THEN CMD$ = "INVALID"
  IF COMMAND% = 10 THEN CMD$ = "PORT TEST"
  IF COMMAND% = 11 THEN CMD$ = "OFFSET SCAN"
  IF COMMAND% > 11 THEN CMD$ = "INVALID"
  IF COMMAND% >= &H100 THEN CMD$ = "CHANNEL CODE"
  IF COMMAND% >= &H200 THEN CMD$ = "DIGITAL OUT"
  IF COMMAND% >= &H300 THEN CMD$ = "TRACK/HOLD"
  IF COMMAND% >= &H400 THEN CMD$ = "INVALID"
  IF COMMAND% >= &H500 THEN CMD$ = "SET AVERAGE"
  IF COMMAND% >= &H600 THEN CMD$ = "GET CONVERSION"
  IF COMMAND% >= &H700 THEN CMD$ = "INVALID"
  IF COMMAND% >= &HC00 THEN CMD$ = "SETUP CALIBRATION"
  IF COMMAND% >= &HD00 THEN CMD$ = "SEND CALIBRATION"
  IF COMMAND% >= &HE00 THEN CMD$ = "SET TIME-STAMP SCAN RATE"
  IF COMMAND% >= &HF00 THEN CMD$ = "INVALID"
  IF COMMAND% = &H3000 THEN CMD$ = "SET INTERRUPT"
  IF COMMAND% = &H3001 THEN CMD$ = "START INTERRUPT"
  IF COMMAND% = &H3002 THEN CMD$ = "RESTORE INTERRUPT"
  IF COMMAND% = &H4001 THEN CMD$ = "STOP TIME-STAMPED SCAN"
  IF COMMAND% = &H4002 THEN CMD$ = "CANCEL TIME-STAMPED SCAN"
  IF COMMAND% = &H4003 THEN
    CMD$ = "SEND DIGITAL OUTPUT IN TIME-STAMP SCANNING"
  END IF
  IF COMMAND% = &H4004 THEN
    CMD$ = "REQUEST DIGITAL INPUT IN TIME-STAMP SCANNING"
  END IF
  IF COMMAND% >= &H5000 THEN CMD$ = "GET NUMBER OF WORDS IN BUFFER"
  IF COMMAND% >= &H6000 THEN CMD$ = "FULLSCALE CAL"
  IF COMMAND% >= &H7000 THEN CMD$ = "OFFSET CAL"
  IF COMMAND% < 0 THEN
    COMMAND& = COMMAND% + 65536
    IF COMMAND& >= &H8000& THEN CMD$ = "END TIME-STAMP SCAN"
    IF COMMAND& >= &H9000& THEN CMD$ = "GET TIME-STAMP SCAN"
    IF COMMAND& >= &HA000& THEN CMD$ = "GET CORRECTION FILES"
    IF COMMAND& >= &HB000& THEN CMD$ = "READ LAST BYTE"
    IF COMMAND& >= &HC000& THEN CMD$ = "SET RATE"
    IF COMMAND& >= &HD000& THEN CMD$ = "SET SCAN CHANNELS"
    IF COMMAND& >= &HE000& THEN CMD$ = "SEND SETUP"
    IF COMMAND& >= &HF000& THEN CMD$ = "SET PORT"
  END IF

  ER$(16) = "error bit raised while reading"            '10h
  ER$(17) = "timeout, no new data to read"              '11h
  ER$(18) = "no response to reset, reset failed"        '12h
  ER$(19) = "timeout, not ready to write"               '13h
  ER$(20) = "error bit raised while writing"            '14h
  ER$(21) = "invalid command word"                      '15h
  ER$(22) = "dac data out of range"                     '16h
  ER$(23) = "checksum wrong"                            '17h
  'ER$(24) + (25)  not used
  ER$(26) = "invalid average loop shift counter"        '1Ah
  ER$(27) = "buffer overflow"                           '1Bh
  ER$(28) = "buffer empty"                              '1Ch
  ER$(29) = "digital input pending"                     '1Dh
  ER$(30) = "bad time-stamped scan"                     '1Eh

  ERCODE$(2) = "  ComPending Error"
  ERCODE$(5) = "  Waitpacket Error"
  ERCODE$(7) = "  Scan Overrun Error"
  ERCODE$(10) = "  Buffer Full Error"

  PRINT "an error occured in "; CMD$; " COMMAND"
  PRINT "RESPONSE = "; HEX$(RESPONSE&)
  PRINT : PRINT "ERRORCODE% "; ERRORCODE%; " "; ER$(ERRORCODE%)
  PRINT "  at "; TIME$; "  on  "; DATE$
  IF ERRORCODE% = 16 OR ERRORCODE% = 20 THEN
    IF RESPONSE& < 16 THEN
      ERCODE% = RESPONSE& AND &HF
      PRINT USING "Microcode error ##"; ERCODE%
      PRINT ERCODE$(ERCODE%)
      IF ERCODE% = 10 THEN
        PRINT "The host computer did not read all requested data,"
        PRINT "or it fell behind the Model 250 in scanning mode."
        PRINT "You may need to slow down the A/D data rate."
      END IF
    END IF
  END IF
  PRINT "PRESS ESC TO CONTINUE "
GETERREXIT:
  IF INKEY$ <> CHR$(27) THEN GOTO GETERREXIT
END SUB

SUB LOADCALIBRATIONS
  CLS
  GetCalFilesCMD% = &HA000   'get long-term calibration and factory cal arrays
  FactErr% = 0: CalErr% = 0
  FACTCAL$ = "FACT" + RIGHT$(NEWCAL$, 8)
  LETTER% = ASC(MID$(NEWCAL$, 4, 1))
  SERIALNUM% = VAL(MID$(NEWCAL$, 5, 4))
  PRINT "LOADING CALIBRATION FILES"
  COMMAND% = GetCalFilesCMD% + LETTER%
  SERIALHI% = (SERIALNUM% \ 100)
  SERIALCODE% = SERIALNUM% - SERIALHI% * 100 + SERIALHI% * 256
  RESPONSE& = SERIALCODE%
  CALL M250(COMMAND%, RESPONSE&, ERRORCODE%)
  IF ERRORCODE% <> 0 THEN GOTO FileError       'different error handler
  GOTO LOADCALEXIT

FileError:                'error handler for file errors
  IF (ERRORCODE% AND &HFF) <> 0 THEN
    CLS
    CalErr% = -1
    IF (ERRORCODE% AND &HFF) = 1 THEN
      PRINT "CALIBRATION FILE "; NEWCAL$; " NOT FOUND ";
    ELSE
      PRINT NEWCAL$; " WRONG SIZE, NOT LOADED ";
    END IF
    COLOR 0, 7: PRINT "Accuracy can not be guaranteed"; : COLOR 7, 0
    PRINT : PRINT
    IF NOT (InLongFlag%) THEN
      PRINT "Select long-term calibration to load calibration files"
    END IF
    SHELL "dir 250*.ca? /on /p"
    PRINT "press ESC to continue":
CALERROR:
    IF INKEY$ <> CHR$(27) THEN GOTO CALERROR
  END IF
  IF ERRORCODE% > &HFF THEN
    CLS
    FactErr% = -1
    IF (ERRORCODE% \ &H100) = 1 THEN
      PRINT "FACTORY CAL FILE "; FACTCAL$; " NOT FOUND ";
    ELSE
      PRINT FACTCAL$; " WRONG SIZE, NOT LOADED ";
    END IF
    COLOR 0, 7: PRINT "Accuracy can not be guaranteed"; : COLOR 7, 0
    PRINT : PRINT
    IF NOT (InLongFlag%) THEN
      PRINT "Select long-term calibration to load calibration files"
    END IF
    SHELL "dir FACT*.ca? /on /p"
    PRINT "press ESC to continue":
FACTERROR:
    IF INKEY$ <> CHR$(27) THEN GOTO FACTERROR
  END IF
LOADCALEXIT:
  IF NOT (CalErr%) THEN CURRENTCAL$ = NEWCAL$ ELSE CURRENTCAL$ = ""
END SUB

SUB SELFCAL
 SetCalCMD% = &HC00             'setup long-term calibration process
 StartCalCMD% = &HD00           'start long term self-calibration
 TrackHoldCMD% = &H300          'set sampling mode
SELFCAL1:
  CLS
  InLongFlag% = -1
  LOCATE 24, 20: PRINT "Current calibration ";
  IF CalErr% THEN
    COLOR 0, 7: PRINT "Uncalibrated"; : COLOR 7, 0
  ELSE
    PRINT MID$(CURRENTCAL$, 4, 5);
    IF FactErr% THEN
      LOCATE 25, 19
      COLOR 0, 7: PRINT "No factory calibration found"; : COLOR 7, 0
    END IF
  END IF
  LOCATE 6, 21: PRINT "Long-term Calibration Menu"
  LOCATE 10, 20: PRINT "1) Load calibration file"
  LOCATE 12, 20: PRINT "2) Start self-calibration"
  LOCATE 14, 18: PRINT "Esc) Return to Main Menu";
  A$ = UCASE$(INPUT$(1))
  IF A$ = "1" THEN
    GOSUB CHANGEID
    IF LEN(NEWCAL$) <> 0 AND LEN(S$) <> 0 THEN
      CALL LOADCALIBRATIONS
    END IF
  END IF
  IF A$ = "2" THEN GOSUB STARTSELFCAL
  IF A$ = CHR$(27) THEN A$ = "": InLongFlag% = 0: GOTO ENDSELFCAL
  GOTO SELFCAL1

STARTSELFCAL:
  CLS
  GOSUB CHANGEID
  IF LEN(NEWCAL$) = 0 THEN GOTO EXITSELFCAL
  PRINT
  PRINT "Press C to create new "; NEWCAL$; " file, or any other key for menu"
  A$ = UCASE$(INPUT$(1))
  IF A$ <> "C" THEN GOTO EXITSELFCAL
  ORGCHAN% = CHANNEL%
  HOLD% = 1
  COMMAND% = TrackHoldCMD% + HOLD%          'Set Track/hold mode to HOLD%
  CALL M250(COMMAND%, RESPONSE&, ERRORCODE%)
  IF ERRORCODE% <> 0 THEN GOTO ENDSELFCAL
  GOSUB CREATEBACKUP
  LOCATE 3, 5: PRINT "TEST DATA wil increment until it reaches 0FFF"
  LOCATE 25, 5
  PRINT "This may take an hour. Press any key to cancel calibration";
  TESTDATA% = &H1
  LOCATE 16, 55: PRINT "(IN HEX)"
  LOCATE 15, 46: PRINT "DIFFERENCE"
  LOCATE 15, 63: PRINT "TEST DATA"
  DIM COR%(4095)
  GOSUB SENDCAL
  IF A$ <> "" THEN   'cancel if keypress
    CLS
    PRINT "Calibration canceled.  No new calibration file created"
    PRINT "Press any key to continue": A$ = INPUT$(1)
    GOTO EXITSELFCAL
  END IF
  DEF SEG = VARSEG(COR%(0))
  BSAVE NEWCAL$, VARPTR(COR%(0)), 8192
  DEF SEG
  CURRENTCAL$ = NEWCAL$
EXITSELFCAL:
  ERASE COR%
RETURN

SENDCAL:
  MONODELAY% = 150
  AVRGLOOP% = 1600
  AVRGLOOP1% = 1600

  COMMAND% = SetCalCMD%
  DATA1& = TESTDATA% * 65536
  CALL M250(COMMAND%, DATA1&, ERRORCODE%)
  IF ERRORCODE% <> 0 THEN GOTO ENDSELFCAL
 
  COMMAND% = SetCalCMD% + 1
  DATA1& = AVRGLOOP% + (AVRGLOOP1% * 65536)
  CALL M250(COMMAND%, DATA1&, ERRORCODE%)
  IF ERRORCODE% <> 0 THEN GOTO ENDSELFCAL

  COMMAND% = SetCalCMD% + 2
  DSTARTSEG% = VARSEG(COR%(0))
  COROFFSET% = VARPTR(COR%(0))
  DATA1& = MONODELAY% + (DSTARTSEG% * 65536)
  CALL M250(COMMAND%, DATA1&, ERRORCODE%)
  IF ERRORCODE% <> 0 THEN GOTO ENDSELFCAL

  COMMAND% = StartCalCMD%
  RESPONSE& = COROFFSET%
  CALL M250(COMMAND%, RESPONSE&, ERRORCODE%)
  IF ERRORCODE% <> 0 THEN GOTO ENDSELFCAL
  A$ = INKEY$        'absorb keystroke, if any
RETURN


CHANGEID:
  CLS
  NEWCAL$ = CURRENTCAL$
  PRINT "Current calibration file name:  "; CURRENTCAL$;
  IF CURRENTCAL$ = "" THEN PRINT "NONE"
  LOCATE 3, 1
  PRINT "Enter one letter plus 4 digit serial number "
  INPUT "(or return to leave unchanged) "; S$
  IF S$ <> "" THEN
    L$ = UCASE$(LEFT$(S$, 1))
    IF L$ >= "A" AND L$ <= "Z" THEN
      LETTER% = ASC(L$)
      S% = VAL(RIGHT$(S$, LEN(S$) - 1))
      IF S% >= 0 AND S% < 10000 THEN SERIALNUM% = S%
      S$ = RIGHT$(STR$(SERIALNUM%), LEN(STR$(SERIALNUM%)) - 1)
      TEMPLEN = LEN(S$): FOR N = 1 TO (4 - TEMPLEN): S$ = "0" + S$: NEXT N
      NEWCAL$ = "250" + CHR$(LETTER%) + S$ + ".CAL"
    ELSE
      NEWCAL$ = ""
    END IF
  END IF
RETURN

CREATEBACKUP:
  BACKUP$ = LEFT$(NEWCAL$, (LEN(NEWCAL$) - 1)) + "~"
  PRINT "making backup... ";
  SHELL "COPY " + NEWCAL$ + " " + BACKUP$
  CLS
RETURN

ENDSELFCAL:
END SUB

